
#include <windows.h>
#include <stdio.h>

void ConvertBitmap(LPCSTR lpFile);
void MakeFontsFile(void);
void Bitmap2Bin(LPSTR lpFile);
BOOL ConvertWindowsFont(void);

int __stdcall WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, int)
{
  MakeFontsFile();
  return 0;
}



#define SZ_INIT (sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 8)

#define READ_CHAR(x, y, order) \
  fseek(fout, SZ_INIT + ((order) * 8 * 4), SEEK_SET); \
  for(i = ((y) * 8); i < (((y) * 8) + 8); i++) { \
  fseek(fin, SZ_INIT + (i * 28) + (x), SEEK_SET);  \
  c = getc(fin); \
  c |= 0xFFFFFF00; \
  fwrite(&c, 4, 1, fout); }

#define READ_CHARS(x, y, order) \
  fseek(fout, SZ_INIT + ((order) * 8 * 4), SEEK_SET); \
  for(i = ((y) * 8); i < (((y) * 8) + 8); i++) { \
  fseek(fin, SZ_INIT + (i * 28) + ((x) / 2), SEEK_SET);  \
  c = getc(fin); \
  (x & 0x01) ?  (c <<= 4, c |= 0xFFFFFF0F) : c |= 0xFFFFFF0F; \
  fwrite(&c, 4, 1, fout); }

void ConvertBitmap(LPCSTR lpFile)
{
  FILE            *fin, *fout;
  BITMAPFILEHEADER bfh;
  BITMAPINFOHEADER bih;
  RGBQUAD          pal[2];
  int              i, c, j;
  
  fin = fopen(lpFile, "rb");
  if(!fin) return;
  fout = fopen("_out_convert.bmp", "wb");

  fread(&bfh, sizeof(BITMAPFILEHEADER), 1, fin);
  fwrite(&bfh, sizeof(BITMAPFILEHEADER), 1, fout);
  fread(&bih, sizeof(BITMAPINFOHEADER), 1, fin);
  bih.biWidth     = 32;
  bih.biHeight    = 128 * 8;
  bih.biSizeImage = 0;
  fwrite(&bih, sizeof(BITMAPINFOHEADER), 1, fout);
  fread(pal, 4, 2, fin);
  fwrite(pal, 4, 2, fout);

  c = 0xFF;
  for(i = 0; i < 128; i++)
  {
    putc(0x00, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x7E, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x7E, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x7E, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x7E, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x7E, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x7E, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
    putc(0x00, fout); putc(c, fout); putc(c, fout); putc(c, fout); 
  }
  fflush(fout);

  for(j = 0x09; j < 0x0E; j++)
  { READ_CHAR(0, 5, j); }
  for(j = 0; j < 12; j++)
  { READ_CHAR(j + 4, 5, 0x20 + j); }
  for(j = 0; j < 16; j++)
  { READ_CHAR(j, 4, 0x2C + j); }
  for(j = 0; j < 16; j++)
  { READ_CHAR(j, 3, 0x3C + j); }
  for(j = 0; j < 15; j++)
  { READ_CHAR(j, 2, 0x4C + j); }
  READ_CHAR(j, 2, 0x61);
  for(j = 0; j < 16; j++)
  { READ_CHAR(j, 1, 0x62 + j); }
  for(j = 0; j < 9; j++)
  { READ_CHAR(j, 0, 0x72 + j); }

  fclose(fout);
  fclose(fin);
}


void MakeFontsFile(void)
{
  const DWORD id = 0x004E4647;
  BYTE        buf[2048];
  struct {
    char        name[32];
    DWORD       type;
    UINT        adr;
    COLORREF    c1, c2;
  } fhdr;
  DWORD       n, n1, n2, n3, i;

  const char     ffile1[][64] = {
    "outrun_font.bin",
    "ghouls&ghosts_font.bin",
    "fzone2_font.bin",
    "smonacogp2_font.bin",
    "cassino_games_font.bin",
    "spiderman_font.bin",
    "" };
  const char     fname1[][32] = {
    "Out Run",
    "Ghouls 'N Ghosts",
    "Fantasy Zone 2",
    "Super Monaco GP II",
    "Cassino Games",
    "Spiderman",
    "" };
  const COLORREF fclr1[] = {
    RGB(255, 255, 255), RGB(0, 0, 0),
    RGB(255, 189, 0),   RGB(255, 0, 0),
    RGB(255, 255, 255), RGB(0, 0, 0),
    RGB(255, 255, 255), RGB(128, 128, 128),
    RGB(0, 0, 255),     RGB(255, 255, 255),
    RGB(255, 255, 255), RGB(128, 128, 128)
  };

  const char  ffile2[][64] = {
    "shining_force_2_font.bin",
    "mk2.bin",
    "" };
  const char  fname2[][32] = {
    "Shining Force 2",
    "Mortal Kombat",
    "" };
  const COLORREF fclr2[] = {
    RGB(255, 255, 255),
    RGB(255, 255, 255) };
  
  const char  ffile3[][64] = {
    "default_font.bin", "" };
  const char  fname3[][32] = {
    "Default", "" };
  const COLORREF fclr3[] = {
    RGB(255, 255, 255) };

  FILE *fin, *fout;
  fout = fopen("font.bin", "wb");
  fwrite(&id, 4, 1, fout);
  n = n1 = n2 = n3 = 0;
  while(*ffile1[n1]) n1++, n++;
  while(*ffile2[n2]) n2++, n++;
  while(*ffile3[n3]) n3++, n++;
  fwrite(&n, 4, 1, fout);
  for(i = 0; i < (n * sizeof(fhdr)); i++) putc(0, fout);
  for(i = 0; i < n1; i++)
  {
    fin = fopen(ffile1[i], "rb");
    memset(&fhdr, 0, sizeof(fhdr));
    strcpy(fhdr.name, fname1[i]);
    fhdr.type = 1;
    fhdr.adr  = 8 + (n * sizeof(fhdr)) + (i * 2048);
    fhdr.c1   = fclr1[i * 2];
    fhdr.c2   = fclr1[(i * 2) + 1];
    fseek(fout, 8 + (i * sizeof(fhdr)), SEEK_SET);
    fwrite(&fhdr, sizeof(fhdr), 1, fout);
    fseek(fout, fhdr.adr, SEEK_SET);
    fread(buf, 2048, 1, fin);
    fwrite(buf, 2048, 1, fout);
  }
  for(i = 0; i < n2; i++)
  {
    fin = fopen(ffile2[i], "rb");
    memset(&fhdr, 0, sizeof(fhdr));
    strcpy(fhdr.name, fname2[i]);
    fhdr.type = 2;
    fhdr.adr  = 8 + (n * sizeof(fhdr)) + (n1 * 2048) + (i * 1024);
    fhdr.c1   = fclr2[i];
    fseek(fout, 8 + ((n1 + i) * sizeof(fhdr)), SEEK_SET);
    fwrite(&fhdr, sizeof(fhdr), 1, fout);
    fseek(fout, fhdr.adr, SEEK_SET);
    fread(buf, 1024, 1, fin);
    fwrite(buf, 1024, 1, fout);
  }
  for(i = 0; i < n3; i++)
  {
    fin = fopen(ffile3[i], "rb");
    memset(&fhdr, 0, sizeof(fhdr));
    strcpy(fhdr.name, fname3[i]);
    fhdr.type = 3;
    fhdr.adr  = 8 + (n * sizeof(fhdr)) + (n1 * 2048) + (n2 * 1024) + (i * 512);
    fhdr.c1   = fclr3[i];
    fseek(fout, 8 + ((n1 + n2 + i) * sizeof(fhdr)), SEEK_SET);
    fwrite(&fhdr, sizeof(fhdr), 1, fout);
    fseek(fout, fhdr.adr, SEEK_SET);
    fread(buf, 512, 1, fin);
    fwrite(buf, 512, 1, fout);
  }
  
  fclose(fout);
}


void Bitmap2Bin(LPSTR lpFile)
{
  FILE *fin, *fout, *fbin;
  int   i, j, c;
  char  pszTmp[256];

  fin = fopen(lpFile, "rb");
  if(!fin) return;
  i = strlen(lpFile);
  strcpy(pszTmp, lpFile);
  strcpy(pszTmp + i - 3, "bin");
  fbin = fopen(pszTmp, "wb");
  strcpy(pszTmp + i - 3, "txt");
  fout = fopen(pszTmp, "wt");
  fseek(fin, SZ_INIT, SEEK_SET);

  for(i = 0; i < 128; i++)
  {
    fprintf(fout, "  db ");
    for(j = 0; j < 8; j++)
    {
      c = getc(fin);
      putc(c, fbin);
      getc(fin); getc(fin); getc(fin);
      fprintf(fout, "0x%2.2X", c);
      if(j != 7) fprintf(fout, ", ");
    }
    if(i > 32 && i < 127 && i != '\\' && i != '/') fprintf(fout, " ; %c\n", i);
    else putc('\n', fout);
  }

  fclose(fbin);
  fclose(fout);
  fclose(fin);
}


BOOL ConvertWindowsFont(void)
{
  HFONT   hFont;
  LOGFONT lf;
  ZeroMemory(&lf, sizeof(LOGFONT));
  lf.lfHeight = 14;
  lf.lfWeight = FW_THIN;
  lstrcpy(lf.lfFaceName, "Arial Narrow");
  hFont = CreateFontIndirect(&lf);
  if(!hFont) return 0;
  HBITMAP hBitmap;
  HDC     hDC;
  RECT    rc;
  hBitmap = CreateBitmap(32, 6 * 256, 1, 1, NULL);
  if(!hBitmap) return 0;
  hDC = CreateCompatibleDC(NULL);
  if(!hDC) return 0;
  SelectObject(hDC, hBitmap);
  SelectObject(hDC, hFont);
  SetBkMode(hDC, TRANSPARENT);
  rc.top = 0; rc.bottom = 256 * 6;
  rc.left = 0; rc.right = 32;
  FillRect(hDC, &rc, (HBRUSH)GetStockObject(WHITE_BRUSH));
  char txt[4];
  txt[1] = 0;
  for(int i = 0; i < 256; i++)
  { txt[0] = i; TextOut(hDC, 0, ((256 - i) * 6) - 10, txt, 1); }
  BITMAPFILEHEADER bfh;
  BITMAPINFOHEADER bih;
  RGBQUAD          pal[2];
  ZeroMemory(&bfh, sizeof(BITMAPFILEHEADER));
  ZeroMemory(&bih, sizeof(BITMAPINFOHEADER));
  ZeroMemory(pal, 8);
  bfh.bfType     = MAKEWORD('B', 'M');
  bfh.bfOffBits  = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 8;
  bih.biSize     = sizeof(BITMAPINFOHEADER);
  bih.biPlanes   = 1;
  bih.biWidth    = 32;
  bih.biHeight   = 6 * 256;
  bih.biBitCount = 1;
  pal[1].rgbRed = 0xFF; pal[1].rgbGreen = 0xFF; pal[1].rgbBlue = 0xFF;
  HANDLE hFile;
  hFile = CreateFile("out.bmp", GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE |
    FILE_FLAG_SEQUENTIAL_SCAN, NULL);
  if(hFile == INVALID_HANDLE_VALUE) return 0;
  DWORD dw;
  WriteFile(hFile, &bfh, sizeof(BITMAPFILEHEADER), &dw, NULL);
  WriteFile(hFile, &bih, sizeof(BITMAPINFOHEADER), &dw, NULL);
  WriteFile(hFile, pal, 8, &dw, NULL);
  BYTE  buf[4 * 6 * 256];
  GetDIBits(hDC, hBitmap, 0, 6 * 256, (LPVOID)buf, (LPBITMAPINFO)&bih, DIB_RGB_COLORS);
  WriteFile(hFile, buf, 4 * 6 * 256, &dw, NULL);
  CloseHandle(hFile);
  DeleteDC(hDC);
  DeleteObject(hBitmap);
  DeleteObject(hFont);

  return 1;
}