/* cfgproc.cpp - projeto gens+. Rodrigo Cardoso.
   contm os procedures das janelas relacionadas  configurao do emulador */

#include <windows.h>
#include <shlobj.h>
#include "gmain.h"
#include "gmenu.h"
#include "gintro.h"
#include "grom.h"
#include "config.h"
#include "state.h"
#include "resource.h"

LRESULT CALLBACK OptionProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK DirectoriesProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK FilesProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK InterfaceProc(HWND, UINT, WPARAM, LPARAM);

#define ITEM_LANGUAGE(id, str, def)                     \
{                                                       \
  LANGUAGE_STRING(str, def, Str_Tmp, 1024);             \
  SetDlgItemText(hDlg, id, Str_Tmp);                    \
}

#define COMBO_LANGUAGE(id, str, def)                              \
{                                                                 \
  LANGUAGE_STRING(str, def, Str_Tmp, 1024);                       \
  SendDlgItemMessage(hDlg, id, CB_ADDSTRING, 0, (LPARAM)Str_Tmp); \
}

#define NPAGES 4

void Show_Config_Window(void)
{
  HWND            hPropSheet;
  MSG             msg;
  PROPSHEETHEADER psh;
  PROPSHEETPAGE   ppsp[NPAGES];
  int             i;
  WNDPROC         pProc[NPAGES] =
  { OptionProc, DirectoriesProc, FilesProc, InterfaceProc };
  const char      pszTitleID[NPAGES][32] =
  { "General", "Directories", "Files", "Interface" };
  char            pszMainTitle[32];
  char            pszTitle[NPAGES][32];

  memset(&psh, 0, sizeof(PROPSHEETHEADER));
  psh.dwSize     = sizeof(PROPSHEETHEADER);
  psh.dwFlags    = PSH_USEICONID | PSH_HASHELP | PSH_PROPSHEETPAGE | PSH_MODELESS;
  psh.hwndParent = ghWnd;
  psh.hInstance  = ghInstance;
  psh.pszIcon    = MAKEINTRESOURCE(IDI_GENS00);
  LANGUAGE_STRING("Configuration", "Configuration", pszMainTitle, 32);
  psh.pszCaption = pszMainTitle;
  psh.nPages     = NPAGES;
  psh.ppsp       = ppsp;

  memset(ppsp, 0, sizeof(PROPSHEETPAGE) * NPAGES);
  for(i = 0; i < NPAGES; i++)
  {
    ppsp[i].dwSize      = sizeof(PROPSHEETPAGE);
    ppsp[i].dwFlags     = PSP_HASHELP | PSP_USETITLE;
    ppsp[i].hInstance   = ghInstance;
    ppsp[i].pszTemplate = MAKEINTRESOURCE(IDD_OPTION + i);
    ppsp[i].pfnDlgProc  = (DLGPROC)pProc[i];
    LANGUAGE_STRING(pszTitleID[i], pszTitleID[i], pszTitle[i], 32);
    ppsp[i].pszTitle    = pszTitle[i];
  }

  hPropSheet = (HWND)PropertySheet(&psh);

  if(hPropSheet)
  {
    Center_Dialog(hPropSheet);
    EnableWindow(ghWnd, FALSE);
    while(GetMessage(&msg, NULL, 0, 0))
    {
      if(!SendMessage(hPropSheet, PSM_GETCURRENTPAGEHWND, 0, 0))
      { break; }
      else if(!SendMessage(hPropSheet, PSM_ISDIALOGMESSAGE, 0, (LPARAM)&msg))
      {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
    EnableWindow(ghWnd, TRUE);
    DestroyWindow(hPropSheet);
  }
}


LRESULT CALLBACK OptionProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  DWORD dwID;
  int   res;

  switch(uMsg)
  {
    case WM_INITDIALOG:
      ITEM_LANGUAGE(IDC_GROUP_EMULATOR, "Emulator", "Emulator");
      ITEM_LANGUAGE(IDC_CHECK_AUTOFIXCS, "Auto Fix Checksum", "Autofix checksum");
      ITEM_LANGUAGE(IDC_CHECK_COMPRESS_STATES, "Compress Save States", "Compress save states");
      ITEM_LANGUAGE(IDC_CHECK_CONTROL_CONFIG, "Multiple Control Config", "Use one controller configuration for each game");
      ITEM_LANGUAGE(IDC_CHECK_MULTIPLE_INSTANCES, "Multiple Instances", "Allow multiple instances of the program");
      ITEM_LANGUAGE(IDC_CHECK_COMPRESS_SHOTS, "Compress Screen Shots", "Compress screen shots (GIF)");
      ITEM_LANGUAGE(IDC_CHECK_STATE_PREVIEW, "State files preview", "Add states files preview shot");
      ITEM_LANGUAGE(IDC_CHECK_AUTO_PAUSE, "Auto pause", "Auto pause emulation when lose focus");
      ITEM_LANGUAGE(IDC_CHECK_SEGACD_LED, "Show LED", "Show Sega CD LED");
      ITEM_LANGUAGE(IDC_CHECK_SWITCH_FULLSCREEN, "Switch Full Screen", "Switch to full screen mode when load a rom");
      ITEM_LANGUAGE(IDC_CHECK_DISABLE_SCREENSAVER, "Disable Screen Saver", "Disable screen saver during emulation");
      ITEM_LANGUAGE(IDC_GROUP_INTRO, "Intro Effect", "Intro effect");
      ITEM_LANGUAGE(IDC_STATIC_INTRO_TYPE, "Intro Type", "Type:");
      ITEM_LANGUAGE(IDC_STATIC_INTRO_COLOR, "Intro Color", "Color:");
      ITEM_LANGUAGE(IDC_STATIC_INTRO_NOTE, "Intro Note", "Note: not all intro types uses color.");

      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "No Effect", "No effect");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "Gens Logo", "Gens logo");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "Strange Effect", "Strange effect");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "Genesis Bios Intro", "Genesis bios");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "TV Effect", "TV effect");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "Circle", "Circle");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "Burn", "Burn");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_TYPE, "Smoke", "Smoke");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "White", "White");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Blue", "Blue");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Green", "Green");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Red", "Red");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Yellow", "Yellow");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Pink", "Pink");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Cyan", "Cyan");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Black", "Black");
      COMBO_LANGUAGE(IDC_COMBO_INTRO_COLOR, "Other Color", "Other color...");

      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 0, RGB(0xFF, 0xFF, 0xFF));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 1, RGB(0x00, 0x00, 0xFF));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 2, RGB(0x00, 0xFF, 0x00));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 3, RGB(0xFF, 0x00, 0x00));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 4, RGB(0xFF, 0xFF, 0x00));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 5, RGB(0xFF, 0x00, 0xFF));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 6, RGB(0x00, 0xFF, 0xFF));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 7, RGB(0x00, 0x00, 0x00));
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 8, Intro_Color);

      SendDlgItemMessage(hDlg, IDC_CHECK_AUTOFIXCS, BM_SETCHECK, Auto_Fix_CS ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_COMPRESS_STATES, BM_SETCHECK, Compress_Save_States ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_CONTROL_CONFIG, BM_SETCHECK, Multiple_Controls_Config ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_MULTIPLE_INSTANCES, BM_SETCHECK, Run_Multiple_Instances ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_COMPRESS_SHOTS, BM_SETCHECK, Compress_ScrShots ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_STATE_PREVIEW, BM_SETCHECK, Use_State_MiniShot ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_AUTO_PAUSE, BM_SETCHECK, Auto_Pause ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_SEGACD_LED, BM_SETCHECK, Show_LED ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_SWITCH_FULLSCREEN, BM_SETCHECK, Switch_Full_Screen ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_DISABLE_SCREENSAVER, BM_SETCHECK, Disable_Screen_Saver ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_TYPE, CB_SETCURSEL, Intro_Style, 0);
      switch(Intro_Color)
      {
        case RGB(0xFF, 0xFF, 0xFF): res = 0; break;
        case RGB(0x00, 0x00, 0xFF): res = 1; break;
        case RGB(0x00, 0xFF, 0x00): res = 2; break;
        case RGB(0xFF, 0x00, 0x00): res = 3; break;
        case RGB(0xFF, 0xFF, 0x00): res = 4; break;
        case RGB(0xFF, 0x00, 0xFF): res = 5; break;
        case RGB(0x00, 0xFF, 0xFF): res = 6; break;
        case RGB(0x00, 0x00, 0x00): res = 7; break;
        default:                    res = 8; break;
      }
      SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETCURSEL, res, 0);
      return 1;

    case WM_COMMAND:
      dwID = LOWORD(wParam);
      switch(dwID)
      {
        case IDC_CHECK_AUTOFIXCS:
        case IDC_CHECK_COMPRESS_STATES:
        case IDC_CHECK_CONTROL_CONFIG:
        case IDC_CHECK_MULTIPLE_INSTANCES:
        case IDC_CHECK_COMPRESS_SHOTS:
        case IDC_CHECK_STATE_PREVIEW:
        case IDC_CHECK_AUTO_PAUSE:
        case IDC_CHECK_SEGACD_LED:
        case IDC_CHECK_SWITCH_FULLSCREEN:
        case IDC_CHECK_DISABLE_SCREENSAVER:
          SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;
        case IDC_COMBO_INTRO_TYPE:
          if(HIWORD(wParam) == CBN_SELENDOK)
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
        case IDC_COMBO_INTRO_COLOR:
          if(HIWORD(wParam) == CBN_SELENDOK)
          {
            res = SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_GETCURSEL, 0, 0);
            if(res == 8)
            {
              CHOOSECOLOR cc;
              COLORREF    pal[16];
              pal[0]  = RGB(0xFF, 0xFF, 0xFF); pal[1]  = RGB(0xFF, 0x00, 0x00);
              pal[2]  = RGB(0xFF, 0x80, 0x00); pal[3]  = RGB(0xFF, 0xFF, 0x00);
              pal[4]  = RGB(0x80, 0xFF, 0x00); pal[5]  = RGB(0x00, 0xFF, 0x00);
              pal[6]  = RGB(0x00, 0xFF, 0x80); pal[7]  = RGB(0x00, 0xFF, 0xFF);
              pal[8]  = RGB(0x00, 0x00, 0x00); pal[9]  = RGB(0x80, 0x80, 0x80);
              pal[10] = RGB(0x00, 0x80, 0xFF); pal[11] = RGB(0x00, 0x00, 0xFF);
              pal[12] = RGB(0x80, 0x00, 0xFF); pal[13] = RGB(0xFF, 0x00, 0xFF);
              pal[14] = RGB(0xFF, 0x00, 0x80); pal[15] = RGB(0x80, 0x00, 0x80);
              memset(&cc, 0, sizeof(CHOOSECOLOR));
              cc.lStructSize  = sizeof(CHOOSECOLOR);
              cc.hwndOwner    = hDlg;
              cc.Flags        = CC_FULLOPEN | CC_RGBINIT;
              cc.rgbResult    = SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_GETITEMDATA, 8, 0);
              cc.lpCustColors = pal;
              if(ChooseColor(&cc))
              { SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_SETITEMDATA, 8, cc.rgbResult);
                SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0); }
            }
            else
              SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          }
          break;
      }
      break;

    case WM_NOTIFY:
      if(((LPNMHDR)lParam)->code == PSN_APPLY)
      {
        Auto_Fix_CS = SendDlgItemMessage(hDlg, IDC_CHECK_AUTOFIXCS, BM_GETCHECK, 0, 0);
        Compress_Save_States = SendDlgItemMessage(hDlg, IDC_CHECK_COMPRESS_STATES, BM_GETCHECK, 0, 0);
        Multiple_Controls_Config = SendDlgItemMessage(hDlg, IDC_CHECK_CONTROL_CONFIG, BM_GETCHECK, 0, 0);
        Run_Multiple_Instances = SendDlgItemMessage(hDlg, IDC_CHECK_MULTIPLE_INSTANCES, BM_GETCHECK,  0, 0);
        Compress_ScrShots = SendDlgItemMessage(hDlg, IDC_CHECK_COMPRESS_SHOTS, BM_GETCHECK, 0, 0);
        Use_State_MiniShot = SendDlgItemMessage(hDlg, IDC_CHECK_STATE_PREVIEW, BM_GETCHECK, 0, 0);
        Auto_Pause = SendDlgItemMessage(hDlg, IDC_CHECK_AUTO_PAUSE, BM_GETCHECK, 0, 0);
        Show_LED = SendDlgItemMessage(hDlg, IDC_CHECK_SEGACD_LED, BM_GETCHECK, 0, 0);
        Switch_Full_Screen = SendDlgItemMessage(hDlg, IDC_CHECK_SWITCH_FULLSCREEN, BM_GETCHECK, 0, 0);
        Disable_Screen_Saver = SendDlgItemMessage(hDlg, IDC_CHECK_DISABLE_SCREENSAVER, BM_GETCHECK, 0, 0);
        res = SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_TYPE, CB_GETCURSEL, 0, 0);
        if(res > -1) Intro_Style = res;
        res = SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_GETCURSEL, 0, 0);
        if(res > -1)
        { Intro_Color = SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_GETITEMDATA, res, 0);
          Intro_ColorW = Color_24_to_16(Intro_Color); }
      }
      break;

    case WM_DRAWITEM:
      if(LOWORD(wParam) == IDC_COMBO_INTRO_COLOR)
      {
        int              x, y;
        LPDRAWITEMSTRUCT pdis;
        COLORREF         clrItem, clrForeground, clrBackground;
        TEXTMETRIC       tm;
        HBRUSH           hBrush;
        RECT             rc;

        dwID = LOWORD(wParam);
        pdis = (LPDRAWITEMSTRUCT)lParam;
        if(pdis->itemID == -1) return 1;

        clrForeground = SetTextColor(pdis->hDC,
          GetSysColor(pdis->itemState & ODS_SELECTED ?
          COLOR_HIGHLIGHTTEXT : COLOR_WINDOWTEXT));
        clrBackground = SetBkColor(pdis->hDC,
          GetSysColor(pdis->itemState & ODS_SELECTED ?
          COLOR_HIGHLIGHT : COLOR_WINDOW));
        GetTextMetrics(pdis->hDC, &tm);
        y = (pdis->rcItem.bottom + pdis->rcItem.top -  tm.tmHeight) / 2;
        x = LOWORD(GetDialogBaseUnits()) * 4;
        SendMessage(pdis->hwndItem, CB_GETLBTEXT, pdis->itemID, (LPARAM)Str_Tmp);
        ExtTextOut(pdis->hDC, x, y, ETO_CLIPPED | ETO_OPAQUE, &pdis->rcItem,
          Str_Tmp, strlen(Str_Tmp), NULL);

        clrItem   = SendDlgItemMessage(hDlg, IDC_COMBO_INTRO_COLOR, CB_GETITEMDATA,
                                       pdis->itemID, 0);
        rc.left   = LOWORD(GetDialogBaseUnits()) / 2;
        rc.right  = x - (rc.left * 2);
        y         = HIWORD(GetDialogBaseUnits()) / 8;
        rc.top    = pdis->rcItem.top + y;
        rc.bottom = pdis->rcItem.bottom - y;
        hBrush = CreateSolidBrush(clrItem);
        if(hBrush)
        { FillRect(pdis->hDC, &rc, hBrush); DeleteObject(hBrush); }
        hBrush = CreateSolidBrush(GetSysColor(COLOR_WINDOWTEXT));
        if(hBrush)
        { FrameRect(pdis->hDC, &rc, hBrush); DeleteObject(hBrush); }
        return 1;
      }
  }

  return 0;
}


#define GETDIR(id, buf)                                                 \
 len = GetDlgItemText(hDlg, id, buf, 1024);                             \
 if(len > 0)                                                            \
 { if(buf[len - 1] != '\\' && buf[len - 1] != '/')                      \
   { buf[len++] = '\\'; buf[len++] = 0; } }                             \
 else if(len == 0)                                                      \
 { strcpy(buf, ".\\"); }

LRESULT CALLBACK DirectoriesProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  DWORD dwID;

  switch(uMsg)
  {
    case WM_INITDIALOG:
      ITEM_LANGUAGE(IDC_GROUP_MAIN_DIR, "Emulator Folders", "Emulator folders");
      ITEM_LANGUAGE(IDC_STATIC_SAVE, "Save static", "Save &state:");
      ITEM_LANGUAGE(IDC_STATIC_SRAM, "Sram static", "S&RAM backup:");
      ITEM_LANGUAGE(IDC_STATIC_BRAM, "Bram static", "&BRAM backup:");
      ITEM_LANGUAGE(IDC_STATIC_DUMP, "Dump static", "Sound &dump:");
      ITEM_LANGUAGE(IDC_STATIC_SHOT, "Shot static", "Scr&een shot:");
      ITEM_LANGUAGE(IDC_STATIC_PATCH, "Patch static", "&PAT patch:");
      ITEM_LANGUAGE(IDC_STATIC_IPS, "IPS static", "&IPS patch:");
      ITEM_LANGUAGE(IDC_STATIC_CHEATS, "Cheats static", "&Cheats:");
      ITEM_LANGUAGE(IDC_STATIC_CONFIG, "Config static", "C&onfig:");
      ITEM_LANGUAGE(IDC_STATIC_ROMS_DIR, "Roms Folder", "Roms:");
      ITEM_LANGUAGE(IDC_CHECK_LAST_OPEN_ROM, "Use Last Opened Rom Directory", "Use last opened rom directory");

      SendDlgItemMessage(hDlg, IDC_EDIT_SAVE, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_SRAM, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_BRAM, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_DUMP, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_SHOT, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_PATCH, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_IPS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_CHEATS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_CONFIG, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_ROMS_DIR, EM_LIMITTEXT, GMAX_PATH - 1, 0);

      SetDlgItemText(hDlg, IDC_EDIT_SAVE, State_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_SRAM, SRAM_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_BRAM, BRAM_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_DUMP, Dump_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_SHOT, ScrShot_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_PATCH, Patch_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_IPS, IPS_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_CHEATS, Cheats_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_CONFIG, Config_Dir);
      SetDlgItemText(hDlg, IDC_EDIT_ROMS_DIR, Rom_Dir);
      SendDlgItemMessage(hDlg, IDC_CHECK_LAST_OPEN_ROM, BM_SETCHECK, Use_Last_Open_Rom_Dir ? BST_CHECKED : BST_UNCHECKED, 0);
      return 1;

    case WM_COMMAND:
      dwID = LOWORD(wParam);
      switch(dwID)
      {
        case IDC_EDIT_SAVE:
        case IDC_EDIT_SRAM:
        case IDC_EDIT_BRAM:
        case IDC_EDIT_DUMP:
        case IDC_EDIT_SHOT:
        case IDC_EDIT_PATCH:
        case IDC_EDIT_IPS:
        case IDC_EDIT_CHEATS:
        case IDC_EDIT_CONFIG:
        case IDC_EDIT_ROMS_DIR:
          if(HIWORD(wParam) == EN_CHANGE)
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;

        case IDC_CHECK_LAST_OPEN_ROM:
          SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;

        case IDC_CHANGE_SAVE:
          GetDlgItemText(hDlg, IDC_EDIT_SAVE, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "Save state directory"))
            SetDlgItemText(hDlg, IDC_EDIT_SAVE, Str_Tmp);
          break;

        case IDC_CHANGE_SRAM:
          GetDlgItemText(hDlg, IDC_EDIT_SRAM, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "SRAM backup directory"))
            SetDlgItemText(hDlg, IDC_EDIT_SRAM, Str_Tmp);
          break;

        case IDC_CHANGE_BRAM:
          GetDlgItemText(hDlg, IDC_EDIT_BRAM, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "BRAM backup directory"))
            SetDlgItemText(hDlg, IDC_EDIT_BRAM, Str_Tmp);
          break;

        case IDC_CHANGE_DUMP:
          GetDlgItemText(hDlg, IDC_EDIT_DUMP, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "Sound dump directory"))
            SetDlgItemText(hDlg, IDC_EDIT_DUMP, Str_Tmp);
          break;

        case IDC_CHANGE_SHOT:
          GetDlgItemText(hDlg, IDC_EDIT_SHOT, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "Screen-shot directory"))
            SetDlgItemText(hDlg, IDC_EDIT_SHOT, Str_Tmp);
          break;

        case IDC_CHANGE_PATCH:
          GetDlgItemText(hDlg, IDC_EDIT_PATCH, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "PAT Patch directory"))
            SetDlgItemText(hDlg, IDC_EDIT_PATCH, Str_Tmp);
          break;

        case IDC_CHANGE_IPS:
          GetDlgItemText(hDlg, IDC_EDIT_IPS, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "IPS Patch directory"))
            SetDlgItemText(hDlg, IDC_EDIT_IPS, Str_Tmp);
          break;

        case IDC_CHANGE_CHEATS:
          GetDlgItemText(hDlg, IDC_EDIT_CHEATS, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "Cheats directory"))
            SetDlgItemText(hDlg, IDC_EDIT_CHEATS, Str_Tmp);
          break;

        case IDC_CHANGE_CONFIG:
          GetDlgItemText(hDlg, IDC_EDIT_CONFIG, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "Config directory"))
            SetDlgItemText(hDlg, IDC_EDIT_CONFIG, Str_Tmp);
          break;

        case IDC_CHANGE_ROMS_DIR:
          GetDlgItemText(hDlg, IDC_EDIT_ROMS_DIR, Str_Tmp, 1024);
          if(Select_Dir(hDlg, Str_Tmp, "Config directory"))
            SetDlgItemText(hDlg, IDC_EDIT_ROMS_DIR, Str_Tmp);
          break;
      }
      break;

    case WM_NOTIFY:
      if(((LPNMHDR)lParam)->code == PSN_APPLY)
      {
        int len;
        GETDIR(IDC_EDIT_SAVE, State_Dir);
        GETDIR(IDC_EDIT_SRAM, SRAM_Dir);
        GETDIR(IDC_EDIT_BRAM, BRAM_Dir);
        GETDIR(IDC_EDIT_DUMP, Dump_Dir);
        GETDIR(IDC_EDIT_SHOT, ScrShot_Dir);
        GETDIR(IDC_EDIT_PATCH, Patch_Dir);
        GETDIR(IDC_EDIT_IPS, IPS_Dir);
        GETDIR(IDC_EDIT_CHEATS, Cheats_Dir);
        GETDIR(IDC_EDIT_CONFIG, Config_Dir);
        GETDIR(IDC_EDIT_ROMS_DIR, Rom_Dir);
        Use_Last_Open_Rom_Dir =
          SendDlgItemMessage(hDlg, IDC_CHECK_LAST_OPEN_ROM, BM_GETCHECK, 0, 0);
      }
      break;
  }

  return 0;
}


LRESULT CALLBACK FilesProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  DWORD dwID;
  char Str_Tmp2[1024];

  switch(uMsg)
  {
    case WM_INITDIALOG:
      ITEM_LANGUAGE(IDC_GENESISBIOS_FILE, "Genesis Bios File", "Genesis bios file");
      ITEM_LANGUAGE(IDC_32XBIOS_FILES, "32X Bios Files", "32X bios files");
      ITEM_LANGUAGE(IDC_CDBIOS_FILES, "SEGA CD Bios Files", "SEGA CD bios files");
      ITEM_LANGUAGE(IDC_MISC_FILES, "Other Files", "Other files");
      ITEM_LANGUAGE(IDC_STATIC_GENESISBIOS, "Genesis Bios", "Genesis:");
      ITEM_LANGUAGE(IDC_STATIC_32XGBIOS, "M68000 Bios", "M68000:");
      ITEM_LANGUAGE(IDC_STATIC_32XMBIOS, "M SH2 Bios", "Master SH2:");
      ITEM_LANGUAGE(IDC_STATIC_32XSBIOS, "S SH2 Bios", "Slave SH2:");
      ITEM_LANGUAGE(IDC_STATIC_USBIOS, "US bios", "USA:");
      ITEM_LANGUAGE(IDC_STATIC_EUBIOS, "EU bios", "Europe:");
      ITEM_LANGUAGE(IDC_STATIC_JABIOS, "JA bios", "Japan:");
      ITEM_LANGUAGE(IDC_STATIC_CGOFFLINE, "CGOffline", "CGOffline:");
      ITEM_LANGUAGE(IDC_STATIC_MANUAL, "Manual", "Manual:");

      SendDlgItemMessage(hDlg, IDC_EDIT_GENESISBIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_32XGBIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_32XMBIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_32XSBIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_USBIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_EUBIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_JABIOS, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_CGOFFLINE, EM_LIMITTEXT, GMAX_PATH - 1, 0);
      SendDlgItemMessage(hDlg, IDC_EDIT_MANUAL, EM_LIMITTEXT, GMAX_PATH - 1, 0);

      SetDlgItemText(hDlg, IDC_EDIT_GENESISBIOS, Genesis_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_32XGBIOS, _32X_Genesis_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_32XMBIOS, _32X_Master_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_32XSBIOS, _32X_Slave_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_USBIOS, US_CD_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_EUBIOS, EU_CD_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_JABIOS, JA_CD_Bios);
      SetDlgItemText(hDlg, IDC_EDIT_CGOFFLINE, CGOffline_Path);
      SetDlgItemText(hDlg, IDC_EDIT_MANUAL, Manual_Path);
      return 1;

    case WM_COMMAND:
      dwID = LOWORD(wParam);
      switch(dwID)
      {
        case IDC_EDIT_GENESISBIOS:
        case IDC_EDIT_32XGBIOS:
        case IDC_EDIT_32XMBIOS:
        case IDC_EDIT_32XSBIOS:
        case IDC_EDIT_USBIOS:
        case IDC_EDIT_EUBIOS:
        case IDC_EDIT_JABIOS:
        case IDC_EDIT_CGOFFLINE:
        case IDC_EDIT_MANUAL:
          if(HIWORD(wParam) == EN_CHANGE)
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;

        case ID_CHANGE_GENESISBIOS:
          GetDlgItemText(hDlg, IDC_EDIT_GENESISBIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "Genesis_OS_ROM.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "Genesis bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_GENESISBIOS, Str_Tmp);
          break;

        case ID_CHANGE_32XGBIOS:
          GetDlgItemText(hDlg, IDC_EDIT_32XGBIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "32X_G_bios.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "32X M68000 bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_32XGBIOS, Str_Tmp);
          break;

        case ID_CHANGE_32XMBIOS:
          GetDlgItemText(hDlg, IDC_EDIT_32XMBIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "32X_M_bios.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "32X Master SH2 bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_32XMBIOS, Str_Tmp);
          break;

        case ID_CHANGE_32XSBIOS:
          GetDlgItemText(hDlg, IDC_EDIT_32XSBIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "32X_S_bios.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "32X Slave SH2 bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_32XSBIOS, Str_Tmp);
          break;

        case ID_CHANGE_USBIOS:
          GetDlgItemText(hDlg, IDC_EDIT_USBIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "us_scd1_9210.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "USA CD bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_USBIOS, Str_Tmp);
          break;

        case ID_CHANGE_EUBIOS:
          GetDlgItemText(hDlg, IDC_EDIT_EUBIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "eu_mcd1_9210.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "EUROPEAN CD bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_EUBIOS, Str_Tmp);
          break;

        case ID_CHANGE_JABIOS:
          GetDlgItemText(hDlg, IDC_EDIT_JABIOS, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "jp_mcd1_9112.bin");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "JAPAN CD bios file", "bios files\0*.bin\0\0", "bin"))
            SetDlgItemText(hDlg, IDC_EDIT_JABIOS, Str_Tmp);
          break;

        case ID_CHANGE_CGOFFLINE:
          GetDlgItemText(hDlg, IDC_EDIT_CGOFFLINE, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "GCOffline.chm");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "Genesis Collective - CGOffline file", "html help files\0*.chm\0\0", "chm"))
            SetDlgItemText(hDlg, IDC_EDIT_CGOFFLINE, Str_Tmp);
          break;

        case ID_CHANGE_MANUAL:
          GetDlgItemText(hDlg, IDC_EDIT_MANUAL, Str_Tmp2, GMAX_PATH);
          strcpy(Str_Tmp, "Manual.exe");
          if (Change_File_L(Str_Tmp, Str_Tmp2, "Gens Manual", "executable files\0*.exe\0\0", "exe"))
            SetDlgItemText(hDlg, IDC_EDIT_MANUAL, Str_Tmp);
          break;

      }
      break;

    case WM_NOTIFY:
      if(((LPNMHDR)lParam)->code == PSN_APPLY)
      {
        GetDlgItemText(hDlg, IDC_EDIT_GENESISBIOS, Genesis_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_32XGBIOS, _32X_Genesis_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_32XMBIOS, _32X_Master_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_32XSBIOS, _32X_Slave_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_USBIOS, US_CD_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_EUBIOS, EU_CD_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_JABIOS, JA_CD_Bios, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_CGOFFLINE, CGOffline_Path, GMAX_PATH);
        GetDlgItemText(hDlg, IDC_EDIT_MANUAL, Manual_Path, GMAX_PATH);
      }
      break;
  }

  return 0;
}


LRESULT CALLBACK InterfaceProc(HWND hDlg, UINT uMes, WPARAM wParam, LPARAM lParam)
{
  static FILE *fFont;
  DWORD  dwID;
  int    i;

  switch(uMes)
  {
    case WM_INITDIALOG:
      ITEM_LANGUAGE(IDC_GROUP_FONTS, "Fonts", "Fonts");
      ITEM_LANGUAGE(IDC_STATIC_FONTS, "Fonts Message", "Choose the fonts to print messages on the screen");
      i = LANGUAGE_STRING("Big Font 1", "Big font 1", Str_Tmp, sizeof(Str_Tmp));
      SetDlgItemText(hDlg, IDC_STATIC_FONT1, Str_Tmp);
      if(i--) { if(Str_Tmp[i] == ':') Str_Tmp[i] = 0; }
      SendDlgItemMessage(hDlg, IDC_COMBO_MESSAGE_FONT, CB_ADDSTRING, 0, (LPARAM)Str_Tmp);
      SendDlgItemMessage(hDlg, IDC_COMBO_FPS_FONT, CB_ADDSTRING, 0, (LPARAM)Str_Tmp);
      ITEM_LANGUAGE(IDC_STATIC_COLOR1, "Font 1 Colors", "Colors:");
      i = LANGUAGE_STRING("Big Font 2", "Big font 2", Str_Tmp, sizeof(Str_Tmp));
      SetDlgItemText(hDlg, IDC_STATIC_FONT2, Str_Tmp);
      if(i--) { if(Str_Tmp[i] == ':') Str_Tmp[i] = 0; }
      SendDlgItemMessage(hDlg, IDC_COMBO_MESSAGE_FONT, CB_ADDSTRING, 0, (LPARAM)Str_Tmp);
      SendDlgItemMessage(hDlg, IDC_COMBO_FPS_FONT, CB_ADDSTRING, 0, (LPARAM)Str_Tmp);
      ITEM_LANGUAGE(IDC_STATIC_COLOR2, "Font 2 Color", "Color:");
      i = LANGUAGE_STRING("Small Font", "Small font", Str_Tmp, sizeof(Str_Tmp));
      SetDlgItemText(hDlg, IDC_STATIC_FONT3, Str_Tmp);
      if(i--) { if(Str_Tmp[i] == ':') Str_Tmp[i] = 0; }
      SendDlgItemMessage(hDlg, IDC_COMBO_MESSAGE_FONT, CB_ADDSTRING, 0, (LPARAM)Str_Tmp);
      SendDlgItemMessage(hDlg, IDC_COMBO_FPS_FONT, CB_ADDSTRING, 0, (LPARAM)Str_Tmp);
      ITEM_LANGUAGE(IDC_STATIC_COLOR3, "Font 3 Color", "Color:");
      ITEM_LANGUAGE(IDC_GROUP_MESSAGE, "Messages", "Messages");
      ITEM_LANGUAGE(IDC_STATIC_MESSAGE_FONT, "Message Font", "Font:");
      ITEM_LANGUAGE(IDC_CHECK_ENABLE_MESSAGE, "Enable Message", "Enable");
      ITEM_LANGUAGE(IDC_CHECK_TRANSPARENT_MESSAGE, "Transparent Message", "Transparent");
      ITEM_LANGUAGE(IDC_GROUP_FPS, "FPS", "FPS");
      ITEM_LANGUAGE(IDC_STATIC_FPS_FONT, "FPS Font", "Font:");
      ITEM_LANGUAGE(IDC_CHECK_ENABLE_FPS, "Enable FPS", "Enable");
      ITEM_LANGUAGE(IDC_CHECK_TRANSPARENT_FPS, "Transparent FPS", "Transparent");


      fFont = Get_Fonts_Info(hDlg);
      SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR11), GWL_USERDATA, Font1_Color1);
      SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR12), GWL_USERDATA, Font1_Color2);
      SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR2), GWL_USERDATA, Font2_Color);
      SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR3), GWL_USERDATA, Font3_Color);
      SendDlgItemMessage(hDlg, IDC_COMBO_MESSAGE_FONT, CB_SETCURSEL, Message_Font - 1, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_ENABLE_MESSAGE, BM_SETCHECK, Show_Message ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_TRANSPARENT_MESSAGE, BM_SETCHECK, Message_Transp ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_FPS_FONT, CB_SETCURSEL, FPS_Font - 1, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_ENABLE_FPS, BM_SETCHECK, Show_FPS ? BST_CHECKED : BST_UNCHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_TRANSPARENT_FPS, BM_SETCHECK, FPS_Transp ? BST_CHECKED : BST_UNCHECKED, 0);
      return 1;

    case WM_DESTROY:
      if(fFont) fclose(fFont), fFont = NULL;
      break;

    case WM_COMMAND:
      dwID = LOWORD(wParam);
      switch(dwID)
      {
        case IDC_COMBO_FONT1:
        case IDC_COMBO_FONT2:
        case IDC_COMBO_FONT3:
          if(HIWORD(wParam) == CBN_SELENDOK && fFont)
          {
            FONTHDR fHdr;
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
            i = SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0);
            if(i < 0) break;
            i = SendMessage((HWND)lParam, CB_GETITEMDATA, i, 0);
            if(fseek(fFont, i, SEEK_SET)) break;
            if(fread(&fHdr, sizeof(FONTHDR), 1, fFont) != 1) break;
            switch(dwID)
            {
              case IDC_COMBO_FONT1:
                SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR11), GWL_USERDATA, fHdr.c1);
                SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR12), GWL_USERDATA, fHdr.c2);
                InvalidateRect(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR11), NULL, TRUE);
                InvalidateRect(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR12), NULL, TRUE);
                break;
              case IDC_COMBO_FONT2:
                SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR2), GWL_USERDATA, fHdr.c1);
                InvalidateRect(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR2), NULL, TRUE);
                break;
              case IDC_COMBO_FONT3:
                SetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR3), GWL_USERDATA, fHdr.c1);
                InvalidateRect(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR3), NULL, TRUE);
                break;
            }
          }
          break;

        case IDC_STATIC_CHOOSE_COLOR11:
        case IDC_STATIC_CHOOSE_COLOR12:
        case IDC_STATIC_CHOOSE_COLOR2:
        case IDC_STATIC_CHOOSE_COLOR3:
          if(HIWORD(wParam) == STN_CLICKED)
          {
            CHOOSECOLOR cc;
            COLORREF    pal[16];
            pal[0]  = RGB(0xFF, 0xFF, 0xFF); pal[1]  = RGB(0xFF, 0x00, 0x00);
            pal[2]  = RGB(0xFF, 0x80, 0x00); pal[3]  = RGB(0xFF, 0xFF, 0x00);
            pal[4]  = RGB(0x80, 0xFF, 0x00); pal[5]  = RGB(0x00, 0xFF, 0x00);
            pal[6]  = RGB(0x00, 0xFF, 0x80); pal[7]  = RGB(0x00, 0xFF, 0xFF);
            pal[8]  = RGB(0x00, 0x00, 0x00); pal[9]  = RGB(0x80, 0x80, 0x80);
            pal[10] = RGB(0x00, 0x80, 0xFF); pal[11] = RGB(0x00, 0x00, 0xFF);
            pal[12] = RGB(0x80, 0x00, 0xFF); pal[13] = RGB(0xFF, 0x00, 0xFF);
            pal[14] = RGB(0xFF, 0x00, 0x80); pal[15] = RGB(0x80, 0x00, 0x80);
            memset(&cc, 0, sizeof(CHOOSECOLOR));
            cc.lStructSize  = sizeof(CHOOSECOLOR);
            cc.hwndOwner    = hDlg;
            cc.Flags        = CC_FULLOPEN | CC_RGBINIT;
            cc.rgbResult    = GetWindowLong(GetDlgItem(hDlg, dwID), GWL_USERDATA);
            cc.lpCustColors = pal;
            if(ChooseColor(&cc))
            { SetWindowLong((HWND)lParam, GWL_USERDATA, cc.rgbResult);
              InvalidateRect((HWND)lParam, NULL, TRUE);
              SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0); }
          }
          break;

        case IDC_COMBO_MESSAGE_FONT:
        case IDC_COMBO_FPS_FONT:
          if(HIWORD(wParam) == CBN_SELENDOK)
            SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;

        case IDC_CHECK_ENABLE_MESSAGE:
        case IDC_CHECK_TRANSPARENT_MESSAGE:
        case IDC_CHECK_ENABLE_FPS:
        case IDC_CHECK_TRANSPARENT_FPS:
          SendMessage(GetParent(hDlg), PSM_CHANGED, (WPARAM)hDlg, 0);
          break;
      }
      break;

    case WM_CTLCOLORSTATIC:
      dwID = GetDlgCtrlID((HWND)lParam);
      switch(dwID)
      {
        case IDC_STATIC_CHOOSE_COLOR11:
        case IDC_STATIC_CHOOSE_COLOR12:
        case IDC_STATIC_CHOOSE_COLOR2:
        case IDC_STATIC_CHOOSE_COLOR3:
          return (LRESULT)CreateSolidBrush(GetWindowLong((HWND)lParam, GWL_USERDATA));
      }
      break;

    case WM_NOTIFY:
      if(((LPNMHDR)lParam)->code == PSN_APPLY)
      {
        GetDlgItemText(hDlg, IDC_COMBO_FONT1, Font1_Name, GFONT_NAME_SIZE);
        Font1_Color1   = GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR11), GWL_USERDATA);
        Font1_Color2   = GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR12), GWL_USERDATA);
        Font1_Color1W  = Color_24_to_16(Font1_Color1);
        Font1_Color2W  = Color_24_to_16(Font1_Color2);
        GetDlgItemText(hDlg, IDC_COMBO_FONT2, Font2_Name, GFONT_NAME_SIZE);
        Font2_Color    = GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR2), GWL_USERDATA);
        Font2_ColorW   = Color_24_to_16(Font2_Color);
        GetDlgItemText(hDlg, IDC_COMBO_FONT3, Font3_Name, GFONT_NAME_SIZE);
        Font3_Color    = GetWindowLong(GetDlgItem(hDlg, IDC_STATIC_CHOOSE_COLOR3), GWL_USERDATA);
        Load_Fonts_From_File(fFont, FALSE);
        Font3_ColorW   = Color_24_to_16(Font3_Color);
        Message_Font   = SendDlgItemMessage(hDlg, IDC_COMBO_MESSAGE_FONT, CB_GETCURSEL, 0, 0) + 1;
        Show_Message   = SendDlgItemMessage(hDlg, IDC_CHECK_ENABLE_MESSAGE, BM_GETCHECK, 0, 0);
        Message_Transp = SendDlgItemMessage(hDlg, IDC_CHECK_TRANSPARENT_MESSAGE, BM_GETCHECK, 0, 0);
        if(Message_Transp)
        {
          switch(Message_Font)
          { case 2:  Print_Message = Print_Text2_Transp; Message_Color = (DWORD)Font2_ColorW; break;
            case 3:  Print_Message = Print_Text3_Transp; Message_Color = (DWORD)Font3_ColorW; break;
            default: Print_Message = Print_Text1_Transp; Message_Color = MAKELONG(Font1_Color1W, Font1_Color2W); break; }
        }
        else
        {
          switch(Message_Font)
          { case 2:  Print_Message = Print_Text2; Message_Color = (DWORD)Font2_ColorW; break;
            case 3:  Print_Message = Print_Text3; Message_Color = (DWORD)Font3_ColorW; break;
            default: Print_Message = Print_Text1; Message_Color = MAKELONG(Font1_Color1W, Font1_Color2W); break; }
        }
        FPS_Font       = SendDlgItemMessage(hDlg, IDC_COMBO_FPS_FONT, CB_GETCURSEL, 0, 0) + 1;
        Show_FPS       = SendDlgItemMessage(hDlg, IDC_CHECK_ENABLE_FPS, BM_GETCHECK, 0, 0);
        FPS_Transp     = SendDlgItemMessage(hDlg, IDC_CHECK_TRANSPARENT_FPS, BM_GETCHECK, 0, 0);
        if(FPS_Transp)
        {
          switch(FPS_Font)
          { case 2:  Print_FPS = Print_Text2_Transp; FPS_Color = (DWORD)Font2_ColorW; break;
            case 3:  Print_FPS = Print_Text3_Transp; FPS_Color = (DWORD)Font3_ColorW; break;
            default: Print_FPS = Print_Text1_Transp; FPS_Color = MAKELONG(Font1_Color1W, Font1_Color2W); break; }
        }
        else
        {
          switch(FPS_Font)
          { case 2:  Print_FPS = Print_Text2; FPS_Color = (DWORD)Font2_ColorW; break;
            case 3:  Print_FPS = Print_Text3; FPS_Color = (DWORD)Font3_ColorW; break;
            default: Print_FPS = Print_Text1; FPS_Color = MAKELONG(Font1_Color1W, Font1_Color2W); break; }
        }
      }
      break;
  }

  return 0;
}


BOOL Select_Dir(HWND hParent, LPSTR pDest, LPCSTR pTitle)
{
  BROWSEINFO   bi;
  LPITEMIDLIST piid;
  LPMALLOC     pm;
  int          i, c = 0;

  memset(&bi, 0, sizeof(BROWSEINFO));
  bi.hwndOwner      = hParent;
  bi.pszDisplayName = Str_Tmp;
  bi.lpszTitle      = pTitle;
  bi.ulFlags        = BIF_RETURNONLYFSDIRS;
  piid = SHBrowseForFolder(&bi);
  if(piid)
  {
    if(SHGetPathFromIDList(piid, pDest))
    {
      i = strlen(pDest);
      if(i)
      {
        c = pDest[i - 1];
        if(c != '\\' && c != '/' && c != ':')
        { pDest[i++] = '\\'; pDest[i] = 0; }
        c = 1;
      }
    }
    if(!SHGetMalloc(&pm))
    { pm->Free(piid); pm->Release(); }
  }

  return c;
}
