/*
    O arquivo de fontes segue a seguinte formatao:
* byte 0  - 3:  deve ser o valor GFONT_FILE_TYPE (identifica o arquivo);
* byte 4  - 7:  nmero de fontes contidas no arquivo.
  Segue os nomes das fontes com seus respectivos indereos no arquivo:
* byte 0  - 31: nome da fonte;
* byte 32 - 35: tipo da fonte (1 -> 8x8x2 corpo e borda/sombra
                               2 -> 8x8x2 somente corpo
                               3 -> 4x8x2 somente corpo);
* byte 36 - 39: endereo da fonte no arquivo;
* byte 40 - 43: cor 1 da fonte
* byte 44 - 47: cor 2 da fonte (esse valor s tem importncia se a fonte for do tipo 1.

   seguir vem os dados da fonte, que podem ocupar 2048 bytes (tipo 1), 1024 bytes
(tipo 2) ou  512 bytes (tipo 3).

typedef struct _FONTHDR {
  char     pszName[32];      // nome da fonte
  DWORD    dwType;           // tipo da fonte (1, 2 ou 3)
  UINT     uFileAdr;         // endereo da fonte no arquivo de fontes
  COLORREF c1;               // cor 1 da fonte
  COLORREF c2;               // cor 2 da fonte
} FONTHDR;

typedef struct _FONTDATA {
  BYTE     pBits[512:1024:2048];
} FONTDATA;

typedef struct _FONTFILE {
  DWORD    dwType;           // deve ser FONT_FILE_TYPE
  DWORD    nFonts;           // nmero de fontes no arquivo
  FONTHDR  pFontHdr[nFonts]; // cabealho das fontes
  FONTDATA pData;
} FONTFILE;

*/

#include "gmain.h"
#include "state.h"
#include "resource.h"

extern "C" BYTE Font1_Shadow[1024];
extern "C" BYTE Font1_Body  [1024];
extern "C" BYTE Font2       [1024];
extern "C" BYTE Font3       [512];

// carrega as fontes do arquivo, se tudo correr bem retorna TRUE
BOOL __stdcall Load_Fonts_From_File(FILE *pFile, BOOL bLoadColor)
{
  FONTHDR fHdr;
  DWORD   i, n;
  BOOL    bClose = FALSE;

  if(!pFile)
  {
    sprintf(Str_Tmp, "%s%s", Gens_Path, GFONTS_FILE_NAME);
    pFile = fopen(Str_Tmp, "rb");
    if(!pFile) return FALSE;
    bClose = TRUE;
  }
  else if(fseek(pFile, 0, SEEK_SET)) return FALSE;

  __try
  {
    if(fread(&n, 4, 1, pFile) != 1) return FALSE;
    if(n != GFONTS_FILE_TYPE)       return FALSE; // arquivo invlido
    if(fread(&n, 4, 1, pFile) != 1) return FALSE;
    if(n > 1024)                    return FALSE; // nmero mximo de fontes
    // lendo os nomes das fontes
    for(i = 0; i < n; i++)
    {
      if(fread(&fHdr, sizeof(FONTHDR), 1, pFile) != 1) return FALSE;
      switch(fHdr.dwType)
      {
        case 1:
          if(!strcmpi(Font1_Name, fHdr.pszName))
          {
            if(fseek(pFile, fHdr.uFileAdr, SEEK_SET)) return FALSE;
            if(fread(State_Buffer, 2048, 1, pFile) != 1) return FALSE;
            if(bLoadColor) { Font1_Color1 = fHdr.c1; Font1_Color2 = fHdr.c2; }
            memcpy(Font1_Shadow, State_Buffer, 2048);
            if(fseek(pFile, 8 + ((i + 1) * sizeof(FONTHDR)), SEEK_SET)) return FALSE;
          }
          break;
        case 2:
          if(!strcmpi(Font2_Name, fHdr.pszName))
          {
            if(fseek(pFile, fHdr.uFileAdr, SEEK_SET)) return FALSE;
            if(fread(State_Buffer, 1024, 1, pFile) != 1) return FALSE;
            if(bLoadColor) { Font2_Color = fHdr.c1; }
            memcpy(Font2, State_Buffer, 1024);
            if(fseek(pFile, 8 + ((i + 1) * sizeof(FONTHDR)), SEEK_SET)) return FALSE;
          }
          break;
        case 3:
          if(!strcmpi(Font3_Name, fHdr.pszName))
          {
            if(fseek(pFile, fHdr.uFileAdr, SEEK_SET)) return FALSE;
            if(fread(State_Buffer, 512, 1, pFile) != 1) return FALSE;
            if(bLoadColor) { Font3_Color = fHdr.c1; }
            memcpy(Font3, State_Buffer , 512);
            if(fseek(pFile, 8 + ((i + 1) * sizeof(FONTHDR)), SEEK_SET)) return FALSE;
          }
          break;
        default:
          break;
      }
    }
  }
  __finally
  {
    if(bClose) fclose(pFile);
  }

  return TRUE;
}


// l o nome de todas as fontes para comboboxes
FILE *__stdcall Get_Fonts_Info(HWND hDlg)
{
  FILE   *pFile;
  DWORD   n, i;
  FONTHDR fHdr;
  int     item1, item2, item3;
  HWND    hCombo;

  sprintf(Str_Tmp, "%s%s", Gens_Path, GFONTS_FILE_NAME);
  pFile = fopen(Str_Tmp, "rbS");
  if(!pFile) return NULL;

  if(fread(&n, 4, 1, pFile) != 1) return fclose(pFile), NULL;
  if(n != GFONTS_FILE_TYPE)       return fclose(pFile), NULL; // arquivo invlido
  if(fread(&n, 4, 1, pFile) != 1) return fclose(pFile), NULL;
  if(n > 1024) return fclose(pFile), NULL; // o nmero mximo de fontes foi ultrapassado
  // lendo os nomes das fontes
  item1 = item2 = item3 = 0;
  for(i = 0; i < n; i++)
  {
    if(fread(&fHdr, sizeof(FONTHDR), 1, pFile) != 1) return fclose(pFile), NULL;
    switch(fHdr.dwType)
    {
      case 1:
        hCombo = GetDlgItem(hDlg, IDC_COMBO_FONT1);
        SendMessage(hCombo, CB_INSERTSTRING, item1, (LPARAM)fHdr.pszName);
        SendMessage(hCombo, CB_SETITEMDATA, item1, ftell(pFile) - sizeof(FONTHDR));
        if(!strcmpi(Font1_Name, fHdr.pszName))
          SendMessage(hCombo, CB_SETCURSEL, item1, 0);
        item1++;
        break;
      case 2:
        hCombo = GetDlgItem(hDlg, IDC_COMBO_FONT2);
        SendMessage(hCombo, CB_INSERTSTRING, item2, (LPARAM)fHdr.pszName);
        SendMessage(hCombo, CB_SETITEMDATA, item2, ftell(pFile) - sizeof(FONTHDR));
        if(!strcmpi(Font2_Name, fHdr.pszName))
          SendMessage(hCombo, CB_SETCURSEL, item2, 0);
        item2++;
        break;
      case 3:
        hCombo = GetDlgItem(hDlg, IDC_COMBO_FONT3);
        SendMessage(hCombo, CB_INSERTSTRING, item3, (LPARAM)fHdr.pszName);
        SendMessage(hCombo, CB_SETITEMDATA, item3, ftell(pFile) - sizeof(FONTHDR));
        if(!strcmpi(Font3_Name, fHdr.pszName))
          SendMessage(hCombo, CB_SETCURSEL, item3, 0);
        item3++;
        break;
      default:
        return fclose(pFile), NULL;
    }
  }

  return pFile;
}
