
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include "config.h"
#include "cpu_68k.h"
#include "cpu_sh2.h"
#include "sh2.h"
#include "z80.h"
#include "cd_aspi.h"
#include "gens.h"
#include "gmain.h"
#include "gddraw.h"
#include "gdsound.h"
#include "gdinput.h"
#include "vdp_io.h"
#include "vdp_rend.h"
#include "vdp_32X.h"
#include "grom.h"
#include "mem_M68K.h"
#include "mem_S68K.h"
#include "mem_SH2.h"
#include "mem_Z80.h"
#include "ym2612.h"
#include "psg.h"
#include "pcm.h"
#include "pwm.h"
#include "ggenie.h"
#include "gio.h"
#include "misc.h"
#include "cd_sys.h"
#include "zlib.h"
#include "state.h"
#include "sms.h"

BYTE State_Buffer[MAX_STATE_FILE_LENGHT];

//MK2k
int numMemstates = 0;
int memstateSize = 0;
int memstateAllocated = 0;
int maxMemTakenByMemstates = 64 * 1024 * 1024; // 64MB Memory taken by Memstates

struct Memstate *ptr_Memstates, *prev_Memstate, *next_Memstate;

char cTemp[255];

void writeLog(char *cText)
{
	FILE *fp = fopen("log.txt", "a");
	if(fp > 0)
	{
      fprintf(fp, cText);
	  fprintf(fp, "\n");
	  fclose(fp);
	}
}

void allocate_Memstates(int datasize)
{
	int maxnumstates = 100;

	maxnumstates = maxMemTakenByMemstates / datasize;

	if(memstateAllocated == 1)
      return;

	struct Memstate *created_state, *first_state;
	int i;
	memstateSize = GENESIS_STATE_FILE_LENGHT_EX; // 152912byte
	numMemstates = maxnumstates;
	if(numMemstates <= 0)
	{
		memstateAllocated = 0;
		return;
	}
	else
	{
		memstateAllocated = 1;
	}

	// reserves first State
	created_state = (struct Memstate *) malloc(sizeof(Memstate));
	created_state->hasData = false;
	created_state->buf = (BYTE *) malloc(datasize);
	first_state = created_state;
	prev_Memstate = first_state;

	for(i=1; i < numMemstates; i++)
	{
		created_state = (struct Memstate *) malloc(sizeof(Memstate));
		created_state->hasData = false;
		created_state->buf = (BYTE *) malloc(datasize);
		created_state->prev = prev_Memstate;
		prev_Memstate->next = created_state;
		prev_Memstate = created_state;
	}

	// make list circular
	created_state->next = first_state;
	first_state->prev = created_state;
	ptr_Memstates = first_state;
}

void free_Memstates()
{
	if(memstateAllocated == 0)
		return;

	struct Memstate *now_state;

	now_state = ptr_Memstates;
	prev_Memstate = now_state->prev;
	prev_Memstate->next = NULL;

	while(true)
	{
		if(now_state == NULL)
			break;
		next_Memstate = now_state->next;
		free(now_state->buf);
		free(now_state);
		now_state = next_Memstate;
	}

	memstateAllocated = 0;
}

void save_Memstate()
{
	if(memstateAllocated == 0)
		return;
	// hier ptr_Memstates->buf saven!
	Save_Memstate(ptr_Memstates->buf);
	ptr_Memstates->hasData = true;
	ptr_Memstates = ptr_Memstates->next;
}

int load_Memstate()
{
	if(memstateAllocated == 0)
		return 0;

	int ret_val = 0;
	prev_Memstate = ptr_Memstates->prev;

	if(prev_Memstate->hasData)
	{
		// hier aus prev_Memstate->buf lesen
		Load_Memstate(prev_Memstate->buf);
		prev_Memstate->hasData = false;
		ptr_Memstates = ptr_Memstates->prev;
		ret_val = 1;
	}

	return ret_val;
}

int Change_File_S(LPSTR Dest, LPSTR Dir, LPSTR Titre, LPSTR Filter, LPSTR Ext)
{
  OPENFILENAME ofn;

  SetCurrentDirectory(Gens_Path);

  if(!strcmp(Dest, ""))
  {
    strcpy(Dest, "default.");
    strcat(Dest, Ext);
  }

  memset(&ofn, 0, sizeof(OPENFILENAME));
  ofn.lStructSize     = sizeof(OPENFILENAME);
  ofn.hwndOwner       = ghWnd;
  ofn.hInstance       = ghInstance;
  ofn.lpstrFile       = Dest;
  ofn.nMaxFile        = GMAX_PATH;
  ofn.lpstrFilter     = Filter;
  ofn.nFilterIndex    = 1;
  ofn.lpstrInitialDir = Dir;
  ofn.lpstrTitle      = Titre;
  ofn.lpstrDefExt     = Ext;
  ofn.Flags           = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;

  if(GetSaveFileName(&ofn)) return 1;

  return 0;
}


int Change_File_L(LPSTR Dest, LPSTR Dir, LPSTR Titre, LPSTR Filter, LPSTR Ext)
{
  OPENFILENAME ofn;

  SetCurrentDirectory(Gens_Path);

  if(!strcmp(Dest, ""))
  {
    strcpy(Dest, "default.");
    strcat(Dest, Ext);
  }

  memset(&ofn, 0, sizeof(OPENFILENAME));
  ofn.lStructSize     = sizeof(OPENFILENAME);
  ofn.hwndOwner       = ghWnd;
  ofn.hInstance       = ghInstance;
  ofn.lpstrFile       = Dest;
  ofn.nMaxFile        = GMAX_PATH;
  ofn.lpstrFilter     = Filter;
  ofn.nFilterIndex    = 1;
  ofn.lpstrInitialDir = Dir;
  ofn.lpstrTitle      = Titre;
  ofn.lpstrDefExt     = Ext;
  ofn.Flags           = OFN_HIDEREADONLY;

  if(GetOpenFileName(&ofn)) return 1;

  return 0;
}


HANDLE Get_State_File()
{
  char  name[1024];
  Get_State_File_Name(name);
  return CreateFile(name, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
}


void Get_State_File_Name(LPSTR name)
{
  char gen_ext[4] = "gsX";
  char sms_ext[4] = "msX";
  LPSTR ext;

  ext = SMS_Started ? sms_ext : gen_ext;
  SetCurrentDirectory(Gens_Path);
  ext[2] = '0' + Current_State;
  sprintf(name, "%s%s.%s", State_Dir, Rom_Name, ext);
}

int Load_State(LPSTR Name)
{
  GZFILE f;
  LPBYTE buf;
  int len;

/*
  // Sega CD save state not suported yet
  if(SegaCD_Started)
  { Put_Info("SEGA CD STATE NOT SUPORTED", 2000); return 1; }
*/

  if     (SMS_Started)     len = SMS_STATE_FILE_LENGHT;
  else if(Genesis_Started) len = GENESIS_STATE_FILE_LENGHT_EX;
  else if(SegaCD_Started)  len = SEGACD_STATE_FILE_LENGHT;
  else if(_32X_Started)    len = G32X_STATE_FILE_LENGHT;
  else return 0;

  buf = State_Buffer;

  if((f = gzopen(Name, "rb")) == NULL) return 0;

  memset(buf, 0, len);
  if(gzread(f, buf, len) > 0)
  {
    if(SMS_Started)
    {
      Import_SMS(buf);
    }
    else
    {
      z80_Reset(&M_Z80);
//    main68k_reset();
//    YM2612ResetChip(0);
//    Reset_VDP();

      Import_Genesis(buf);
      buf += 0x22500;
      if(SegaCD_Started)
      {
        Import_SegaCD(buf);
        buf += 0xE0000;
      }
      if(_32X_Started)
      {
        Import_32X(buf);
        buf += 0x82A00;
      }

      CRam_Flag = 1;
      VRam_Flag = 1;
    }

    Flag_Clr_Scr = 1;
    Erase_MiniShot();
    sprintf(Str_Tmp, "STATE %d LOADED", Current_State);
    Put_Info(Str_Tmp, 2000);
  }

  gzclose(f);

  return 1;
}

int Load_Memstate(BYTE *memBuf)
{
	BYTE *buf;
	int len;

  if     (SMS_Started)     len = SMS_STATE_FILE_LENGHT;
  else if(Genesis_Started) len = GENESIS_STATE_FILE_LENGHT_EX;
  else if(SegaCD_Started)  len = SEGACD_STATE_FILE_LENGHT;
  else if(_32X_Started)    len = G32X_STATE_FILE_LENGHT;
  else return 0;

	buf = memBuf;

	if(SMS_Started)
	{
		Import_SMS(memBuf);
	}
	else
	{
		z80_Reset(&M_Z80);
//    main68k_reset();
//    YM2612ResetChip(0);
//    Reset_VDP();

		Import_Genesis(memBuf);

		buf += 0x22500;
		if(SegaCD_Started)
		{
			Import_SegaCD(buf);
			buf += 0xE0000;
		}
		if(_32X_Started)
		{
			Import_32X(buf);
			buf += 0x82A00;
		}

		CRam_Flag = 1;
		VRam_Flag = 1;
	}

  return 1;
}

int Save_State(LPSTR Name)
{
  HANDLE  f;
  GZFILE  gz;
  LPBYTE  buf;
  int     len;
  DWORD   d;

/*
  // Sega CD save state not suported yet
  if(SegaCD_Started)
  { Put_Info("SEGA CD STATE NOT SUPORTED", 2000); return 1; }
*/

  buf = State_Buffer;

  if(Compress_Save_States)
  { if((gz = gzopen(Name, "wb5")) == NULL) return 0; }
  else
  {
    f = CreateFile(Name, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);
    if(f == INVALID_HANDLE_VALUE) return 0;
  }

  if     (SMS_Started)     len = SMS_STATE_FILE_LENGHT;
  else if(Genesis_Started) len = GENESIS_STATE_FILE_LENGHT;
  else if(SegaCD_Started)  len = SEGACD_STATE_FILE_LENGHT;
  else if(_32X_Started)    len = G32X_STATE_FILE_LENGHT;

  memset(buf, 0, len);

  if(SMS_Started)
  {
    Export_SMS(buf);
  }
  else
  {
    Export_Genesis(buf);
    buf += 0x22500;
    if(SegaCD_Started)
    {
      Export_SegaCD(buf);
      buf += 0xE0000;
    }
    if(_32X_Started)
    {
      Export_32X(buf);
      buf += 0x82A00;
    }
  }

  if(Compress_Save_States)
  {
    gzwrite(gz, State_Buffer, len);
    gzclose(gz);
    if(Use_State_MiniShot)
    {
      f = CreateFile(Name, GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
      if(f != INVALID_HANDLE_VALUE)
      { SetFilePointer(f, 0, NULL, FILE_END);
        Create_MiniShot();
        WriteFile(f, MiniShot_Data, sizeof(MiniShot_Data), &d, NULL);
        CloseHandle(f); }
    }
  }
  else
  {
    WriteFile(f, State_Buffer, len, &d, NULL);
    if(Use_State_MiniShot)
    { Create_MiniShot();
      WriteFile(f, MiniShot_Data, sizeof(MiniShot_Data), &d, NULL); }
    CloseHandle(f);
  }

  sprintf(Str_Tmp, "STATE %d SAVED", Current_State);
  Put_Info(Str_Tmp, 3000);
  Erase_MiniShot();

  return 1;
}

int Save_Memstate(BYTE *memBuf)
{
	BYTE  *buf;
  int     len;

  buf = memBuf;

  /*
  if(Compress_Save_States)
  { if((gz = gzopen(Name, "wb5")) == NULL) return 0; }
  else
  {
    f = CreateFile(Name, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL);
    if(f == INVALID_HANDLE_VALUE) return 0;
  }
  */

  if     (SMS_Started)     len = SMS_STATE_FILE_LENGHT;
  else if(Genesis_Started) len = GENESIS_STATE_FILE_LENGHT;
  else if(SegaCD_Started)  len = SEGACD_STATE_FILE_LENGHT;
  else if(_32X_Started)    len = G32X_STATE_FILE_LENGHT;

  memset(memBuf, 0, len);

  if(SMS_Started)
  {
    Export_SMS(memBuf);
  }
  else
  {
    Export_Genesis(memBuf);
		buf += 0x22500;
    if(SegaCD_Started)
    {
      Export_SegaCD(buf);
      //buf += 0xE0000;
    }
    if(_32X_Started)
    {
      Export_32X(buf);
      //buf += 0x82A00;
    }
	}

  /*
  if(Compress_Save_States)
  {
    gzwrite(gz, State_Buffer, len);
    gzclose(gz);
    if(Use_State_MiniShot)
    {
      f = CreateFile(Name, GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
      if(f != INVALID_HANDLE_VALUE)
      { SetFilePointer(f, 0, NULL, FILE_END);
        Create_MiniShot();
        WriteFile(f, MiniShot_Data, sizeof(MiniShot_Data), &d, NULL);
        CloseHandle(f); }
    }
  }
  else
  {
    WriteFile(f, State_Buffer, len, &d, NULL);
    if(Use_State_MiniShot)
    { Create_MiniShot();
      WriteFile(f, MiniShot_Data, sizeof(MiniShot_Data), &d, NULL); }
    CloseHandle(f);
  }

  sprintf(Str_Tmp, "STATE %d SAVED", Current_State);
  Put_Info(Str_Tmp, 3000);
  Erase_MiniShot();
  */

  return 1;
}


/*

GST genecyst save file

Range        Size   Description
-----------  -----  -----------
00000-00002  3      "GST"
00006-00007  2      "\xE0\x40"
000FA-00112  24     VDP registers
00112-00191  128    Color RAM
00192-001E1  80     Vertical scroll RAM
001E4-003E3  512    YM2612 registers
00474-02473  8192   Z80 RAM
02478-12477  65536  68K RAM
12478-22477  65536  Video RAM

main 68000 registers
--------------------

00080-0009F : D0-D7
000A0-000BF : A0-A7
000C8 : PC
000D0 : SR
000D2 : USP
000D6 : SSP

Z80 registers
-------------

00404 : AF
00408 : BC
0040C : DE
00410 : HL
00414 : IX
00418 : IY
0041C : PC
00420 : SP
00424 : AF'
00428 : BC'
0042C : DE'
00430 : HL'

00434 : I
00435 : Unknow
00436 : IFF1 = IFF2
00437 : Unknow

The 'R' register is not supported.

Z80 State
---------

00438 : Z80 RESET
00439 : Z80 BUSREQ
0043A : Unknow
0043B : Unknow

0043C : Z80 BANK (DWORD)

Gens and Kega ADD
-----------------

00040 : last VDP Control data written (DWORD)
00044 : second write flag (1 for second write)
00045 : DMA Fill flag (1 mean next data write will cause a DMA fill)
00048 : VDP write address (DWORD)

00050 : Version       (Genecyst=0 ; Kega=5 ; Gens=5)
00051 : Emulator ID   (Genecyst=0 ; Kega=0 ; Gens=1)
00052 : System ID     (Genesis=0 ; SegaCD=1 ; 32X=2 ; SegaCD32X=3)

00060-00070 : PSG registers (WORD).


SEGA CD
-------

+00000-00FFF : Gate array & sub 68K
+01000-80FFF : Prg RAM
+81000-C0FFF : Word RAM (2M mode arrangement)
+C1000-D0FFF : PCM RAM
+D1000-DFFFF : CDD & CDC data (16 kB cache include)

32X
---

main SH2
--------

+00000-00FFF : cache
+01000-011FF : IO registers
+01200-0123F : R0-R15
+01240 : SR
+01244 : GBR
+01248 : VBR
+0124C : MACL
+01250 : MACH
+01254 : PR
+01258 : PC
+0125C : State

sub SH2
-------

+01400-023FF : cache
+02400-025FF : IO registers
+02600-0263F : R0-R15
+02640 : SR
+02644 : GBR
+02648 : VBR
+0264C : MACL
+02650 : MACH
+02654 : PR
+02658 : PC
+0265C : State

others
------
// Fix 32X save state :
// enregistrer correctement les registres systmes ...

+02700 : ADEN bit (bit 0)
+02701 : FM bit (bit 7)
+02702 : Master SH2 INT mask register
+02703 : Slave SH2 INT mask register
+02704 : 68000 32X rom bank register
+02705 : RV (Rom to VRAM DMA allowed) bit (bit 0)
+02710-0273F : FIFO stuff (not yet done)
+02740-0274F : 32X communication buffer
+02750-02759 : PWM registers
+02760-0276F : 32X VDP registers
+02800-029FF : 32X palette
+02A00-429FF : SDRAM
+42A00-829FF : FB1 & FB2

*/

void Import_Genesis(LPBYTE Data)
{
  BYTE Reg_1[0x200], Version, *src;
  int i;

  VDP_Int = 0;
  DMAT_Lenght = 0;
  Version = Data[0x50];

  for(i = 0; i < 0x80; i++) CRam[i] = Data[i + 0x112];
  for(i = 0; i < 0x50; i++) VSRam[i] = Data[i + 0x192];
  for(i = 0; i < 0x2000; i++) Ram_Z80[i] = Data[i + 0x474];

  for(i = 0; i < 0x10000; i += 2)
  {
    Ram_68k[i + 0] = Data[i + 0x2478 + 1];
    Ram_68k[i + 1] = Data[i + 0x2478 + 0];
  }

  for(i = 0; i < 0x10000; i += 2)
  {
    VRam[i + 0] = Data[i + 0x12478 + 1];
    VRam[i + 1] = Data[i + 0x12478 + 0];
  }

  for(i = 0; i < 0x200; i++) Reg_1[i] = Data[i + 0x1E4];
  YM2612_Restore(Reg_1);

  if((Version >= 2) && (Version < 4))
  {
    src = (LPBYTE) &Ctrl;
    for(i = 0; i < 7 * 4; i++) *src++ = Data[i + 0x30];

    Z80_State &= ~6;
    if(Data[0x440] & 1) Z80_State |= 2;
    if(Data[0x444] & 1) Z80_State |= 4;

    src = (LPBYTE) &Bank_Z80;
    for(i = 0; i < 4; i++) *src++ = Data[i + 0x448];

    src = (LPBYTE) &PSG_Save;
    for(i = 0; i < 8 * 4; i++) *src++ = Data[i + 0x224B8];
    PSG_Restore_State();
  }
  else if((Version >= 4) || (Version == 0))     // New version compatible with Kega.
  {
    Z80_State &= ~6;

    if(Version == 4)
    {
      M_Z80.IM = Data[0x437];
      M_Z80.IFF.b.IFF1 = (Data[0x438] & 1) << 2;
      M_Z80.IFF.b.IFF2 = (Data[0x438] & 1) << 2;

      Z80_State |= (Data[0x439] & 1) << 1;
    }
    else
    {
      M_Z80.IM = 1;
      M_Z80.IFF.b.IFF1 = (Data[0x436] & 1) << 2;
      M_Z80.IFF.b.IFF2 = (Data[0x436] & 1) << 2;

      Z80_State |= ((Data[0x439] & 1) ^ 1) << 1;
      Z80_State |= ((Data[0x438] & 1) ^ 1) << 2;
    }

    src = (LPBYTE) &Ctrl;
    for(i = 0; i < 7 * 4; i++) *src++ = 0;

    Write_VDP_Ctrl(Data[0x40] + (Data[0x41] << 8));
    Write_VDP_Ctrl(Data[0x42] + (Data[0x43] << 8));

    Ctrl.Flag = Data[0x44];
    Ctrl.DMA = (Data[0x45] & 1) << 2;
    Ctrl.Address = Data[0x48] + (Data[0x49] << 8);

    src = (LPBYTE) &Bank_Z80;
    for(i = 0; i < 4; i++) *src++ = Data[i + 0x43C];

    if(Version >= 4)
    {
      for(i = 0; i < 8; i++) PSG_Save[i] = Data[i * 2 + 0x60] + (Data[i * 2 + 0x61] << 8);
      PSG_Restore_State();
    }
  }

  z80_Set_AF(&M_Z80, Data[0x404] + (Data[0x405] << 8));
  M_Z80.BC.w.BC = Data[0x408] + (Data[0x409] << 8);
  M_Z80.DE.w.DE = Data[0x40C] + (Data[0x40D] << 8);
  M_Z80.HL.w.HL = Data[0x410] + (Data[0x411] << 8);
  M_Z80.IX.w.IX = Data[0x414] + (Data[0x415] << 8);
  M_Z80.IY.w.IY = Data[0x418] + (Data[0x419] << 8);
  z80_Set_PC(&M_Z80, Data[0x41C] + (Data[0x41D] << 8));
  M_Z80.SP.w.SP = Data[0x420] + (Data[0x421] << 8);
  z80_Set_AF2(&M_Z80, Data[0x424] + (Data[0x425] << 8));
  M_Z80.BC2.w.BC2 = Data[0x428] + (Data[0x429] << 8);
  M_Z80.DE2.w.DE2 = Data[0x42C] + (Data[0x42D] << 8);
  M_Z80.HL2.w.HL2 = Data[0x430] + (Data[0x431] << 8);
  M_Z80.I = Data[0x434] & 0xFF;

  main68k_GetContext(&Context_68K);

  for(i = 0; i < 24; i++) Set_VDP_Reg(i, Data[0xFA + i]);

  src = (LPBYTE) &Context_68K.dreg[0];
  for(i = 0; i < 8 * 2 * 4; i++) *src++ = Data[0x80 + i];

  src = (LPBYTE) &Context_68K.pc;
  for(i = 0; i < 4; i++) *src++ = Data[0xC8 + i];

  src = (LPBYTE) &Context_68K.sr;
  for(i = 0; i < 2; i++) *src++ = Data[0xD0 + i];

  if((Version >= 3) || (Version == 0))
  {
    if(Data[0xD1] & 0x20)
    {
      // Supervisor
      src = (LPBYTE) &Context_68K.asp;
      for(i = 0; i < 2; i++) *src++ = Data[0xD2 + i];
    }
    else
    {
      // User
      src = (LPBYTE) &Context_68K.asp;
      for(i = 0; i < 2; i++) *src++ = Data[0xD6 + i];
    }
  }

  main68k_SetContext(&Context_68K);
}


void Export_Genesis(LPBYTE Data)
{
  BYTE Reg_1[0x200], *src;
  int i;

  while (DMAT_Lenght) Update_DMA();    // Be sure to finish DMA before save

  Data[0x00] = 'G';
  Data[0x01] = 'S';
  Data[0x02] = 'T';
  Data[0x03] = 0x40;
  Data[0x04] = 0xE0;

  Data[0x50] = 5;    // Version
  Data[0x51] = 0;    // Gens

  PSG_Save_State();

  for(i = 0; i < 8; i++)
  {
    Data[0x60 + i * 2] = PSG_Save[i] & 0xFF;
    Data[0x61 + i * 2] = (PSG_Save[i] >> 8) & 0xFF;
  }

  main68k_GetContext(&Context_68K);

  src = (LPBYTE) &(Context_68K.dreg[0]);
  for(i = 0; i < 8 * 2 * 4; i++) Data[0x80 + i] = *src++;

  src = (LPBYTE) &(Context_68K.pc);
  for(i = 0; i < 4; i++) Data[0xC8 + i] = *src++;

  src = (LPBYTE) &(Context_68K.sr);
  for(i = 0; i < 2; i++) Data[0xD0 + i] = *src++;

  if(Context_68K.sr & 0x2000)
  {
    src = (LPBYTE) &(Context_68K.asp);
    for(i = 0; i < 4; i++) Data[0xD2 + i] = *src++;
    src = (LPBYTE) &(Context_68K.areg[7]);
    for(i = 0; i < 4; i++) Data[0xD6 + i] = *src++;
  }
  else
  {
    src = (LPBYTE) &(Context_68K.asp);
    for(i = 0; i < 4; i++) Data[0xD6 + i] = *src++;
    src = (LPBYTE) &(Context_68K.areg[7]);
    for(i = 0; i < 4; i++) Data[0xD2 + i] = *src++;
  }

  src = (LPBYTE) &(Ctrl.Data);
  for(i = 0; i < 4; i++) Data[0x40 + i] = *src++;

  Data[0x44] = Ctrl.Flag;
  Data[0x45] = (Ctrl.DMA >> 2) & 1;

  Data[0x48] = Ctrl.Address & 0xFF;
  Data[0x49] = (Ctrl.Address >> 8) & 0xFF;

  VDP_Reg.DMA_Lenght_L = VDP_Reg.DMA_Lenght & 0xFF;
  VDP_Reg.DMA_Lenght_H = (VDP_Reg.DMA_Lenght >> 8) & 0xFF;

  VDP_Reg.DMA_Src_Adr_L = VDP_Reg.DMA_Address & 0xFF;
  VDP_Reg.DMA_Src_Adr_M = (VDP_Reg.DMA_Address >> 8) & 0xFF;
  VDP_Reg.DMA_Src_Adr_H = (VDP_Reg.DMA_Address >> 16) & 0xFF;

  VDP_Reg.DMA_Src_Adr_H |= Ctrl.DMA_Mode & 0xC0;

  src = (LPBYTE) &(VDP_Reg.Set1);
  for(i = 0; i < 24; i++)
  {
    Data[0xFA + i] = *src;
    src += 4;
  }

  for(i = 0; i < 0x80; i++) Data[i + 0x112] = CRam[i];
  for(i = 0; i < 0x50; i++) Data[i + 0x192] = VSRam[i];

  YM2612_Save(Reg_1);
  for(i = 0; i < 0x200; i++) Data[i + 0x1E4] = Reg_1[i];

  Data[0x404] = (BYTE) (z80_Get_AF(&M_Z80) & 0xFF);
  Data[0x405] = (BYTE) (z80_Get_AF(&M_Z80) >> 8);
  Data[0x408] = (BYTE) (M_Z80.BC.w.BC & 0xFF);
  Data[0x409] = (BYTE) (M_Z80.BC.w.BC >> 8);
  Data[0x40C] = (BYTE) (M_Z80.DE.w.DE & 0xFF);
  Data[0x40D] = (BYTE) (M_Z80.DE.w.DE >> 8);
  Data[0x410] = (BYTE) (M_Z80.HL.w.HL & 0xFF);
  Data[0x411] = (BYTE) (M_Z80.HL.w.HL >> 8);
  Data[0x414] = (BYTE) (M_Z80.IX.w.IX & 0xFF);
  Data[0x415] = (BYTE) (M_Z80.IX.w.IX >> 8);
  Data[0x418] = (BYTE) (M_Z80.IY.w.IY & 0xFF);
  Data[0x419] = (BYTE) (M_Z80.IY.w.IY >> 8);
  Data[0x41C] = (BYTE) (z80_Get_PC(&M_Z80) & 0xFF);
  Data[0x41D] = (BYTE) (z80_Get_PC(&M_Z80) >> 8);
  Data[0x420] = (BYTE) (M_Z80.SP.w.SP & 0xFF);
  Data[0x421] = (BYTE) (M_Z80.SP.w.SP >> 8);
  Data[0x424] = (BYTE) (z80_Get_AF2(&M_Z80) & 0xFF);
  Data[0x425] = (BYTE) (z80_Get_AF2(&M_Z80) >> 8);
  Data[0x428] = (BYTE) (M_Z80.BC2.w.BC2 & 0xFF);
  Data[0x429] = (BYTE) (M_Z80.BC2.w.BC2 >> 8);
  Data[0x42C] = (BYTE) (M_Z80.DE2.w.DE2 & 0xFF);
  Data[0x42D] = (BYTE) (M_Z80.DE2.w.DE2 >> 8);
  Data[0x430] = (BYTE) (M_Z80.HL2.w.HL2 & 0xFF);
  Data[0x431] = (BYTE) (M_Z80.HL2.w.HL2 >> 8);
  Data[0x434] = (BYTE) (M_Z80.I);
  Data[0x436] = (BYTE) (M_Z80.IFF.b.IFF1 >> 2);

  Data[0x438] = (BYTE) (((Z80_State & 4) >> 2) ^ 1);
  Data[0x439] = (BYTE) (((Z80_State & 2) >> 1) ^ 1);

  src = (LPBYTE) &Bank_Z80;
  for(i = 0; i < 4; i++) Data[0x43C + i] = *src++;

  for(i = 0; i < 0x2000; i++) Data[i + 0x474] = Ram_Z80[i];

  for(i = 0; i < 0x10000; i += 2)
  {
    Data[i + 0x2478 + 1] = Ram_68k[i + 0];
    Data[i + 0x2478 + 0] = Ram_68k[i + 1];
  }

  for(i = 0; i < 0x10000; i += 2)
  {
    Data[i + 0x12478 + 1] = VRam[i + 0];
    Data[i + 0x12478 + 0] = VRam[i + 1];
  }
}


void Import_32X(LPBYTE Data)
{
  LPBYTE src;
  int i;

  for(i = 0; i < 0x1000; i++) M_SH2.Cache[i] = Data[i];
  for(i = 0; i < 0x200; i++) SH2_Write_Byte(&M_SH2, 0xFFFFFE00 + i, Data[0x1000 + i]);

  src = (LPBYTE) &(M_SH2.R[0]);
  for(i = 0; i < 16 * 4; i++) *src++ = Data[0x1200 + i];

  SH2_Set_SR(&M_SH2, (Data[0x1243] << 24) + (Data[0x1242] << 16) + (Data[0x1241] << 8) + Data[0x1240]);
  SH2_Set_GBR(&M_SH2, (Data[0x1247] << 24) + (Data[0x1246] << 16) + (Data[0x1245] << 8) + Data[0x1244]);
  SH2_Set_VBR(&M_SH2, (Data[0x124B] << 24) + (Data[0x124A] << 16) + (Data[0x1249] << 8) + Data[0x1248]);
  SH2_Set_MACL(&M_SH2, (Data[0x124F] << 24) + (Data[0x124E] << 16) + (Data[0x124D] << 8) + Data[0x124C]);
  SH2_Set_MACH(&M_SH2, (Data[0x1253] << 24) + (Data[0x1252] << 16) + (Data[0x1251] << 8) + Data[0x1250]);
  SH2_Set_PR(&M_SH2, (Data[0x1257] << 24) + (Data[0x1256] << 16) + (Data[0x1255] << 8) + Data[0x1254]);
  SH2_Set_PC(&M_SH2, (Data[0x125B] << 24) + (Data[0x125A] << 16) + (Data[0x1259] << 8) + Data[0x1258]);
  M_SH2.Status = (Data[0x125F] << 24) + (Data[0x125E] << 16) + (Data[0x125D] << 8) + Data[0x125C];

  for(i = 0; i < 0x1000; i++) S_SH2.Cache[i] = Data[i + 0x1400];
  for(i = 0; i < 0x200; i++) SH2_Write_Byte(&S_SH2, 0xFFFFFE00 + i, Data[0x2400 + i]);

  src = (LPBYTE) &(S_SH2.R[0]);
  for(i = 0; i < 16 * 4; i++) *src++ = Data[0x2600 + i];

  SH2_Set_SR(&S_SH2, (Data[0x2643] << 24) + (Data[0x2642] << 16) + (Data[0x2641] << 8) + Data[0x2640]);
  SH2_Set_GBR(&S_SH2, (Data[0x2647] << 24) + (Data[0x2646] << 16) + (Data[0x2645] << 8) + Data[0x2644]);
  SH2_Set_VBR(&S_SH2, (Data[0x264B] << 24) + (Data[0x264A] << 16) + (Data[0x2649] << 8) + Data[0x2648]);
  SH2_Set_MACL(&S_SH2, (Data[0x264F] << 24) + (Data[0x264E] << 16) + (Data[0x264D] << 8) + Data[0x264C]);
  SH2_Set_MACH(&S_SH2, (Data[0x2653] << 24) + (Data[0x2652] << 16) + (Data[0x2651] << 8) + Data[0x2650]);
  SH2_Set_PR(&S_SH2, (Data[0x2657] << 24) + (Data[0x2656] << 16) + (Data[0x2655] << 8) + Data[0x2654]);
  SH2_Set_PC(&S_SH2, (Data[0x265B] << 24) + (Data[0x265A] << 16) + (Data[0x2659] << 8) + Data[0x2658]);
  S_SH2.Status = (Data[0x265F] << 24) + (Data[0x265E] << 16) + (Data[0x265D] << 8) + Data[0x265C];

  _32X_ADEN = Data[0x2700] & 1;
  _32X_RV = Data[0x2705] & 1;
  M68K_32X_Mode();
  _32X_FM = Data[0x2701] & 0x80;
  _32X_Set_FB();
  _32X_MINT = Data[0x2702];
  _32X_SINT = Data[0x2703];
  Bank_SH2 = Data[0x2704];
  M68K_Set_32X_Rom_Bank();

  /*******
  FIFO stuff to add here...
   *******/
  for(i = 0; i < 0x10; i++) SH2_Write_Byte(&M_SH2, 0x4020 + i, Data[0x2740 + i]);
  for(i = 0; i < 4; i++) SH2_Write_Byte(&M_SH2, 0x4030 + i, Data[0x2750 + i]);
  /*******
  Extra PWM stuff to add here...
   *******/

  // Do it to allow VDP write on 32X side
  _32X_FM = 0x80;
  _32X_Set_FB();

  for(i = 0; i < 0x10; i++) SH2_Write_Byte(&M_SH2, 0x4100 + i, Data[0x2760 + i]);
  for(i = 0; i < 0x200; i += 2) SH2_Write_Word(&M_SH2, 0x4200 + i, Data[0x2800 + i + 0] + (Data[0x2800 + i + 1] << 8));
  for(i = 0; i < 0x40000; i++) SH2_Write_Byte(&M_SH2, 0x6000000 + i, Data[0x2A00 + i]);

  _32X_FM = Data[0x2701] & 0x80;
  _32X_Set_FB();

  if(SH2_Read_Word(&M_SH2, 0x410A) & 1)
  {
    for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i] = Data[0x42A00 + i];
    for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i + 0x20000] = Data[0x62A00 + i];
  }
  else
  {
    for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i + 0x20000] = Data[0x42A00 + i];
    for(i = 0; i < 0x20000; i++) _32X_VDP_Ram[i] = Data[0x62A00 + i];
  }
}


void Export_32X(LPBYTE Data)
{
  LPBYTE src;
  int i;

  for(i = 0; i < 0x1000; i++) Data[i] = M_SH2.Cache[i];
  for(i = 0; i < 0x200; i++) Data[0x1000 + i] = M_SH2.IO_Reg[i];

  src = (LPBYTE) &(M_SH2.R[0]);
  for(i = 0; i < 16 * 4; i++) Data[0x1200 + i] = *src++;

  i = SH2_Get_SR(&M_SH2);
  Data[0x1240] = (BYTE) (i & 0xFF);
  Data[0x1241] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x1242] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x1243] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_GBR(&M_SH2);
  Data[0x1244] = (BYTE) (i & 0xFF);
  Data[0x1245] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x1246] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x1247] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_VBR(&M_SH2);
  Data[0x1248] = (BYTE) (i & 0xFF);
  Data[0x1249] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x124A] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x124B] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_MACL(&M_SH2);
  Data[0x124C] = (BYTE) (i & 0xFF);
  Data[0x124D] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x124E] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x124F] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_MACH(&M_SH2);
  Data[0x1250] = (BYTE) (i & 0xFF);
  Data[0x1251] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x1252] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x1253] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_PR(&M_SH2);
  Data[0x1254] = (BYTE) (i & 0xFF);
  Data[0x1255] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x1256] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x1257] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_PC(&M_SH2);
  Data[0x1258] = (BYTE) (i & 0xFF);
  Data[0x1259] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x125A] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x125B] = (BYTE) ((i >> 24) & 0xFF);
  i = M_SH2.Status;
  Data[0x125C] = (BYTE) (i & 0xFF);
  Data[0x125D] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x125E] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x125F] = (BYTE) ((i >> 24) & 0xFF);

  for(i = 0; i < 0x1000; i++) Data[0x1400 + i] = S_SH2.Cache[i];
  for(i = 0; i < 0x200; i++) Data[0x2400 + i] = S_SH2.IO_Reg[i];

  src = (LPBYTE) &(S_SH2.R[0]);
  for(i = 0; i < 16 * 4; i++) Data[0x2600 + i] = *src++;

  i = SH2_Get_SR(&S_SH2);
  Data[0x2640] = (BYTE) (i & 0xFF);
  Data[0x2641] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x2642] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x2643] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_GBR(&S_SH2);
  Data[0x2644] = (BYTE) (i & 0xFF);
  Data[0x2645] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x2646] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x2647] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_VBR(&S_SH2);
  Data[0x2648] = (BYTE) (i & 0xFF);
  Data[0x2649] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x264A] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x264B] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_MACL(&S_SH2);
  Data[0x264C] = (BYTE) (i & 0xFF);
  Data[0x264D] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x264E] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x264F] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_MACH(&S_SH2);
  Data[0x2650] = (BYTE) (i & 0xFF);
  Data[0x2651] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x2652] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x2653] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_PR(&S_SH2);
  Data[0x2654] = (BYTE) (i & 0xFF);
  Data[0x2655] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x2656] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x2657] = (BYTE) ((i >> 24) & 0xFF);
  i = SH2_Get_PC(&S_SH2);
  Data[0x2658] = (BYTE) (i & 0xFF);
  Data[0x2659] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x265A] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x265B] = (BYTE) ((i >> 24) & 0xFF);
  i = S_SH2.Status;
  Data[0x265C] = (BYTE) (i & 0xFF);
  Data[0x265D] = (BYTE) ((i >> 8) & 0xFF);
  Data[0x265E] = (BYTE) ((i >> 16) & 0xFF);
  Data[0x265F] = (BYTE) ((i >> 24) & 0xFF);

  // Do it to allow VDP write on 32X side
  Data[0x2700] = _32X_ADEN & 1;
  Data[0x2705] = _32X_RV & 1;
  Data[0x2701] = _32X_FM & 0x80;
  Data[0x2702] = _32X_MINT;
  Data[0x2703] = _32X_SINT;
  Data[0x2704] = Bank_SH2;

  /*******
  FIFO stuff to add here...
   *******/
  for(i = 0; i < 0x10; i++) Data[0x2740 + i] = SH2_Read_Byte(&M_SH2, 0x4020 + i);
  for(i = 0; i < 4; i++) Data[0x2750 + i] = SH2_Read_Byte(&M_SH2, 0x4030 + i);
  /*******
  Extra PWM stuff to add here...
   *******/

  // Do it to allow VDP write on 32X side
  _32X_FM = 0x80;
  _32X_Set_FB();

  for(i = 0; i < 0x10; i++) Data[0x2760 + i] = SH2_Read_Byte(&M_SH2, 0x4100 + i);
  for(i = 0; i < 0x200; i += 2)
  {
    Data[0x2800 + i + 0] = (BYTE) (SH2_Read_Word(&M_SH2, 0x4200 + i) & 0xFF);
    Data[0x2800 + i + 1] = (BYTE) (SH2_Read_Word(&M_SH2, 0x4200 + i) >> 8);
  }
  for(i = 0; i < 0x40000; i++) Data[0x2A00 + i] = SH2_Read_Byte(&M_SH2, 0x6000000 + i);

  _32X_FM = Data[0x2701] & 0x80;
  _32X_Set_FB();

  if(SH2_Read_Word(&M_SH2, 0x410A) & 1)
  {
    for(i = 0; i < 0x20000; i++) Data[0x42A00 + i] = _32X_VDP_Ram[i];
    for(i = 0; i < 0x20000; i++) Data[0x62A00 + i] = _32X_VDP_Ram[i + 0x20000];
  }
  else
  {
    for(i = 0; i < 0x20000; i++) Data[0x42A00 + i] = _32X_VDP_Ram[i + 0x20000];
    for(i = 0; i < 0x20000; i++) Data[0x62A00 + i] = _32X_VDP_Ram[i];
  }
}


void Import_SegaCD(LPBYTE Data)
{
  S68000CONTEXT Context_sub68K;
  LPBYTE src;
  int i;

  sub68k_GetContext(&Context_sub68K);

  src = (LPBYTE) &Context_sub68K.dreg[0];
  for(i = 0; i < 8 * 2 * 4; i++) *src++ = Data[i];

  src = (LPBYTE) &Context_sub68K.pc;
  for(i = 0; i < 4; i++) *src++ = Data[0x48 + i];

  src = (LPBYTE) &Context_sub68K.sr;
  for(i = 0; i < 2; i++) *src++ = Data[0x50 + i];

  sub68k_SetContext(&Context_sub68K);

  src = (LPBYTE)Ram_Prg;
  for(i = 0; i < 0x7FFFF; i++) src[i] = Data[0x01000 + i];

  src = (LPBYTE)Ram_Word_2M;
  for(i = 0; i < 0x3FFFF; i++) src[i] = Data[0x81000 + i];

  src = (LPBYTE)Ram_PCM;
  for(i = 0; i < 0x0FFFF; i++) src[i] = Data[0xC1000 + i];
}


void Export_SegaCD(LPBYTE Data)
{
  S68000CONTEXT Context_sub68K;
  LPBYTE src;
  int i;

  sub68k_GetContext(&Context_sub68K);

  src = (LPBYTE) &(Context_sub68K.dreg[0]);
  for(i = 0; i < 8 * 2 * 4; i++) Data[i] = *src++;

  src = (LPBYTE) &(Context_sub68K.pc);
  for(i = 0; i < 4; i++) Data[0x48 + i] = *src++;

  src = (LPBYTE) &(Context_sub68K.sr);
  for(i = 0; i < 2; i++) Data[0x50 + i] = *src++;

  if(Context_sub68K.sr & 0x2000)
  {
    src = (LPBYTE) &(Context_sub68K.asp);
    for(i = 0; i < 4; i++) Data[0x52 + i] = *src++;
    src = (LPBYTE) &(Context_sub68K.areg[7]);
    for(i = 0; i < 4; i++) Data[0x56 + i] = *src++;
  }
  else
  {
    src = (LPBYTE) &(Context_sub68K.asp);
    for(i = 0; i < 4; i++) Data[0x56 + i] = *src++;
    src = (LPBYTE) &(Context_sub68K.areg[7]);
    for(i = 0; i < 4; i++) Data[0x52 + i] = *src++;
  }

  src = (LPBYTE)Ram_Prg;
  for(i = 0; i < 0x7FFFF; i++) Data[0x01000 + i] = *src++;

  src = (LPBYTE)Ram_Word_2M;
  for(i = 0; i < 0x3FFFF; i++) Data[0x81000 + i] = *src++;

  src = (LPBYTE)Ram_PCM;
  for(i = 0; i < 0x0FFFF; i++) Data[0xC1000 + i] = *src++;
}


int Load_SRAM(void)
{
  int           bResult;
  HANDLE        SRAM_File;
  char          Name[GMAX_PATH + 32];
  unsigned long Bytes_Read;

  SetCurrentDirectory(Gens_Path);

  memset(SRAM, 0, 64 * 1024);

  sprintf(Name, "%s%s%s", SRAM_Dir, Rom_Name, ".srm");

  SRAM_File = CreateFile(Name, GENERIC_READ, FILE_SHARE_READ,
    NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

  if(SRAM_File == INVALID_HANDLE_VALUE) return 0;

  bResult = ReadFile(SRAM_File, SRAM, 64 * 1024, &Bytes_Read, NULL);

  CloseHandle(SRAM_File);

  sprintf(Str_Tmp, "SRAM loaded from %s", Name);
  Put_Info(Str_Tmp, 2000);

  return 1;
}


int Save_SRAM(void)
{
  HANDLE SRAM_File;
  int bResult, size_to_save, i;
  char Name[2048];
  unsigned long Bytes_Write;

  SetCurrentDirectory(Gens_Path);

  i = (64 * 1024) - 1;
  while ((i >= 0) && (SRAM[i] == 0)) i--;

  if(i < 0) return 0;

  i++;

  size_to_save = 1;
  while (i > size_to_save) size_to_save <<= 1;

  sprintf(Name, "%s%s%s", SRAM_Dir, Rom_Name, ".srm");

  SRAM_File = CreateFile(Name, GENERIC_WRITE, NULL,
    NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

  if(SRAM_File == INVALID_HANDLE_VALUE) return 0;

  bResult = WriteFile(SRAM_File, SRAM, size_to_save, &Bytes_Write, NULL);

  CloseHandle(SRAM_File);

  sprintf(Str_Tmp, "SRAM saved in %s", Name);
  Put_Info(Str_Tmp, 2000);

  return 1;
}


void S_Format_BRAM(LPBYTE buf)
{
  memset(buf, 0x5F, 11);

  buf[0x0F] = 0x40;

  buf[0x11] = 0x7D;
  buf[0x13] = 0x7D;
  buf[0x15] = 0x7D;
  buf[0x17] = 0x7D;

  sprintf((LPSTR ) &buf[0x20], "SEGA CD ROM");
  sprintf((LPSTR ) &buf[0x30], "RAM CARTRIDGE");

  buf[0x24] = 0x5F;
  buf[0x27] = 0x5F;

  buf[0x2C] = 0x01;

  buf[0x33] = 0x5F;
  buf[0x3D] = 0x5F;
  buf[0x3E] = 0x5F;
  buf[0x3F] = 0x5F;
}


void Format_Backup_Ram(void)
{
  memset(Ram_Backup, 0, 8 * 1024);

  S_Format_BRAM(&Ram_Backup[0x1FC0]);

  memset(Ram_Backup_Ex, 0, 64 * 1024);
}


int Load_BRAM(void)
{
  HANDLE BRAM_File;
  int bResult;
  char Name[2048];
  unsigned long Bytes_Read;

  Format_Backup_Ram();

  SetCurrentDirectory(Gens_Path);

  strcpy(Name, BRAM_Dir);
  strcat(Name, Rom_Name);
  strcat(Name, ".brm");

  BRAM_File = CreateFile(Name, GENERIC_READ, FILE_SHARE_READ,
    NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

  if(BRAM_File == INVALID_HANDLE_VALUE) return 0;

  bResult = ReadFile(BRAM_File, Ram_Backup, 8 * 1024, &Bytes_Read, NULL);
  bResult = ReadFile(BRAM_File, Ram_Backup_Ex, (8 << BRAM_Ex_Size) * 1024, &Bytes_Read, NULL);

  CloseHandle(BRAM_File);

  strcpy(Str_Tmp, "BRAM loaded from ");
  strcat(Str_Tmp, Name);
  Put_Info(Str_Tmp, 2000);

  return 1;
}


int Save_BRAM(void)
{
  HANDLE BRAM_File;
  int bResult;
  char Name[2048];
  unsigned long Bytes_Write;

  SetCurrentDirectory(Gens_Path);

  strcpy(Name, BRAM_Dir);
  strcat(Name, Rom_Name);
  strcat(Name, ".brm");

  BRAM_File = CreateFile(Name, GENERIC_WRITE, NULL,
    NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

  if(BRAM_File == INVALID_HANDLE_VALUE) return 0;

  bResult = WriteFile(BRAM_File, Ram_Backup, 8 * 1024, &Bytes_Write, NULL);
  bResult = WriteFile(BRAM_File, Ram_Backup_Ex, (8 << BRAM_Ex_Size) * 1024, &Bytes_Write, NULL);

  CloseHandle(BRAM_File);

  strcpy(Str_Tmp, "BRAM saved in ");
  strcat(Str_Tmp, Name);
  Put_Info(Str_Tmp, 2000);

  return 1;
}
