/**********************************************************/
/*                                                        */
/* Z80 emulator 0.99                                      */
/* Copyright 2002 Stphane Dallongeville                  */
/* Used for the genesis emulation in Gens                 */
/*                                                        */
/**********************************************************/

#ifndef _Z80_H_
#define _Z80_H_

#include <windows.h>

#ifdef __cplusplus
extern "C" {
#endif


/******************************/
/* Compiler dependant defines */
/******************************/

#ifndef FASTCALL
#define FASTCALL __fastcall
#endif


/****************************/
/* Structures & definitions */
/****************************/


typedef BYTE FASTCALL Z80_RB(DWORD adr);
typedef WORD FASTCALL Z80_RW(DWORD adr);
typedef void FASTCALL Z80_WB(DWORD adr, BYTE data);
typedef void FASTCALL Z80_WW(DWORD adr, WORD data);


struct _Z80_context
{
  union
  {
    struct
    {
      BYTE A;
      BYTE F;
      BYTE x;
      BYTE FXY;
    } b;
    struct
    {
      WORD AF;
      WORD FXYW;
    } w;
    DWORD d;
  } AF;
  union
  {
    struct
    {
      BYTE C;
      BYTE B;
      WORD x;
    } b;
    struct
    {
      WORD BC;
      WORD x;
    } w;
    DWORD d;
  } BC;
  union
  {
    struct
    {
      BYTE E;
      BYTE D;
      WORD x;
    } b;
    struct
    {
      WORD DE;
      WORD x;
    } w;
    DWORD d;
  } DE;
  union
  {
    struct
    {
      BYTE L;
      BYTE H;
      WORD x;
    } b;
    struct
    {
      WORD HL;
      WORD x;
    } w;
    DWORD d;
  } HL;
  union
  {
    struct
    {
      BYTE IXL;
      BYTE IXH;
      WORD x;
    } b;
    struct
    {
      WORD IX;
      WORD x;
    } w;
    DWORD d;
  } IX;
  union
  {
    struct
    {
      BYTE IYL;
      BYTE IYH;
      WORD x;
    } b;
    struct
    {
      WORD IY;
      WORD x;
    } w;
    DWORD d;
  } IY;
  union
  {
    struct
    {
      BYTE PCL;
      BYTE PCH;
      WORD x;
    } b;
    struct
    {
      WORD PC;
      WORD x;
    } w;
    DWORD d;
  } PC;
  union
  {
    struct
    {
      BYTE SPL;
      BYTE SPH;
      WORD x;
    } b;
    struct
    {
      WORD SP;
      WORD x;
    } w;
    DWORD d;
  } SP;
  union
  {
    struct
    {
      BYTE A2;
      BYTE F2;
      BYTE x;
      BYTE FXY2;
    } b;
    struct
    {
      WORD AF2;
      WORD FXYW2;
    } w;
    DWORD d;
  } AF2;
  union
  {
    struct
    {
      BYTE C2;
      BYTE B2;
      WORD x;
    } b;
    struct
    {
      WORD BC2;
      WORD x;
    } w;
    DWORD d;
  } BC2;
  union
  {
    struct
    {
      BYTE E2;
      BYTE D2;
      WORD x;
    } b;
    struct
    {
      WORD DE2;
      WORD x;
    } w;
    DWORD d;
  } DE2;
  union
  {
    struct
    {
      BYTE L2;
      BYTE H2;
      WORD x;
    } b;
    struct
    {
      WORD HL2;
      WORD x;
    } w;
    DWORD d;
  } HL2;
  union
  {
    struct
    {
      BYTE IFF1;
      BYTE IFF2;
      WORD x;
    } b;
    struct
    {
      WORD IFF;
      WORD x;
    } w;
    DWORD d;
  } IFF;
  union
  {
    struct
    {
      BYTE R1;
      BYTE R2;
      WORD x;
    } b;
    struct
    {
      WORD R;
      WORD x;
    } w;
    DWORD d;
  } R;

  BYTE I;
  BYTE IM;
  BYTE IntVect;
  BYTE IntLine;

  DWORD Status;
  DWORD BasePC;
  DWORD TmpSav0;
  DWORD TmpSav1;

  DWORD CycleCnt;
  DWORD CycleTD;
  DWORD CycleIO;
  DWORD CycleSup;

  Z80_RB *ReadB[0x100];
  Z80_WB *WriteB[0x100];
  BYTE *Fetch[0x100];
  Z80_RW *ReadW[0x100];
  Z80_WW *WriteW[0x100];

  Z80_WB *IN_C;
  Z80_RB *OUT_C;
  DWORD RetIC;
  DWORD IntAckC;
};

typedef struct _Z80_context Z80_CONTEXT;


/*************************/
/* Publics Z80 variables */
/*************************/


extern Z80_CONTEXT M_Z80;


/*************************/
/* Publics Z80 functions */
/*************************/


DWORD FASTCALL z80_Init(Z80_CONTEXT *z80);
DWORD FASTCALL z80_Reset(Z80_CONTEXT *z80);

DWORD z80_Add_ReadB(Z80_CONTEXT *z80, DWORD low_adr, DWORD high_adr, Z80_RB *Func);
DWORD z80_Add_ReadW(Z80_CONTEXT *z80, DWORD low_adr, DWORD high_adr, Z80_RW *Func);
DWORD z80_Add_WriteB(Z80_CONTEXT *z80, DWORD low_adr, DWORD high_adr, Z80_WB *Func);
DWORD z80_Add_WriteW(Z80_CONTEXT *z80, DWORD low_adr, DWORD high_adr, Z80_WW *Func);
DWORD z80_Add_Fetch(Z80_CONTEXT *z80, DWORD low_adr, DWORD high_adr, BYTE *Region);

DWORD FASTCALL z80_Read_Odo(Z80_CONTEXT *z80);
void FASTCALL z80_Clear_Odo(Z80_CONTEXT *z80);
void FASTCALL z80_Set_Odo(Z80_CONTEXT *z80, DWORD Odo);
void FASTCALL z80_Add_Cycles(Z80_CONTEXT *z80, DWORD cycles);

DWORD FASTCALL z80_Exec(Z80_CONTEXT *z80, int odo);

DWORD FASTCALL z80_NMI(Z80_CONTEXT *z80);
DWORD FASTCALL z80_Interrupt(Z80_CONTEXT *z80, DWORD vector);

DWORD FASTCALL z80_Get_PC(Z80_CONTEXT *z80);
DWORD FASTCALL z80_Get_AF(Z80_CONTEXT *z80);
DWORD FASTCALL z80_Get_AF2(Z80_CONTEXT *z80);
DWORD FASTCALL z80_Set_PC(Z80_CONTEXT *z80, DWORD PC);
DWORD FASTCALL z80_Set_AF(Z80_CONTEXT *z80, DWORD AF);
DWORD FASTCALL z80_Set_AF2(Z80_CONTEXT *z80, DWORD AF2);


#ifdef __cplusplus
};
#endif

#endif