
#include <stdio.h>
#include "resource.h"
#include "gdinput.h"
#include "gio.h"
#include "gmain.h"
#include "grom.h"
#include "keycode.h"
#include "gmain.h"
#include "movie.h"

#define KEYDOWN(key) (Keys[key] & 0x80)
#define MAX_JOYS 8

LPDIRECTINPUT lpDI;
LPDIRECTINPUTDEVICE lpDIDKeyboard;
LPDIRECTINPUTDEVICE lpDIDMouse;
char Phrase[1024];
int Nb_Joys = 0;
static IDirectInputDevice2 *Joy_ID[MAX_JOYS] = {NULL};
static DIJOYSTATE Joy_State[MAX_JOYS] = {{0}};
long MouseX, MouseY;
unsigned char Keys[256];
unsigned char Kaillera_Keys[16];
LPCONTROL_KEYS Control = (LPCONTROL_KEYS)(&Controller_1_Type);


struct K_Def Keys_Def[8] =
{
	{
		DIK_RETURN,
		DIK_RSHIFT,
		DIK_Z,
		DIK_X,
		DIK_C,
		DIK_A,
		DIK_S,
		DIK_D,
		DIK_UP,
		DIK_DOWN,
		DIK_LEFT,
		DIK_RIGHT,
		DIK_V //Rewound: Added Default Rewind Key
	},
  {
		DIK_U,
		DIK_T,
		DIK_K,
		DIK_L,
		DIK_M,
		DIK_I,
		DIK_O,
		DIK_P,
		DIK_Y,
		DIK_H,
		DIK_G,
		DIK_J,
		DIK_B //Rewound: Added Default Rewind Key
	}
};


void End_Input()
{
  int i;

  if(lpDI)
  {
    if(lpDIDMouse)
    {
      lpDIDMouse->Release();
      lpDIDMouse = NULL;
    }

    if(lpDIDKeyboard)
    {
      lpDIDKeyboard->Release();
      lpDIDKeyboard = NULL;
    }

    for(i = 0; i < MAX_JOYS; i++)
    {
      if(Joy_ID[i])
      {
        Joy_ID[i]->Unacquire();
        Joy_ID[i]->Release();
      }
    }

    Nb_Joys = 0;
    lpDI->Release();
    lpDI = NULL;
  }
}


BOOL CALLBACK InitJoystick(LPCDIDEVICEINSTANCE lpDIIJoy, LPVOID pvRef)
{
  HRESULT             rval;
  LPDIRECTINPUTDEVICE lpDIJoy;
  DIPROPRANGE         diprg;
  int i;

  if(Nb_Joys >= MAX_JOYS) return DIENUM_STOP;

  Joy_ID[Nb_Joys] = NULL;

  rval = lpDI->CreateDevice(lpDIIJoy->guidInstance, &lpDIJoy, NULL);
  if(rval != DI_OK)
  {
    MessageBox(ghWnd, "IDirectInput::CreateDevice FAILED", "error joystick", MB_OK);
    return DIENUM_CONTINUE;
  }

  rval = lpDIJoy->QueryInterface(IID_IDirectInputDevice2, (void **)&Joy_ID[Nb_Joys]);
  lpDIJoy->Release();
  if(rval != DI_OK)
  {
    MessageBox(ghWnd, "IDirectInputDevice2::QueryInterface FAILED", "error joystick", MB_OK);
      Joy_ID[Nb_Joys] = NULL;
      return DIENUM_CONTINUE;
  }

  rval = Joy_ID[Nb_Joys]->SetDataFormat(&c_dfDIJoystick);
  if(rval != DI_OK)
  {
    MessageBox(ghWnd, "IDirectInputDevice::SetDataFormat FAILED", "error joystick", MB_OK);
    Joy_ID[Nb_Joys]->Release();
    Joy_ID[Nb_Joys] = NULL;
    return DIENUM_CONTINUE;
  }

  rval = Joy_ID[Nb_Joys]->SetCooperativeLevel((HWND)pvRef, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);

  if(rval != DI_OK)
  {
    MessageBox(ghWnd, "IDirectInputDevice::SetCooperativeLevel FAILED", "error joystick", MB_OK);
    Joy_ID[Nb_Joys]->Release();
    Joy_ID[Nb_Joys] = NULL;
    return DIENUM_CONTINUE;
  }

  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_X;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -1000;
  diprg.lMax              = +1000;

  rval = Joy_ID[Nb_Joys]->SetProperty(DIPROP_RANGE, &diprg.diph);
  if((rval != DI_OK) && (rval != DI_PROPNOEFFECT))
    MessageBox(ghWnd, "IDirectInputDevice::SetProperty() (X-Axis) FAILED", "error joystick", MB_OK);

  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_Y;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -1000;
  diprg.lMax              = +1000;

  rval = Joy_ID[Nb_Joys]->SetProperty(DIPROP_RANGE, &diprg.diph);
  if((rval != DI_OK) && (rval != DI_PROPNOEFFECT))
    MessageBox(ghWnd, "IDirectInputDevice::SetProperty() (Y-Axis) FAILED", "error joystick", MB_OK);

  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_Z;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -1000;
  diprg.lMax              = +1000;

  Joy_ID[Nb_Joys]->SetProperty(DIPROP_RANGE, &diprg.diph);

  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_RX;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -1000;
  diprg.lMax              = +1000;

  Joy_ID[Nb_Joys]->SetProperty(DIPROP_RANGE, &diprg.diph);

  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_RY;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -1000;
  diprg.lMax              = +1000;

  Joy_ID[Nb_Joys]->SetProperty(DIPROP_RANGE, &diprg.diph);

  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_RZ;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -1000;
  diprg.lMax              = +1000;

  Joy_ID[Nb_Joys]->SetProperty(DIPROP_RANGE, &diprg.diph);

  for(i = 0; i < 10; i++)
  {
    rval = Joy_ID[Nb_Joys]->Acquire();
    if(rval == DI_OK) break;
    Sleep(10);
  }

  Nb_Joys++;

  return DIENUM_CONTINUE;
}


int Init_Input(HINSTANCE hInst, HWND hWnd)
{
  int i;
  HRESULT rval;

  End_Input();

  rval = DirectInputCreate(hInst, DIRECTINPUT_VERSION, &lpDI, NULL);
  if(rval != DI_OK)
  {
    MessageBox(hWnd, "DirectInput failed ...You must have DirectX 5", "Error", MB_OK);
    return 0;
  }

  Nb_Joys = 0;

  for(i = 0; i < MAX_JOYS; i++) Joy_ID[i] = NULL;

  rval = lpDI->EnumDevices(DIDEVTYPE_JOYSTICK, &InitJoystick, hWnd, DIEDFL_ATTACHEDONLY);
  if(rval != DI_OK) return 0;

//  rval = lpDI->CreateDevice(GUID_SysMouse, &lpDIDMouse, NULL);
  rval = lpDI->CreateDevice(GUID_SysKeyboard, &lpDIDKeyboard, NULL);
  if(rval != DI_OK) return 0;

//  rval = lpDIDMouse->SetCooperativeLevel(hWnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);
  rval = lpDIDKeyboard->SetCooperativeLevel(hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
  if(rval != DI_OK) return 0;

//  rval = lpDIDMouse->SetDataFormat(&c_dfDIMouse);
  rval = lpDIDKeyboard->SetDataFormat(&c_dfDIKeyboard);
  if(rval != DI_OK) return 0;

//  rval = lpDIDMouse->Acquire();
  for(i = 0; i < 10; i++)
  {
    rval = lpDIDKeyboard->Acquire();
    if(rval == DI_OK) break;
    Sleep(10);
  }

  return 1;
}


void Restore_Input()
{
//  lpDIDMouse->Acquire();
  lpDIDKeyboard->Acquire();
}


void Update_Input()
{
//  DIMOUSESTATE MouseState;
  HRESULT rval;
  int i;

  rval = lpDIDKeyboard->GetDeviceState(256, &Keys);

  if((rval == DIERR_INPUTLOST) | (rval == DIERR_NOTACQUIRED))
    Restore_Input();

  for(i = 0; i < Nb_Joys; i++)
  {
    if(Joy_ID[i])
    {
      Joy_ID[i]->Poll();
      rval = Joy_ID[i]->GetDeviceState(sizeof(Joy_State[i]), &Joy_State[i]);
      if(rval != DI_OK) Joy_ID[i]->Acquire();
    }
  }

//  rval = lpDIDMouse->GetDeviceState(sizeof(MouseState), &MouseState);

//  if((rval == DIERR_INPUTLOST) | (rval == DIERR_NOTACQUIRED))
//    Restore_Input();

//  MouseX = MouseState.lX;
//  MouseY = MouseState.lY;
}


int Check_Key_Pressed(unsigned int key)
{
  int Num_Joy;

  if(key < 0x100)
  {
    if KEYDOWN(key) return(1);
  }
  else
  {
    Num_Joy = ((key >> 8) & 0xF);

    if(Joy_ID[Num_Joy])
    {
      if(key & 0x80)      // Test POV Joys
      {
        switch(key & 0x0F)
        {
          case 1:
            if( (Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] >= 0 &&
                 Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] <= 4500) ||
                (Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] >= 31500 &&
                 Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] <= 36000) )
            { return 1; }
            break;
          case 2:
            if( Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] >= 4500 &&
                Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] <= 13500 )
            { return 1; }
            break;
          case 3:
            if( Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] >= 13500 &&
                Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] <= 22500 )
            { return 1; }
            break;
          case 4:
            if( Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] >= 22500 &&
                Joy_State[Num_Joy].rgdwPOV[(key >> 4) & 3] <= 31500 )
            { return 1; }
            break;
        }

      }
      else if(key & 0x70)    // Test Button Joys
      {
        if(Joy_State[Num_Joy].rgbButtons[(key & 0xFF) - 0x10]) return 1;
      }
      else
      {
        switch(key & 0xF)
        {
          case 1:
            if(Joy_State[Num_Joy].lY < -500)
            { return 1; }
            break;
          case 2:
            if(Joy_State[Num_Joy].lY > +500)
            { return 1; }
            break;
          case 3:
            if(Joy_State[Num_Joy].lX < -500)
            { return 1; }
            break;
          case 4:
            if(Joy_State[Num_Joy].lX > +500)
            { return 1; }
            break;
          case 5:
            if(Joy_State[Num_Joy].lZ < -500)
            { return 1; }
            break;
          case 6:
            if(Joy_State[Num_Joy].lZ > +500)
            { return 1; }
            break;
          case 7:
            if(Joy_State[Num_Joy].lRx < -500)
            { return 1; }
            break;
          case 8:
            if(Joy_State[Num_Joy].lRx > +500)
            { return 1; }
            break;
          case 9:
            if(Joy_State[Num_Joy].lRy < -500)
            { return 1; }
            break;
          case 10:
            if(Joy_State[Num_Joy].lRy > +500)
            { return 1; }
            break;
          case 11:
            if(Joy_State[Num_Joy].lRz < -500)
            { return 1; }
            break;
          case 12:
            if(Joy_State[Num_Joy].lRz > +500)
            { return 1; }
            break;
        }
      }
    }
  }

  return 0;
}


unsigned int Get_Key(void)
{
  int i, j;

  while(1)
  {
    Update_Input();

    for(i = 1; i < 256; i++)
      if KEYDOWN(i) return i;

    for(i = 0; i < Nb_Joys; i++)
    {
      if(Joy_ID[i])
      {
        if(Joy_State[i].lY < -500)
          return(0x1000 + (0x100 * i) + 0x1);

        if(Joy_State[i].lY > +500)
          return(0x1000 + (0x100 * i) + 0x2);

        if(Joy_State[i].lX < -500)
          return(0x1000 + (0x100 * i) + 0x3);

        if(Joy_State[i].lX > +500)
          return(0x1000 + (0x100 * i) + 0x4);

        if(Joy_State[i].lZ < -500)
          return(0x1000 + (0x100 * i) + 0x5);

        if(Joy_State[i].lZ > +500)
          return(0x1000 + (0x100 * i) + 0x6);

        if(Joy_State[i].lRx < -500)
          return(0x1000 + (0x100 * i) + 0x7);

        if(Joy_State[i].lRx > +500)
          return(0x1000 + (0x100 * i) + 0x8);

        if(Joy_State[i].lRy < -500)
          return(0x1000 + (0x100 * i) + 0x9);

        if(Joy_State[i].lRy > +500)
          return(0x1000 + (0x100 * i) + 0xA);

        if(Joy_State[i].lRz < -500)
          return(0x1000 + (0x100 * i) + 0xB);

        if(Joy_State[i].lRz > +500)
          return(0x1000 + (0x100 * i) + 0xC);

        // POV
        for(j = 0; j < 4; j++)
        {
          if( (Joy_State[i].rgdwPOV[j] >= 0     && Joy_State[i].rgdwPOV[j] <= 4500) ||
              (Joy_State[i].rgdwPOV[j] >= 31500 && Joy_State[i].rgdwPOV[j] <= 36000) )
            return(0x1080 + (0x100 * i) + (0x10 * j) + 0x1);
        }

        for(j = 0; j < 4; j++)
        {
          if( (Joy_State[i].rgdwPOV[j] >= 4500  && Joy_State[i].rgdwPOV[j] <= 13500) )
            return(0x1080 + (0x100 * i) + (0x10 * j) + 0x2);
        }

        for(j = 0; j < 4; j++)
        {
          if( (Joy_State[i].rgdwPOV[j] >= 13500  && Joy_State[i].rgdwPOV[j] <= 22500) )
            return(0x1080 + (0x100 * i) + (0x10 * j) + 0x3);
        }

        for(j = 0; j < 4; j++)
        {
          if( (Joy_State[i].rgdwPOV[j] >= 22500  && Joy_State[i].rgdwPOV[j] <= 31500) )
            return(0x1080 + (0x100 * i) + (0x10 * j) + 0x4);
        }

        for(j = 0; j < 32; j++)
        {
          if(Joy_State[i].rgbButtons[j])
            return(0x1010 + (0x100 * i) + j);
        }
      }
    }
  }
}

int __stdcall RewindRequested(void)
{
	if(Check_Key_Pressed(Keys_Def[0].Rewind))
		return 1;
	else
		return 0;
}

void __stdcall Update_Controllers(void)
{
  int i, j;

  // movies implement
  if(Movie_Status & MOVIE_PLAYING)
  { MoviePlayingStuff(); return; }

  Update_Input();

  for(i = 0; i < 8; i++)
  {
    if(Check_Key_Pressed(Keys_Def[i].Up))
    {
      Control[i].Up = 0;
      Control[i].Down = 1;
    }
    else
    {
      Control[i].Up = 1;
      if(Check_Key_Pressed(Keys_Def[i].Down)) Control[i].Down = 0;
      else Control[i].Down = 1;
    }

    if(Check_Key_Pressed(Keys_Def[i].Left))
    {
      Control[i].Left = 0;
      Control[i].Right = 1;
    }
    else
    {
      Control[i].Left = 1;
      if(Check_Key_Pressed(Keys_Def[i].Right)) Control[i].Right = 0;
      else Control[i].Right = 1;
    }

    if(Check_Key_Pressed(Keys_Def[i].Start)) Control[i].Start = 0;
    else Control[i].Start = 1;

    if(Check_Key_Pressed(Keys_Def[i].A)) Control[i].A = 0;
    else Control[i].A = 1;

    if(Check_Key_Pressed(Keys_Def[i].B)) Control[i].B = 0;
    else Control[i].B = 1;

    if(Check_Key_Pressed(Keys_Def[i].C)) Control[i].C = 0;
    else Control[i].C = 1;

    if(Control[i].Type & 1)
    {
      if(Check_Key_Pressed(Keys_Def[i].Mode)) Control[i].Mode = 0;
      else Control[i].Mode = 1;

      if(Check_Key_Pressed(Keys_Def[i].X)) Control[i].X = 0;
      else Control[i].X = 1;

      if(Check_Key_Pressed(Keys_Def[i].Y)) Control[i].Y = 0;
      else Control[i].Y = 1;

      if(Check_Key_Pressed(Keys_Def[i].Z)) Control[i].Z = 0;
      else Control[i].Z = 1;
    }

    // hotkeys
    for(j = 0; j < 4; j++)
    {
      if(Check_Key_Pressed(Keys_Def[i].HotKey[j]))
      {
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_UP)    Control[i].Up    = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_LEFT)  Control[i].Left  = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_RIGHT) Control[i].Right = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_DOWN)  Control[i].Down  = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_A)     Control[i].A     = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_B)     Control[i].B     = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_C)     Control[i].C     = 0;
        if(Keys_Def[i].HotKey_Buttons[j] & HKB_START) Control[i].Start = 0;
        if(Control[i].Type & 1)
        {
          if(Keys_Def[i].HotKey_Buttons[j] & HKB_X)    Control[i].X    = 0;
          if(Keys_Def[i].HotKey_Buttons[j] & HKB_Y)    Control[i].Y    = 0;
          if(Keys_Def[i].HotKey_Buttons[j] & HKB_Z)    Control[i].Z    = 0;
          if(Keys_Def[i].HotKey_Buttons[j] & HKB_MODE) Control[i].Mode = 0;
        }
      }
    }
  }

  if(Movie_Status & MOVIE_RECORDING)
  { MovieRecordingStuff(); }
}


#define _GETKEY(key, mes)       \
  SetWindowText(Txt1, mes);     \
  key = Get_Key();              \
  if(key == K_ESC) { key = 0; } \
  Sleep(250);

int Setting_Keys(HWND hset, int Player, int TypeP)
{
  HWND Txt1, Txt2;
  MSG m;

  Sleep(250);
  Txt1 = GetDlgItem(hset, IDC_STATIC_TEXT1);
  Txt2 = GetDlgItem(hset, IDC_STATIC_TEXT2);
  if(Txt1 == NULL) return 0;
  if(Txt2 == NULL) return 0;

  SetWindowText(Txt2, "ESC FOR NONE...");

  _GETKEY(Keys_Def[Player].Up,   "INPUT KEY FOR UP");
  _GETKEY(Keys_Def[Player].Down, "INPUT KEY FOR DOWN");
  _GETKEY(Keys_Def[Player].Left, "INPUT KEY FOR LEFT");
  _GETKEY(Keys_Def[Player].Right, "INPUT KEY FOR RIGHT");
  _GETKEY(Keys_Def[Player].Start, "INPUT KEY FOR START");
  _GETKEY(Keys_Def[Player].A, "INPUT KEY FOR A");
  _GETKEY(Keys_Def[Player].B, "INPUT KEY FOR B");
  _GETKEY(Keys_Def[Player].C, "INPUT KEY FOR C");

	_GETKEY(Keys_Def[Player].Rewind, "INPUT KEY FOR REWiND"); //Rewound: Added Key Profile

  if(TypeP)
  {
    _GETKEY(Keys_Def[Player].Mode, "INPUT KEY FOR MODE");
    _GETKEY(Keys_Def[Player].X, "INPUT KEY FOR X");
    _GETKEY(Keys_Def[Player].Y, "INPUT KEY FOR Y");
    _GETKEY(Keys_Def[Player].Z, "INPUT KEY FOR Z");

  }

  _GETKEY(Keys_Def[Player].HotKey[0], "INPUT KEY FOR HOTKEY 1");
  _GETKEY(Keys_Def[Player].HotKey[1], "INPUT KEY FOR HOTKEY 2");
  _GETKEY(Keys_Def[Player].HotKey[2], "INPUT KEY FOR HOTKEY 3");
  _GETKEY(Keys_Def[Player].HotKey[3], "INPUT KEY FOR HOTKEY 4");

  SetWindowText(Txt1, "CONFIGURATION SUCCESSFULL");
  SetWindowText(Txt2, "PRESS A KEY TO CONTINUE ...");
  Get_Key();
  Sleep(500);

  while (PeekMessage(&m, hset, WM_KEYDOWN, WM_KEYDOWN, PM_REMOVE));

  SetWindowText(Txt1, "");
  SetWindowText(Txt2, "");

  return 1;
}


void Scan_Player_Net(int Player)
{
  if(!Player) return;

  Update_Input();

  if(Check_Key_Pressed(Keys_Def[0].Up))
  {
    Kaillera_Keys[0] &= ~0x08;
    Kaillera_Keys[0] |= 0x04;
  }
  else
  {
    Kaillera_Keys[0] |= 0x08;
    if(Check_Key_Pressed(Keys_Def[0].Down)) Kaillera_Keys[0] &= ~0x04;
    else Kaillera_Keys[0] |= 0x04;
  }

  if(Check_Key_Pressed(Keys_Def[0].Left))
  {
    Kaillera_Keys[0] &= ~0x02;
    Kaillera_Keys[0] |= 0x01;
  }
  else
  {
    Kaillera_Keys[0] |= 0x02;
    if(Check_Key_Pressed(Keys_Def[0].Right)) Kaillera_Keys[0] &= ~0x01;
    else Kaillera_Keys[0] |= 0x01;
  }

  if(Check_Key_Pressed(Keys_Def[0].Start)) Kaillera_Keys[0] &= ~0x80;
  else Kaillera_Keys[0] |= 0x80;

  if(Check_Key_Pressed(Keys_Def[0].A)) Kaillera_Keys[0] &= ~0x40;
  else Kaillera_Keys[0] |= 0x40;

  if(Check_Key_Pressed(Keys_Def[0].B)) Kaillera_Keys[0] &= ~0x20;
  else Kaillera_Keys[0] |= 0x20;

  if(Check_Key_Pressed(Keys_Def[0].C)) Kaillera_Keys[0] &= ~0x10;
  else Kaillera_Keys[0] |= 0x10;

  if(Controller_1_Type & 1)
  {
    if(Check_Key_Pressed(Keys_Def[0].Mode)) Kaillera_Keys[1] &= ~0x08;
    else Kaillera_Keys[1] |= 0x08;

    if(Check_Key_Pressed(Keys_Def[0].X)) Kaillera_Keys[1] &= ~0x04;
    else Kaillera_Keys[1] |= 0x04;

    if(Check_Key_Pressed(Keys_Def[0].Y)) Kaillera_Keys[1] &= ~0x02;
    else Kaillera_Keys[1] |= 0x02;

    if(Check_Key_Pressed(Keys_Def[0].Z)) Kaillera_Keys[1] &= ~0x01;
    else Kaillera_Keys[1] |= 0x01;
  }
}


void Update_Controllers_Net(int num_player)
{
  Controller_1_Up = (Kaillera_Keys[0] & 0x08) >> 3;
  Controller_1_Down = (Kaillera_Keys[0] & 0x04) >> 2;
  Controller_1_Left = (Kaillera_Keys[0] & 0x02) >> 1;
  Controller_1_Right = (Kaillera_Keys[0] & 0x01);
  Controller_1_Start = (Kaillera_Keys[0] & 0x80) >> 7;
  Controller_1_A = (Kaillera_Keys[0] & 0x40) >> 6;
  Controller_1_B = (Kaillera_Keys[0] & 0x20) >> 5;
  Controller_1_C = (Kaillera_Keys[0] & 0x10) >> 4;

  if(Controller_1_Type & 1)
  {
    Controller_1_Mode = (Kaillera_Keys[0 + 1] & 0x08) >> 3;
    Controller_1_X = (Kaillera_Keys[0 + 1] & 0x04) >> 2;
    Controller_1_Y = (Kaillera_Keys[0 + 1] & 0x02) >> 1;
    Controller_1_Z = (Kaillera_Keys[0 + 1] & 0x01);
  }

  if(num_player > 2)      // TEAMPLAYER
  {
    Controller_1B_Up = (Kaillera_Keys[2] & 0x08) >> 3;
    Controller_1B_Down = (Kaillera_Keys[2] & 0x04) >> 2;
    Controller_1B_Left = (Kaillera_Keys[2] & 0x02) >> 1;
    Controller_1B_Right = (Kaillera_Keys[2] & 0x01);
    Controller_1B_Start = (Kaillera_Keys[2] & 0x80) >> 7;
    Controller_1B_A = (Kaillera_Keys[2] & 0x40) >> 6;
    Controller_1B_B = (Kaillera_Keys[2] & 0x20) >> 5;
    Controller_1B_C = (Kaillera_Keys[2] & 0x10) >> 4;

    if(Controller_1B_Type & 1)
    {
      Controller_1B_Mode = (Kaillera_Keys[2 + 1] & 0x08) >> 3;
      Controller_1B_X = (Kaillera_Keys[2 + 1] & 0x04) >> 2;
      Controller_1B_Y = (Kaillera_Keys[2 + 1] & 0x02) >> 1;
      Controller_1B_Z = (Kaillera_Keys[2 + 1] & 0x01);
    }

    Controller_1C_Up = (Kaillera_Keys[4] & 0x08) >> 3;
    Controller_1C_Down = (Kaillera_Keys[4] & 0x04) >> 2;
    Controller_1C_Left = (Kaillera_Keys[4] & 0x02) >> 1;
    Controller_1C_Right = (Kaillera_Keys[4] & 0x01);
    Controller_1C_Start = (Kaillera_Keys[4] & 0x80) >> 7;
    Controller_1C_A = (Kaillera_Keys[4] & 0x40) >> 6;
    Controller_1C_B = (Kaillera_Keys[4] & 0x20) >> 5;
    Controller_1C_C = (Kaillera_Keys[4] & 0x10) >> 4;

    if(Controller_1C_Type & 1)
    {
      Controller_1C_Mode = (Kaillera_Keys[4 + 1] & 0x08) >> 3;
      Controller_1C_X = (Kaillera_Keys[4 + 1] & 0x04) >> 2;
      Controller_1C_Y = (Kaillera_Keys[4 + 1] & 0x02) >> 1;
      Controller_1C_Z = (Kaillera_Keys[4 + 1] & 0x01);
    }

    Controller_1D_Up = (Kaillera_Keys[6] & 0x08) >> 3;
    Controller_1D_Down = (Kaillera_Keys[6] & 0x04) >> 2;
    Controller_1D_Left = (Kaillera_Keys[6] & 0x02) >> 1;
    Controller_1D_Right = (Kaillera_Keys[6] & 0x01);
    Controller_1D_Start = (Kaillera_Keys[6] & 0x80) >> 7;
    Controller_1D_A = (Kaillera_Keys[6] & 0x40) >> 6;
    Controller_1D_B = (Kaillera_Keys[6] & 0x20) >> 5;
    Controller_1D_C = (Kaillera_Keys[6] & 0x10) >> 4;

    if(Controller_1D_Type & 1)
    {
      Controller_1D_Mode = (Kaillera_Keys[6 + 1] & 0x08) >> 3;
      Controller_1D_X = (Kaillera_Keys[6 + 1] & 0x04) >> 2;
      Controller_1D_Y = (Kaillera_Keys[6 + 1] & 0x02) >> 1;
      Controller_1D_Z = (Kaillera_Keys[6 + 1] & 0x01);
    }
  }
  else
  {
    Controller_2_Up = (Kaillera_Keys[2] & 0x08) >> 3;
    Controller_2_Down = (Kaillera_Keys[2] & 0x04) >> 2;
    Controller_2_Left = (Kaillera_Keys[2] & 0x02) >> 1;
    Controller_2_Right = (Kaillera_Keys[2] & 0x01);
    Controller_2_Start = (Kaillera_Keys[2] & 0x80) >> 7;
    Controller_2_A = (Kaillera_Keys[2] & 0x40) >> 6;
    Controller_2_B = (Kaillera_Keys[2] & 0x20) >> 5;
    Controller_2_C = (Kaillera_Keys[2] & 0x10) >> 4;

    if(Controller_2_Type & 1)
    {
      Controller_2_Mode = (Kaillera_Keys[2 + 1] & 0x08) >> 3;
      Controller_2_X = (Kaillera_Keys[2 + 1] & 0x04) >> 2;
      Controller_2_Y = (Kaillera_Keys[2 + 1] & 0x02) >> 1;
      Controller_2_Z = (Kaillera_Keys[2 + 1] & 0x01);
    }
  }
}


LRESULT CALLBACK ControllerProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  RECT r;
  RECT r2;
  int dx1, dy1, dx2, dy2;
  int i;
  static HWND Tex0 = NULL;

  switch(uMsg)
  {
    case WM_INITDIALOG:
      if(Full_Screen)
      { while(ShowCursor(FALSE) >= 0);
        while(ShowCursor(TRUE) < 0); }

      if(Movie_Status & MOVIE_PLAYING) // must return to the original config
      {
        for(i = 0; i < 8; i++)
        { dx1 = Control[i].Type;
          Control[i].Type = Movie_Save_Joys[i];
          Movie_Save_Joys[i] = dx1; }
      }

      GetWindowRect(ghWnd, &r);
      dx1 = (r.right - r.left) / 2;
      dy1 = (r.bottom - r.top) / 2;

      GetWindowRect(hDlg, &r2);
      dx2 = (r2.right - r2.left) / 2;
      dy2 = (r2.bottom - r2.top) / 2;

      SetWindowPos(hDlg, NULL, r.left + (dx1 - dx2), r.top + (dy1 - dy2), NULL, NULL, SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW);

      Tex0 = GetDlgItem(hDlg, IDC_STATIC_TEXT0);

      if(!Init_Input(ghInstance, hDlg)) return false;

      if(Multiple_Controls_Config)
      {
        GetWindowText(hDlg, Str_Tmp, 968);
        if(*Rom_Name)
        { strcat(Str_Tmp, " - ["); strcat(Str_Tmp, Rom_Name); strcat(Str_Tmp, "]"); }
        else strcat(Str_Tmp, " - [default]");
        SetWindowText(hDlg, Str_Tmp);
      }

      WORD_L(IDC_JOYINFO1, "Controller info 1", "", "Player 1-B 1-C and 1-D are enabled only if a teamplayer is connected to port 1");
      WORD_L(IDC_JOYINFO2, "Controller info 2", "", "Player 2-B 2-C and 2-D are enabled only if a teamplayer is connected to port 2");
      WORD_L(IDC_JOYINFO3, "Controller info 3", "", "Only a few games support teamplayer (games which have 4 players support), so don't forget to use the \"load config\" and \"save config\" possibility :)");

      for(i = 0; i < 2; i++)
      {
        SendDlgItemMessage(hDlg, IDC_COMBO_PORT1 + i, CB_INSERTSTRING, (WPARAM) 0, (LONG) (LPTSTR) "teamplayer");
        SendDlgItemMessage(hDlg, IDC_COMBO_PORT1 + i, CB_INSERTSTRING, (WPARAM) 0, (LONG) (LPTSTR) "pad");
      }

      for(i = 0; i < 8; i++)
      {
        SendDlgItemMessage(hDlg, IDC_COMBO_PADP1 + i, CB_INSERTSTRING, (WPARAM) 0, (LONG) (LPTSTR) "6 buttons");
        SendDlgItemMessage(hDlg, IDC_COMBO_PADP1 + i, CB_INSERTSTRING, (WPARAM) 0, (LONG) (LPTSTR) "3 buttons");
      }

      SendDlgItemMessage(hDlg, IDC_COMBO_PORT1, CB_SETCURSEL, (WPARAM) ((Controller_1_Type >> 4) & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PORT2, CB_SETCURSEL, (WPARAM) ((Controller_2_Type >> 4) & 1), (LPARAM) 0);

      SendDlgItemMessage(hDlg, IDC_COMBO_PADP1, CB_SETCURSEL, (WPARAM) (Controller_1_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP1B, CB_SETCURSEL, (WPARAM) (Controller_1B_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP1C, CB_SETCURSEL, (WPARAM) (Controller_1C_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP1D, CB_SETCURSEL, (WPARAM) (Controller_1D_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP2, CB_SETCURSEL, (WPARAM) (Controller_2_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP2B, CB_SETCURSEL, (WPARAM) (Controller_2B_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP2C, CB_SETCURSEL, (WPARAM) (Controller_2C_Type & 1), (LPARAM) 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PADP2D, CB_SETCURSEL, (WPARAM) (Controller_2D_Type & 1), (LPARAM) 0);
      break;

    case WM_COMMAND:
      switch(wParam)
      {

        case IDOK:
          Controller_1_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PORT1, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_1_Type <<= 4;

          Controller_1_Type |= (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_1B_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1B, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_1C_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1C, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_1D_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1D, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);


          Controller_2_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PORT2, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_2_Type <<= 4;

          Controller_2_Type |= (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_2B_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2B, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_2C_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2C, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);
          Controller_2D_Type = (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2D, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1);

          Make_IO_Table();

        case IDCANCEL:
          if(Movie_Status & MOVIE_PLAYING) // must return to the original config
          {
            for(i = 0; i < 8; i++)
            { dx1 = Control[i].Type;
              Control[i].Type = Movie_Save_Joys[i];
              Movie_Save_Joys[i] = dx1; }
          }
          End_Input();
          EndDialog(hDlg, TRUE);
          break;

        case IDC_BUTTON_SETKEYSP1:
          SetWindowText(Tex0, "SETTING KEYS P1");
          Setting_Keys(hDlg, 0, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(Tex0, "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP1B:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P1B");
          Setting_Keys(hDlg, 2, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1B, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP1C:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P1C");
          Setting_Keys(hDlg, 3, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1C, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP1D:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P1D");
          Setting_Keys(hDlg, 4, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP1D, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP2:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P2");
          Setting_Keys(hDlg, 1, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP2B:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P2B");
          Setting_Keys(hDlg, 5, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2B, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP2C:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P2C");
          Setting_Keys(hDlg, 6, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2C, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case IDC_BUTTON_SETKEYSP2D:
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS P2D");
          Setting_Keys(hDlg, 7, (SendDlgItemMessage(hDlg, IDC_COMBO_PADP2D, CB_GETCURSEL, (WPARAM) 0, (LPARAM) 0) & 1));
          SetWindowText(GetDlgItem(hDlg, IDC_STATIC_TEXT0), "SETTING KEYS");
          break;

        case ID_HELP_HELP:
          if(GetFileAttributes(Manual_Path) & FILE_ATTRIBUTE_DIRECTORY)    // can be used to detect if file exist
          {
            strcpy(Str_Tmp, Manual_Path);
            strcat(Str_Tmp, " helpjoypads.html");
            system(Str_Tmp);
          }
          else
          {
            MessageBox(NULL, "You need to configure the Manual Path to have help", "info", MB_OK);
          }
          break;
      }
      break;

    case WM_CLOSE:
      if(Movie_Status & MOVIE_PLAYING) // must return to the original config
      {
        for(i = 0; i < 8; i++)
        { dx1 = Control[i].Type;
          Control[i].Type = Movie_Save_Joys[i];
          Movie_Save_Joys[i] = dx1; }
      }
      End_Input();
      EndDialog(hDlg, TRUE);
      break;
  }

  return FALSE;
}



void __stdcall _ShowHotkey(HWND hDlg, DWORD hotkey, UINT type)
{
  int i, check, max;
  const int btn_ctrl[12] = { IDC_CHECK_UP, IDC_CHECK_LEFT, IDC_CHECK_RIGHT, IDC_CHECK_DOWN,
    IDC_CHECK_A, IDC_CHECK_B, IDC_CHECK_C, IDC_CHECK_START,
    IDC_CHECK_X, IDC_CHECK_Y, IDC_CHECK_Z, IDC_CHECK_MODE };
  const DWORD btn_value[12] = { HKB_UP, HKB_LEFT, HKB_RIGHT, HKB_DOWN,
    HKB_A, HKB_B, HKB_C, HKB_START, HKB_X, HKB_Y, HKB_Z, HKB_MODE };

  if(type & 1)
  {
    max = 12;
    for(i = 8; i < 12; i++) EnableWindow(GetDlgItem(hDlg, btn_ctrl[i]), TRUE);
  }
  else
  {
    max = 8;
    for(i = 8; i < 12; i++)
    { SendDlgItemMessage(hDlg, btn_ctrl[i], BM_SETCHECK, BST_UNCHECKED, 0);
      EnableWindow(GetDlgItem(hDlg, btn_ctrl[i]), FALSE); }
  }
  for(i = 0; i < max; i++)
  {
    if(btn_value[i] & hotkey) check = BST_CHECKED;
    else                      check = BST_UNCHECKED;
    SendDlgItemMessage(hDlg, btn_ctrl[i], BM_SETCHECK, check, 0);
  }
}

DWORD __stdcall _GetHotkey(HWND hDlg, UINT type)
{
  DWORD hotkey;
  int i, max;
  const int btn_ctrl[12] = { IDC_CHECK_UP, IDC_CHECK_LEFT, IDC_CHECK_RIGHT, IDC_CHECK_DOWN,
    IDC_CHECK_A, IDC_CHECK_B, IDC_CHECK_C, IDC_CHECK_X, IDC_CHECK_Y, IDC_CHECK_Z,
    IDC_CHECK_START, IDC_CHECK_MODE };
  const DWORD btn_value[12] = { HKB_UP, HKB_LEFT, HKB_RIGHT, HKB_DOWN,
    HKB_A, HKB_B, HKB_C, HKB_X, HKB_Y, HKB_Z,
    HKB_START, HKB_MODE };

  hotkey = 0;
  max = (type & 1) ? 12 : 8;
  for(i = 0; i < max; i++)
  {
    if(SendDlgItemMessage(hDlg, btn_ctrl[i], BM_GETCHECK, 0, 0) == BST_CHECKED)
      hotkey |= btn_value[i];
  }

  return hotkey;
}

DWORD Hotkeys_Buf[8][4];

LRESULT CALLBACK HotkeysProc(HWND hDlg, UINT uMes, WPARAM wParam, LPARAM lParam)
{
  DWORD dwID;
  RECT r;
  RECT r2;
  int dx1, dy1, dx2, dy2, i;
  const int p_ctrls[8] = { IDC_CHECK_P1, IDC_CHECK_P1B, IDC_CHECK_P1C, IDC_CHECK_P1D,
    IDC_CHECK_P2, IDC_CHECK_P2B, IDC_CHECK_P2C, IDC_CHECK_P2D };
  const int hk_ctrls[4] = { IDC_CHECK_HOTKEY1, IDC_CHECK_HOTKEY2, IDC_CHECK_HOTKEY3, IDC_CHECK_HOTKEY4 };
  const unsigned int type[8] = { Controller_1_Type, Controller_1B_Type,
    Controller_1C_Type, Controller_1D_Type, Controller_2_Type, Controller_2B_Type,
    Controller_2C_Type, Controller_2D_Type };

  switch(uMes)
  {
    case WM_INITDIALOG:
      GetWindowRect(ghWnd, &r);
      dx1 = (r.right - r.left) / 2;
      dy1 = (r.bottom - r.top) / 2;
      GetWindowRect(hDlg, &r2);
      dx2 = (r2.right - r2.left) / 2;
      dy2 = (r2.bottom - r2.top) / 2;
      SetWindowPos(hDlg, NULL, r.left + (dx1 - dx2), r.top + (dy1 - dy2), NULL, NULL, SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW);
      SendDlgItemMessage(hDlg, IDC_CHECK_P1, BM_SETCHECK, BST_CHECKED, 0);
      SendDlgItemMessage(hDlg, IDC_CHECK_HOTKEY1, BM_SETCHECK, BST_CHECKED, 0);
      for(i = 0; i < 8; i++)
      { for(dx1 = 0; dx1 < 4; dx1++)
        { Hotkeys_Buf[i][dx1] = Keys_Def[i].HotKey_Buttons[dx1]; } }
      _ShowHotkey(hDlg, Hotkeys_Buf[0][0], type[0]);
      return 1;

    case WM_CLOSE:
      EndDialog(hDlg, 0);
      break;

    case WM_COMMAND:
      dwID  = LOWORD(wParam);
      dy1   = 9;
      switch(dwID)
      {
        case IDOK:
          SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_CHECK_HOTKEY1, BST_PUSHED), (LPARAM)GetDlgItem(hDlg, IDC_CHECK_HOTKEY1));
          for(i = 0; i < 8; i++)
          { for(dx1 = 0; dx1 < 4; dx1++)
            { Keys_Def[i].HotKey_Buttons[dx1] = Hotkeys_Buf[i][dx1]; } }
        case IDCANCEL:
          EndDialog(hDlg, 0);
          break;

        case IDC_CHECK_P1:
        case IDC_CHECK_P1B:
        case IDC_CHECK_P1C:
        case IDC_CHECK_P1D:
        case IDC_CHECK_P2:
        case IDC_CHECK_P2B:
        case IDC_CHECK_P2C:
        case IDC_CHECK_P2D:
          for(dy1 = 0; dy1 < 9; dy1++)
          { if(SendDlgItemMessage(hDlg, p_ctrls[dy1], BM_GETCHECK, 0, 0) == BST_CHECKED) break; }
          for(i = 0; i < 9; i++)
          { SendDlgItemMessage(hDlg, p_ctrls[i], BM_SETCHECK, (p_ctrls[i] == dwID) ? BST_CHECKED : BST_UNCHECKED, 0); }
          dwID = IDC_CHECK_HOTKEY1;

        case IDC_CHECK_HOTKEY1:
        case IDC_CHECK_HOTKEY2:
        case IDC_CHECK_HOTKEY3:
        case IDC_CHECK_HOTKEY4:
          if(dy1 > 8)
          {
            for(dy1 = 0; dy1 < 9; dy1++)
            { if(SendDlgItemMessage(hDlg, p_ctrls[dy1], BM_GETCHECK, 0, 0) == BST_CHECKED) break; }
          }
          for(dy2 = 0; dy2 < 9; dy2++)
          { if(SendDlgItemMessage(hDlg, hk_ctrls[dy2], BM_GETCHECK, 0, 0) == BST_CHECKED) break; }

          dx2 = 5;
          for(i = 0; i < 5; i++)
          { SendDlgItemMessage(hDlg, hk_ctrls[i], BM_SETCHECK, (hk_ctrls[i] == dwID) ? dx2 = i, BST_CHECKED : BST_UNCHECKED, 0); }
          for(dx1 = 0; dx1 < 9; dx1++)
          { if(SendDlgItemMessage(hDlg, p_ctrls[dx1], BM_GETCHECK, 0, 0) == BST_CHECKED) break; }
          if(dy1 < 9 && dy2 < 5) Hotkeys_Buf[dy1][dy2] = _GetHotkey(hDlg, type[dy1]);
          if(dx1 < 9 && dx2 < 5) _ShowHotkey(hDlg, Hotkeys_Buf[dx1][dx2], type[dx1]);
          break;
      }
      break;
  }
  return 0;
}
