/******************************************************************************
* In order to make life a little bit easier when using the GIF file format,   *
* this library was written, and which does all the dirty work...              *
*                                                                             *
*                                        Written by Gershon Elber,  Jun. 1989 *
*                                        Hacks by Eric S. Raymond,  Sep. 1992 *
*******************************************************************************
* History:                                                                    *
* 14 Jun 89 - Version 1.0 by Gershon Elber.                                   *
*  3 Sep 90 - Version 1.1 by Gershon Elber (Support for Gif89, Unique names)  *
* 15 Sep 90 - Version 2.0 by Eric S. Raymond (Changes to suoport GIF slurp)   *
* 26 Jun 96 - Version 3.0 by Eric S. Raymond (Full GIF89 support)             *
* 17 Dec 98 - Version 4.0 by Toshio Kuratomi (Fix extension writing code)     *
******************************************************************************/

#ifndef GIFLIB_H
#define GIFLIB_H

#include <windows.h>

#ifndef __MSDOS__
# define __MSDOS__
#endif

#define GIF_ERROR  0
#define GIF_OK     1

#define GIF_FILE_BUFFER_SIZE 16384  // Files uses bigger buffers than usual

typedef BYTE   GIFPIXEL;

#define GIF_MESSAGE(Msg) fprintf(stderr, "\n%s: %s\n", PROGRAM_NAME, Msg)
#define GIF_EXIT(Msg)  { GIF_MESSAGE(Msg); exit(-3); }

typedef struct _GIFCOLOR {
  BYTE Red, Green, Blue;
} GIFCOLOR;

typedef struct _COLORMAPOBJECT
{
  int  ColorCount;
  int BitsPerPixel;
  GIFCOLOR *Colors;    // on malloc(3) heap
} COLORMAPOBJECT;

typedef struct _GIFIMAGEDESC {
  int Left, Top, Width, Height,  /* Current image dimensions. */
  Interlace;      /* Sequential/Interlaced lines. */
  COLORMAPOBJECT *ColorMap;    /* The local color map */
} GIFIMAGEDESC;

typedef struct _GIFFILE {
  int SWidth, SHeight,             // Screen dimensions
  SColorResolution,                // How many colors can we generate?
  SBackGroundColor;                // I hope you understand this one...
  COLORMAPOBJECT *SColorMap;       // NULL if not exists.
  int ImageCount;                  // Number of current image
  GIFIMAGEDESC Image;              // Block describing current image
  struct SavedImage *SavedImages;  // Use this to accumulate file state
  LPVOID UserData;                 // hook to attach user data (TVT)
  LPVOID Private;                  // Don't mess with this!
} GIFFILE;

typedef int GIFRECORDTYPE;

#define UNDEFINED_RECORD_TYPE   0
#define SCREEN_DESC_RECORD_TYPE 1
#define IMAGE_DESC_RECORD_TYPE  2 // Begin with ','
#define EXTENSION_RECORD_TYPE   3 // Begin with '!'
#define TERMINATE_RECORD_TYPE   4 // Begin with ';'

/* DumpScreen2Gif routine constants identify type of window/screen to dump.  */
/* Note all values below 1000 are reserved for the IBMPC different display   */
/* devices (it has many!) and are compatible with the numbering TC2.0        */
/* (Turbo C 2.0 compiler for IBM PC) gives to these devices.         */
typedef enum {
    GIF_DUMP_SGI_WINDOW = 1000,
    GIF_DUMP_X_WINDOW = 1001
} GifScreenDumpType;

// func type to read gif data from arbitrary sources (TVT)
typedef int (__stdcall *GIFINPUTFUNC)(GIFFILE *, LPBYTE, int);

/* func type to write gif data ro arbitrary targets.
 * Returns count of bytes written. (MRB) */
typedef int (__stdcall *GIFOUTPUTFUNC)(GIFFILE *, LPCBYTE, int);
/******************************************************************************
*  GIF89 extension function codes                                             *
******************************************************************************/

#define COMMENT_EXT_FUNC_CODE     0xFE  // comment
#define GRAPHICS_EXT_FUNC_CODE    0xF9  // graphics control
#define PLAINTEXT_EXT_FUNC_CODE   0x01  // plaintext
#define APPLICATION_EXT_FUNC_CODE 0xFF  // application block

/******************************************************************************
* O.K., here are the routines one can access in order to encode GIF file:     *
* (GIF_LIB file EGIF_LIB.C).                  *
******************************************************************************/

GIFFILE *EGifOpenFileName(char *GifFileName);
GIFFILE *EGifOpen(void *userPtr, GIFOUTPUTFUNC writeFunc);
int EGifSpew(GIFFILE *GifFile);
void EGifSetGifVersion(char *Version);
int EGifPutScreenDesc(GIFFILE *GifFile,
  int GifWidth, int GifHeight, int GifColorRes, int GifBackGround,
  COLORMAPOBJECT *GifColorMap);
int EGifPutImageDesc(GIFFILE *GifFile,
  int GifLeft, int GifTop, int Width, int GifHeight, int GifInterlace,
  COLORMAPOBJECT *GifColorMap);
int EGifPutLine(GIFFILE *GifFile, GIFPIXEL *GifLine, int GifLineLen);
int EGifPutPixel(GIFFILE *GifFile, GIFPIXEL GifPixel);
int EGifPutComment(GIFFILE *GifFile, char *GifComment);
int EGifPutExtension(GIFFILE *GifFile, int GifExtCode, int GifExtLen,
              LPVOID GifExtension);
int EGifPutCode(GIFFILE *GifFile, int GifCodeSize,
               BYTE *GifCodeBlock);
int EGifPutCodeNext(GIFFILE *GifFile, BYTE *GifCodeBlock);
int EGifCloseFile(GIFFILE *GifFile);

#define E_GIF_ERR_OPEN_FAILED    1
#define E_GIF_ERR_WRITE_FAILED   2
#define E_GIF_ERR_HAS_SCRN_DSCR  3
#define E_GIF_ERR_HAS_IMAG_DSCR  4
#define E_GIF_ERR_NO_COLOR_MAP   5
#define E_GIF_ERR_DATA_TOO_BIG   6
#define E_GIF_ERR_NOT_ENOUGH_MEM 7
#define E_GIF_ERR_DISK_IS_FULL   8
#define E_GIF_ERR_CLOSE_FAILED   9
#define E_GIF_ERR_NOT_WRITEABLE  10

/******************************************************************************
* O.K., here are the routines one can access in order to decode GIF file:     *
* (GIF_LIB file DGIF_LIB.C).                  *
******************************************************************************/

GIFFILE *DGifOpenFileName(const char *GifFileName);
GIFFILE *DGifOpen(LPVOID userPtr, GIFINPUTFUNC readFunc);  /* new one (TVT) */
int DGifSlurp(GIFFILE *GifFile);
int DGifGetScreenDesc(GIFFILE *GifFile);
int DGifGetRecordType(GIFFILE *GifFile, GIFRECORDTYPE *GifType);
int DGifGetImageDesc(GIFFILE *GifFile);
int DGifGetLine(GIFFILE *GifFile, GIFPIXEL *GifLine, int GifLineLen);
int DGifGetPixel(GIFFILE *GifFile, GIFPIXEL GifPixel);
int DGifGetComment(GIFFILE *GifFile, char *GifComment);
int DGifGetExtension(GIFFILE *GifFile, int *GifExtCode, LPBYTE *GifExtension);
int DGifGetExtensionNext(GIFFILE *GifFile, BYTE **GifExtension);
int DGifGetCode(GIFFILE *GifFile, int *GifCodeSize, LPBYTE *GifCodeBlock);
int DGifGetCodeNext(GIFFILE *GifFile, BYTE **GifCodeBlock);
int DGifGetLZCodes(GIFFILE *GifFile, int *GifCode);
int DGifCloseFile(GIFFILE *GifFile);

#define D_GIF_ERR_OPEN_FAILED    101
#define D_GIF_ERR_READ_FAILED    102
#define D_GIF_ERR_NOT_GIF_FILE   103
#define D_GIF_ERR_NO_SCRN_DSCR   104
#define D_GIF_ERR_NO_IMAG_DSCR   105
#define D_GIF_ERR_NO_COLOR_MAP   106
#define D_GIF_ERR_WRONG_RECORD   107
#define D_GIF_ERR_DATA_TOO_BIG   108
#define D_GIF_ERR_NOT_ENOUGH_MEM 109
#define D_GIF_ERR_CLOSE_FAILED   110
#define D_GIF_ERR_NOT_READABLE   111
#define D_GIF_ERR_IMAGE_DEFECT   112
#define D_GIF_ERR_EOF_TOO_SOON   113

/******************************************************************************
* O.K., here are the routines from GIF_LIB file QUANTIZE.C.          *
******************************************************************************/
int QuantizeBuffer(unsigned int Width, unsigned int Height, int *ColorMapSize,
  BYTE *RedInput, BYTE *GreenInput, BYTE *BlueInput,
  BYTE *OutputBuffer, GIFCOLOR *OutputColorMap);


/******************************************************************************
* O.K., here are the routines from GIF_LIB file QPRINTF.C.          *
******************************************************************************/
extern int GifQuietPrint;

#ifdef USE_VARARGS
extern void GifQprintf();
#else
extern void GifQprintf(char *Format, ...);
#endif /* USE_VARARGS */

/******************************************************************************
* O.K., here are the routines from GIF_LIB file GIF_ERR.C.          *
******************************************************************************/
extern void PrintGifError(void);
extern int GifLastError(void);

/******************************************************************************
* O.K., here are the routines from GIF_LIB file DEV2GIF.C.          *
******************************************************************************/
extern int DumpScreen2Gif(char *FileName,
        int ReqGraphDriver,
        int ReqGraphMode1,
        int ReqGraphMode2,
        int ReqGraphMode3);

/*****************************************************************************
 *
 * Everything below this point is new after version 1.2, supporting `slurp
 * mode' for doing I/O in two big belts with all the image-bashing in core.
 *
 *****************************************************************************/

/******************************************************************************
* Color Map handling from ALLOCGIF.C                *
******************************************************************************/

extern COLORMAPOBJECT *MakeMapObject(int ColorCount, GIFCOLOR *ColorMap);
extern void FreeMapObject(COLORMAPOBJECT *Object);
extern COLORMAPOBJECT *UnionColorMap(COLORMAPOBJECT *ColorIn1,
            COLORMAPOBJECT *ColorIn2,
            GIFPIXEL ColorTransIn2[]);
extern int BitSize(int n);

/******************************************************************************
* Support for the in-core structures allocation (slurp mode).          *
******************************************************************************/

/* This is the in-core version of an extension record */
typedef struct {
    int    ByteCount;
    char  *Bytes;    /* on malloc(3) heap */
    int Function;       /* Holds the type of the Extension block. */
} ExtensionBlock;

/* This holds an image header, its unpacked raster bits, and extensions */
typedef struct SavedImage {
    GIFIMAGEDESC  ImageDesc;

    char    *RasterBits;    /* on malloc(3) heap */

    int      Function; /* DEPRECATED: Use ExtensionBlocks[x].Function
                           * instead */
    int      ExtensionBlockCount;
    ExtensionBlock  *ExtensionBlocks;  /* on malloc(3) heap */
} SavedImage;

extern void ApplyTranslation(SavedImage *Image, GIFPIXEL Translation[]);

extern void MakeExtension(SavedImage *New, int Function);
extern int AddExtensionBlock(SavedImage *New, int Len, char ExtData[]);
extern void FreeExtension(SavedImage *Image);

extern SavedImage *MakeSavedImage(GIFFILE *GifFile, SavedImage *CopyFrom);
extern void FreeSavedImages(GIFFILE *GifFile);

/******************************************************************************
* The library's internal utility font                *
******************************************************************************/

#define GIF_FONT_WIDTH  8
#define GIF_FONT_HEIGHT  8
extern unsigned char AsciiTable[][GIF_FONT_WIDTH];

extern void DrawText(SavedImage *Image,
         const int x, const int y,
         const char *legend,
         const int color);

extern void DrawBox(SavedImage *Image,
         const int x, const int y,
         const int w, const int d,
         const int color);

void DrawRectangle(SavedImage *Image,
         const int x, const int y,
         const int w, const int d,
         const int color);

extern void DrawBoxedText(SavedImage *Image,
         const int x, const int y,
         const char *legend,
         const int border,
         const int bg,
         const int fg);


HBITMAP __stdcall LoadGIF(HINSTANCE, LPCSTR);

#endif // GIFLIB_H
