// Code added by Rodrigo Cardoso. Loads an gif file from resource or disk as a bitmap.

#include <windows.h>
#include <stdio.h>
#include "giflib.h"

typedef struct _BITMAPBUF {
  BITMAPINFOHEADER bih;
  RGBQUAD          pal[256];
  BYTE             bits[4];
} BITMAPBUF, *LPBITMAPBUF;

int __stdcall GifRead(GIFFILE *gFile, LPBYTE pData, int iSize)
{
  int c, i;

  for(i = 0; i < iSize; i++)
  {
    c = getc(((FILE *)gFile->UserData));
    if(c == EOF) break;
    pData[i] = c;
  }

  return i;
}


HBITMAP __stdcall LoadGIF(HINSTANCE hInstance, LPCSTR lpGifName)
{
  FILE       *pFile   = NULL;
  LPBITMAPBUF pBMP    = NULL;
  GIFFILE    *gFile   = NULL;
  HBITMAP     hBitmap = NULL;
  HDC         hDC     = NULL;
  HRSRC       hRes;
  HGLOBAL     hResData;
  LPVOID      pRes;
  DWORD       dwSize;
  int         x, iLineSize;

  if(hInstance) // resource
  {
    char pszTmp[MAX_PATH];
    hRes = FindResource(hInstance, lpGifName, RT_RCDATA);
    if(!hRes)     return NULL;
    dwSize = SizeofResource(hInstance, hRes);
    if(!dwSize)   return NULL;
    hResData = LoadResource(hInstance, hRes);
    if(!hResData) return NULL;
    pRes = (LPBYTE)LockResource(hResData);
    GetTempFileName(".", "gif", 0, pszTmp);
    pFile = fopen(pszTmp, "w+bD");
    if(!pFile)    return NULL;
    if(fwrite(pRes, dwSize, 1, pFile) != 1) goto finish;
    if(fseek(pFile, 0, SEEK_SET))           goto finish;
  }
  else          // file
  {
    pFile = fopen(lpGifName, "rb");
    if(!pFile) return NULL;   // error load image
  }

  gFile = DGifOpen(pFile, GifRead);
  if(!gFile)                                    goto finish;
  if(DGifGetRecordType(gFile, &x) == GIF_ERROR) goto finish;
  if(DGifGetImageDesc(gFile) == GIF_ERROR)      goto finish;
  if(!gFile->SColorMap)                         goto finish;

  iLineSize = gFile->SWidth;
  if(iLineSize & 0x03) iLineSize = (iLineSize & (~0x03)) + 4;

  // alloc memory for the output image
  pBMP = (LPBITMAPBUF)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
    sizeof(BITMAPINFOHEADER) + 1024 + (gFile->SHeight * iLineSize));
  if(!pBMP) goto finish;

  // copy the color palette
  for(x = 0; x < gFile->SColorMap->ColorCount; x++)
  {
    pBMP->pal[x].rgbRed   = gFile->SColorMap->Colors[x].Red;
    pBMP->pal[x].rgbGreen = gFile->SColorMap->Colors[x].Green;
    pBMP->pal[x].rgbBlue  = gFile->SColorMap->Colors[x].Blue;
  }

  pBMP->bih.biSize     = sizeof(BITMAPINFOHEADER);
  pBMP->bih.biBitCount = 8;
  pBMP->bih.biWidth    = gFile->SWidth;
  pBMP->bih.biHeight   = gFile->SHeight;
  pBMP->bih.biPlanes   = 1;
  hDC = CreateDC("DISPLAY", NULL, NULL, NULL);
  hBitmap = CreateDIBitmap(hDC, (LPBITMAPINFOHEADER)pBMP, 0, NULL, (LPBITMAPINFO)pBMP,
    DIB_RGB_COLORS);
  if(!hBitmap) goto finish;

  // reading the image
  for(x = 0; x < gFile->SHeight; x++)
  {
    if(DGifGetLine(gFile, pBMP->bits + ((gFile->SHeight - x - 1) * iLineSize), 0) == GIF_ERROR)
      goto finish;
  }

  SetDIBits(hDC, hBitmap, 0, gFile->SHeight, pBMP->bits, (LPBITMAPINFO)pBMP, DIB_RGB_COLORS);

  finish:
  if(hDC)     DeleteDC(hDC);
  if(pBMP)    HeapFree(GetProcessHeap(), 0, pBMP);
  if(gFile)   DGifCloseFile(gFile);
  if(pFile)   fclose(pFile);

  return hBitmap;
}
