
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include "grom.h"
#include "gdsound.h"
#include "gmain.h"
#include "gmenu.h"
#include "gens.h"
#include "ggenie.h"
#include "cpu_68k.h"
#include "cd_sys.h"
#include "mem_M68K.h"
#include "mem_SH2.h"
#include "vdp_io.h"
#include "state.h"
#include "ccnet.h"
#include "misc.h"
#include "zlib.h"
#include "unzip.h"
#include "unrar.hpp"
#include "wave.h"
#include "script.h"
#include "sgm.h"
#include "vgm.h"
#include "sms.h"
#include "movie.h"

int         File_Type_Index;
LPGROM_INFO Rom_Info = NULL;
char        Rom_Name[G_MAX_ROM_NAME];

void Get_Name_From_Path(char *Full_Path, char *Name)
{
  int i = 0;

  i = strlen(Full_Path) - 1;

  while ((i >= 0) && (Full_Path[i] != '\\')) i--;

  if(i <= 0)
  {
    Name[0] = 0;
  }
  else
  {
    strcpy(Name, &Full_Path[++i]);
  }
}


void Get_Dir_From_Path(char *Full_Path, char *Dir)
{
  int i = 0;

  i = strlen(Full_Path) - 1;

  while ((i >= 0) && (Full_Path[i] != '\\')) i--;

  if(i <= 0)
  {
    Dir[0] = 0;
  }
  else
  {
    strncpy(Dir, Full_Path, ++i);
    Dir[i] = 0;
  }
}


void __stdcall Update_Recent_Rom(LPCSTR pszRom, int iType)
{
  int   i, exist;
  char  szTmp[GMAX_PATH];

  if(pszRom)
  {
    for(exist = i = 0; i < 9; i++)
    {
      if(!*Recent_Rom[i]) break;
      else if(!strcmpi(Recent_Rom[i], pszRom))
      { exist = 1; break; }
    }

    if(exist)
    {
      if(i)
      {
        strcpy(szTmp, Recent_Rom[i]);
        memmove(Recent_Rom[1], Recent_Rom[0], GMAX_PATH * i);
        strcpy(Recent_Rom[0], szTmp);
        memmove(Recent_Rom_Type + 1, Recent_Rom_Type, 4 * i);
        Recent_Rom_Type[0] = iType;
      }
    }
    else
    {
      memmove(Recent_Rom[1], Recent_Rom[0], GMAX_PATH * (GMAX_RECENT_ROM - 1));
      strcpy(Recent_Rom[0], pszRom);
      memmove(Recent_Rom_Type + 1, Recent_Rom_Type, 4 * (GMAX_RECENT_ROM - 1));
      Recent_Rom_Type[0] = iType;
    }
  }
}


void Update_Rom_Dir(char *Path)
{
  if(Use_Last_Open_Rom_Dir) Get_Dir_From_Path(Path, Rom_Dir);
}


void __stdcall Update_Rom_Name(LPCSTR Name)
{
  int i, leng;

  leng = strlen(Name) - 1;

  while ((leng >= 0) && (Name[leng] != '\\')) leng--;

  leng++; i = 0;

  while ((Name[leng]) && (Name[leng] != '.'))
    Rom_Name[i++] = Name[leng++];

  Rom_Name[i] = 0;
}


void Update_CD_Rom_Name(char *Name)
{
  int i, j;

  memcpy(Rom_Name, Name, 48);

  for(i = 0; i < 48; i++)
  {
    if((Rom_Name[i] >= '0') && (Rom_Name[i] <= '9')) continue;
    if(Rom_Name[i] == ' ') continue;
    if((Rom_Name[i] >= 'A') && (Rom_Name[i] <= 'Z')) continue;
    if((Rom_Name[i] >= 'a') && (Rom_Name[i] <= 'z')) continue;
    Rom_Name[i] = ' ';
  }

  for(i = 0; i < 48; i++)
  {
    if(Rom_Name[i] != ' ') i = 100;
  }

  if(i < 100) strcpy(Rom_Name, "no name");

  for(i = 47, j = 48; i >= 0; i--, j--)
  {
    if(Rom_Name[i] != ' ') i = -1;
  }

  Rom_Name[j + 1] = 0;
}


int __stdcall Detect_Format(LPCSTR Name)
{
  GZFILE        f;
  HUNZFILE      zf;
  UNZ_FILE_INFO zinf;
  int           zret, etype = 0, i;
  char          zname[256];
  LPSTR         buf = (LPSTR)State_Buffer;
  LPSTR         ext;

  SetCurrentDirectory(Gens_Path);

  memset(buf, 0, 1024);
  ext = strrchr(Name, '.');

  if(ext && (!strcmpi(".ZIP", ext) || !strcmpi(".ZSG", ext)))
  {
    zf = unzOpen(Name);
    if(!zf) return -1;

    zret = unzGoToFirstFile(zf);
    while(zret == UNZ_OK)
    {
      unzGetCurrentFileInfo(zf, &zinf, zname, 128, NULL, 0, NULL, 0);
      ext = strrchr(zname, '.');
      if(ext)
      {
        if(!strcmpi(".SMD", ext) || !strcmpi(".BIN", ext)  || !strcmpi(".GEN", ext) || !strcmpi(".MD", ext))
        { etype = GENESIS_ROM; break; }
        else if(!strcmpi(".32X", ext)) { etype = _32X_ROM; break; }
        else if(!strcmpi(".SMS", ext)) { etype = SMS_ROM;  break; }
        else if(!strcmpi(".GG", ext))  { etype = GG_ROM;   break; }
      }
      zret = unzGoToNextFile(zf);
    }

    if(zret != UNZ_OK) { unzClose(zf); return 0; } // nenhuma rom vlida no zip
    if(etype == SMS_ROM || etype == GG_ROM)        // ROMS de master e gg so conhecidas pela extenso
    { unzClose(zf); return etype; }
    if(unzOpenCurrentFile(zf) != UNZ_OK) return 0;
    unzReadCurrentFile(zf, buf, 1024);
    unzCloseCurrentFile(zf);
    unzClose(zf);
  }
  // TODO: detect RAR format
  else if(ext && !strcmpi(".RAR", ext))
  {
    return 0;
  }
  else
  {
    // roms SMS/GG so reconhecidas pela extenso
    ext = strrchr(Name, '.');
    if(ext)
    {
      if     (!strcmpi(".SMS", ext)) { return SMS_ROM; }
      else if(!strcmpi(".GG", ext))  { return GG_ROM; }
      else if(!strcmpi(".32X", ext)) { etype = _32X_ROM; }
      else                           { etype = 0; }
    }
    else etype = 0;

    f = gzopen(Name, "rb");
    if(f == NULL) return -1;
    i = f->transparent;
    gzread(f, buf, 1024);
    gzclose(f);
    if(i) // verifaca se o arquivo no  gzip (ISOS no podem ser compactadas)
    {
      if(!strnicmp("SEGADISCSYSTEM", buf, 14))        return SEGACD_IMAGE;      // Sega CD (ISO)
      if(!strnicmp("SEGADISCSYSTEM", buf + 0x10, 14)) return SEGACD_IMAGE + 1;  // Sega CD (BIN)
    }
  }

  i = 0;

  if(strnicmp("SEGA", buf + 0x100, 4))
  {
    // Maybe interleaved
    if(!strnicmp("EA", buf + (0x200 + (0x100 / 2)), 2))             i = 1;
    if(buf[0x08] == 0xAA && buf[0x09] == 0xBB && buf[0x0A] == 0x06) i = 1;
  }

  if(i)    // interleaved
  {
    if(etype == _32X_ROM && buf[0x200 / 2] == 0x4E)      return _32X_ROM + 1;
    if(!strnicmp("3X", buf + (0x200 + (0x105 / 2)), 3)) return _32X_ROM + 1;
  }
  else
  {
    if(etype == _32X_ROM && buf[0x200] == 0x4E) return _32X_ROM;
    if(!strnicmp("32X", buf + 0x105, 3))        return _32X_ROM;
  }

  return GENESIS_ROM + i;
}


void De_Interleave(void)
{
  unsigned char buf[16384];
  unsigned char *Src;
  int i, j, Nb_Blocks, ptr;

  Src = &Rom_Data[0x200];

  Rom_Size -= 512;

  Nb_Blocks = Rom_Size / 16384;

  for(ptr = 0, i = 0; i < Nb_Blocks; i++, ptr += 16384)
  {
    memcpy(buf, Src + ptr, 16384);

    for(j = 0; j < 8192; j++)
    {
      Rom_Data[ptr + (j << 1) + 1] = buf[j];
      Rom_Data[ptr + (j << 1)]     = buf[j + 8192];
    }
  }
}


void __stdcall Fill_Infos_MD(void)
{
  int i;

  // Finally we do the IPS patch here, we can have the translated game name

  IPS_Patching();

  memset(Rom_Info, 0, sizeof(GENESIS_ROM_INFO));

  for(i = 0; i < 16; i++)
    Rom_Info->Genesis.Console_Name[i] = Rom_Data[i + 256];

  for(i = 0; i < 16; i++)
    Rom_Info->Genesis.Copyright[i] = Rom_Data[i + 272];

  for(i = 0; i < 48; i++)
    Rom_Info->Genesis.Rom_Name[i] = Rom_Data[i + 288];

  for(i = 0; i < 48; i++)
    Rom_Info->Genesis.Rom_Name_W[i] = Rom_Data[i + 336];

  Rom_Info->Genesis.Type[0] = Rom_Data[384];
  Rom_Info->Genesis.Type[1] = Rom_Data[385];

  for(i = 0; i < 12; i++)
    Rom_Info->Genesis.Version[i] = Rom_Data[i + 386];

  Rom_Info->Genesis.Checksum = (Rom_Data[398] << 8) | Rom_Data[399];

  for(i = 0; i < 16; i++)
    Rom_Info->Genesis.IO_Support[i] = Rom_Data[i + 400];

  Rom_Info->Genesis.Rom_Start_Adress = Rom_Data[416] << 24;
  Rom_Info->Genesis.Rom_Start_Adress |= Rom_Data[417] << 16;
  Rom_Info->Genesis.Rom_Start_Adress |= Rom_Data[418] << 8;
  Rom_Info->Genesis.Rom_Start_Adress |= Rom_Data[419];

  Rom_Info->Genesis.Rom_End_Adress = Rom_Data[420] << 24;
  Rom_Info->Genesis.Rom_End_Adress |= Rom_Data[421] << 16;
  Rom_Info->Genesis.Rom_End_Adress |= Rom_Data[422] << 8;
  Rom_Info->Genesis.Rom_End_Adress |= Rom_Data[423];

  Rom_Info->Genesis.R_Size = Rom_Info->Genesis.Rom_End_Adress - Rom_Info->Genesis.Rom_Start_Adress + 1;

  for(i = 0; i < 12; i++)
    Rom_Info->Genesis.Ram_Infos[i] = Rom_Data[i + 424];

  Rom_Info->Genesis.Ram_Start_Adress = Rom_Data[436] << 24;
  Rom_Info->Genesis.Ram_Start_Adress |= Rom_Data[437] << 16;
  Rom_Info->Genesis.Ram_Start_Adress |= Rom_Data[438] << 8;
  Rom_Info->Genesis.Ram_Start_Adress |= Rom_Data[439];

  Rom_Info->Genesis.Ram_End_Adress = Rom_Data[440] << 24;
  Rom_Info->Genesis.Ram_End_Adress |= Rom_Data[441] << 16;
  Rom_Info->Genesis.Ram_End_Adress |= Rom_Data[442] << 8;
  Rom_Info->Genesis.Ram_End_Adress |= Rom_Data[443];

  for(i = 0; i < 12; i++)
    Rom_Info->Genesis.Modem_Infos[i] = Rom_Data[i + 444];

  for(i = 0; i < 40; i++)
    Rom_Info->Genesis.Description[i] = Rom_Data[i + 456];

  for(i = 0; i < 3; i++)
    Rom_Info->Genesis.Countries[i] = Rom_Data[i + 496];

  Rom_Info->Genesis.Console_Name[16] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Console_Name);
  Rom_Info->Genesis.Copyright[16] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Copyright);
  Rom_Info->Genesis.Rom_Name[48] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Rom_Name);
  Rom_Info->Genesis.Rom_Name_W[48] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Rom_Name_W);
  Rom_Info->Genesis.Type[2] = 0;
  Rom_Info->Genesis.Version[12] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Version);
  Rom_Info->Genesis.IO_Support[12] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.IO_Support);
  Rom_Info->Genesis.Ram_Infos[12] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Ram_Infos);
  Rom_Info->Genesis.Modem_Infos[12] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Modem_Infos);
  Rom_Info->Genesis.Description[40] = 0;
  Clear_String_Spaces(Rom_Info->Genesis.Description);
  Rom_Info->Genesis.Countries[3] = 0;
}


void __stdcall Fill_Infos_MS(void)
{
}


int __stdcall Get_Rom(HWND hWnd)
{
  char         Name[GMAX_PATH];
  LPSTR        ext;
  OPENFILENAME ofn;
  int          sys, ret;

  SetCurrentDirectory(Gens_Path);

  memset(Name, 0, GMAX_PATH);
  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize     = sizeof(OPENFILENAME);
  ofn.hwndOwner       = hWnd;
  ofn.hInstance       = ghInstance;
  ofn.lpstrFile       = Name;
  ofn.nMaxFile        = GMAX_PATH;
  ofn.lpstrFilter     =
    "All suported files\0*.bin;*.smd;*.gen;*.32x;*.sms;*.gg;*.iso;*.raw;*.zip;*.zsg;*.rar\0" \
    "Genesis roms (smd, bin, gen, zip, zsg, rar)\0*.smd;*.bin;*.gen;*.zip;*.zsg;*.rar\0" \
    "32X roms (32x, zip, rar)\0*.32x;*.zip;*.rar\0" \
    "Master System roms (sms, zip, rar)\0*.sms;*.zip;*.rar\0" \
    "Game Gear roms (gg, zip, rar)\0*.gg;*.zip;*.rar\0"\
    "Sega CD images (iso, bin, raw)\0*.iso;*.bin;*.raw\0" \
    "All files (*.*)\0*.*\0\0";
  ofn.nFilterIndex    = File_Type_Index;
  ofn.lpstrInitialDir = Rom_Dir;
  ofn.lpstrDefExt     = "smd";
  ofn.Flags           = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
  if(!GetOpenFileName(&ofn)) return 0;

  Free_Rom(Game);

  sys = Detect_Format(Name);

  File_Type_Index = ofn.nFilterIndex;
  if((File_Type_Index > 1) && (File_Type_Index < 7))
  {
    sys &= 1;
    sys |= (File_Type_Index - 1) << 1;
  }

  ext = strrchr(Name, '.');
  ret = 1;
  if(ext)
  {
    if(!strcmpi(".ZIP", ext) || !strcmpi(".ZSG", ext))
    { Game = Load_Rom_Zipped(hWnd, Name, sys);
      ret  = 0; }
    else if(!strcmpi(".RAR", ext))
    { Game = Load_Rom_RAR_Compressed(hWnd, Name, sys);
      ret  = 0; }
  }
  if(ret) { Game = Load_Rom(hWnd, Name, sys); }

  if(Game) Game->Type = sys;

  ret = -1;
  switch(sys >> 1)
  {
    default:
    case (GENESIS_ROM >> 1):    // Genesis rom
      allocate_Memstates(GENESIS_STATE_FILE_LENGHT); // Rewound
      if(Game) Genesis_Started = Init_Genesis(&Game->Genesis);
      ret = Genesis_Started;
      break;

    case (_32X_ROM >> 1):    // 32X rom
      allocate_Memstates(G32X_STATE_FILE_LENGHT); // Rewound
      if(Game) _32X_Started = Init_32X(&Game->Genesis);
      ret = _32X_Started;
      break;

    case (SMS_ROM >> 1):
    case (GG_ROM >> 1):
      if(memcmp(Rom_Data + 0x7FF0, "TMR SEGA", 8)) // invalid rom
      { MessageBox(ghWnd, "Invalid SMS/GG rom.", "Error", MB_ICONWARNING); 
        return 0; }
      allocate_Memstates(SMS_STATE_FILE_LENGHT); // Rewound
      if(Game) SMS_Started = SMS_Init(Game);
      ret = SMS_Started;
      break;

    case (SEGACD_IMAGE >> 1):    // Sega CD image
      allocate_Memstates(SEGACD_STATE_FILE_LENGHT); // Rewound
      SegaCD_Started = Init_SegaCD(Name);
      ret = SegaCD_Started;
      break;

    case (SEGACD_32X_IMAGE >> 1):    // Sega CD 32X image
      break;
  }

  // reading controller configuration (if necessary)
  if(Multiple_Controls_Config)
  {
    sprintf(Str_Tmp, "%s%s%s", Config_Dir, Rom_Name, ".ini");
    Load_Control_Config(Str_Tmp);
  }

  if(ret)
  {
    Update_Recent_Rom(Name, sys);
    Update_Rom_Dir(Name);
    Build_Main_Menu();
    if(!Full_Screen && Switch_Full_Screen) Set_Render(hWnd, 1, -1, Render_FS, TRUE);
  }

  return ret;
}


int __stdcall Pre_Load_Rom(HWND hWnd, char *Name)
{
  int    sys, ret;
  LPSTR  ext;

  SetCurrentDirectory(Gens_Path);

  if(GetFileAttributes(Name) & FILE_ATTRIBUTE_DIRECTORY) return 0;

  Free_Rom(Game);
  sys = Detect_Format(Name);

	if(sys < 1)
	{
		return -1;
	}

  if((sys >> 1) < 5)    // Have to load a rom
  {
    ret = 1;
    ext = strrchr(Name, '.');
    if(ext)
    {
      if(!strcmpi(".ZIP", ext) || !strcmpi(".ZSG", ext))
      {
				Game = Load_Rom_Zipped(hWnd, Name, sys);
        ret  = 0; }
      else if(!strcmpi(".RAR", ext))
			{
				Game = Load_Rom_RAR_Compressed(hWnd, Name, sys & 1);
        ret = 0;
			}
    }
    if(ret)
    {
			Game = Load_Rom(hWnd, Name, sys & 1); 
		}
  }

  if(Game) Game->Type = sys;

  ret = 0;
  switch(sys >> 1)
  {
    default:
    case (GENESIS_ROM >> 1):    // Genesis rom
      allocate_Memstates(GENESIS_STATE_FILE_LENGHT); //Rewound
	  if(Game) Genesis_Started = Init_Genesis(&Game->Genesis);
      ret = Genesis_Started;
      break;

    case (_32X_ROM >> 1):    // 32X rom
      allocate_Memstates(G32X_STATE_FILE_LENGHT); //Rewound
      if(Game) _32X_Started = Init_32X(&Game->Genesis);
      ret = _32X_Started;
      break;

    case (SMS_ROM >> 1):
    case (GG_ROM >> 1):
      if(memcmp(Rom_Data + 0x7FF0, "TMR SEGA", 8)) // invalid rom
      { MessageBox(ghWnd, "Invalid SMS/GG rom.", "Error", MB_ICONWARNING); 
        return 0; }
      allocate_Memstates(SMS_STATE_FILE_LENGHT); //Rewound
      if(Game) SMS_Started = SMS_Init(Game);
      ret = SMS_Started;
      break;

    case (SEGACD_IMAGE >> 1):    // Sega CD image
      allocate_Memstates(SEGACD_STATE_FILE_LENGHT); //Rewound
      SegaCD_Started = Init_SegaCD(Name);
      ret = SegaCD_Started;
      break;

    case (SEGACD_32X_IMAGE >> 1):    // Sega CD 32X image
      break;
  }

  // reading controller configuration (if necessary)
  if(Multiple_Controls_Config)
  {
    sprintf(Str_Tmp, "%s%s%s", Config_Dir, Rom_Name, ".ini");
    Load_Control_Config(Str_Tmp);
  }

  if(ret)
  {
    Update_Recent_Rom(Name, sys);
    Update_Rom_Dir(Name);
    Build_Main_Menu();
    if(!Full_Screen && Switch_Full_Screen) Set_Render(hWnd, 1, -1, Render_FS, TRUE);
  }

  return ret;
}


// Rom is already in buffer, we just need to fill rom structure
// and do init stuff...
int __stdcall Load_Rom_CC(LPCSTR Name, int Size)
{
  Rom_Info = (LPGROM_INFO)malloc(sizeof(GROM_INFO));

  if(!Rom_Info) { Game = NULL; return NULL; }

  Update_Rom_Name(Name);
  Rom_Size = Size;
  Fill_Infos_MD();

  Game = Rom_Info;

  Genesis_Started = Init_Genesis(&Game->Genesis);
  Build_Main_Menu();

  return Genesis_Started;
}


LPGROM_INFO __stdcall Load_Bios(HWND hWnd, char *Name)
{
  LPSTR  ext;

  SetCurrentDirectory(Gens_Path);

  if(GetFileAttributes(Name) & FILE_ATTRIBUTE_DIRECTORY) return NULL;

  Free_Rom(Game);

  ext = strrchr(Name, '.');
  if(ext)
  {
    if     (!strcmpi(".ZIP", ext)) return (Game = Load_Rom_Zipped(hWnd, Name, 0));
    else if(!strcmpi(".RAR", ext)) return (Game = Load_Rom_RAR_Compressed(hWnd, Name, 0));
    else                           return (Game = Load_Rom(hWnd, Name, 0));
  }
  else return NULL;
}


LPGROM_INFO __stdcall Load_Rom(HWND hWnd, char *Name, int Type)
{
  HANDLE Rom_File;
  BY_HANDLE_FILE_INFORMATION File_Inf;
  int Size = 0;
  int bResult;
  unsigned long Bytes_Read;

  SetCurrentDirectory(Gens_Path);

  Rom_File = CreateFile(Name, GENERIC_READ, FILE_SHARE_READ,
    NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

  if(Rom_File == INVALID_HANDLE_VALUE)
  { Game = NULL;
    return NULL; }

  if(GetFileInformationByHandle(Rom_File, &File_Inf))
    Size = (File_Inf.nFileSizeHigh << 16) + File_Inf.nFileSizeLow + 65536;

  if((Size - 65536) > ((6 * 1024 * 1024) + 512))
  { Game = NULL;
    CloseHandle(Rom_File);
    return NULL; }

  Rom_Info = (LPGROM_INFO)malloc(sizeof(GROM_INFO));

  if(!Rom_Info)
  { Game = NULL;
    CloseHandle(Rom_File);
    return NULL; }

  memset(Rom_Data, 0, 6 * 1024 * 1024);
  bResult = ReadFile(Rom_File, Rom_Data, Size, &Bytes_Read, NULL);

  CloseHandle(Rom_File);

  Rom_Size = Size - 65536;

  if(Type & 1) De_Interleave();

  Update_Rom_Name(Name);
  if     ((Type >> 1) < 3) Fill_Infos_MD();
  else if((Type >> 1) < 5) // SMS/GG rom
  {
    if((Rom_Size / 512) & 1) // skip image header
    { Rom_Size -= 512;
      memmove(Rom_Data, Rom_Data + 512, Rom_Size); }
    Fill_Infos_MS();
  }

  return Rom_Info;
}


LPGROM_INFO __stdcall Load_Rom_Zipped(HWND hWnd, LPCSTR Name, int Type)
{
  int           Size = 0;
  int           bResult;
  int           Current;
  char          Tmp[256];
  char          File_Name[132];
  UNZ_FILE_INFO Infos;
  HUNZFILE      Rom_File;
  LPSTR         ext;

  SetCurrentDirectory(Gens_Path);

  Rom_File = unzOpen(Name);
  if(!Rom_File)
  { Game = NULL; return NULL; }

  Current = unzGoToFirstFile(Rom_File);
  while(Current == UNZ_OK)
  {
    unzGetCurrentFileInfo(Rom_File, &Infos, File_Name, 128, NULL, 0, NULL, 0);
    ext = strrchr(File_Name, '.');
    if(ext)
    {
      if(!strcmpi(".SMD", ext) || !strcmpi(".BIN", ext) || !strcmpi(".GEN", ext) ||
         !strcmpi(".32X", ext) || !strcmpi(".SMS", ext) || !strcmpi(".GG", ext) || !strcmpi(".MD", ext))
      { Size = Infos.dwUncompSize; break; }
    }
    Current = unzGoToNextFile(Rom_File);
  }

  if(Current != UNZ_OK || !Size)
  {
    MessageBox(hWnd, "No valid roms file found in zip\n", "Error", MB_OK);
    unzClose(Rom_File);
    Game = NULL;
    return NULL;
  }

  if(Size > ((6 * 1024 * 1024) + 512))
  { unzClose(Rom_File);
    Game = NULL;
    return NULL; }

  if(unzOpenCurrentFile(Rom_File) != UNZ_OK)
  { unzClose(Rom_File);
    Game = NULL;
    return NULL; }

  Rom_Info = (LPGROM_INFO)malloc(sizeof(GROM_INFO));
  if(!Rom_Info)
  { unzClose(Rom_File);
    Game = NULL;
    return NULL; }

  memset(Rom_Data, 0, 6 * 1024 * 1024);
  bResult = unzReadCurrentFile(Rom_File, Rom_Data, Size);
  unzCloseCurrentFile(Rom_File);

  if((bResult <= 0) || (bResult != Size))
  {
    sprintf(Tmp, "Error in zip file : \n");

    switch(bResult)
    {
      case UNZ_ERRNO:
        strcat(Tmp, "Unknow...");
        break;
      case UNZ_EOF:
        strcat(Tmp, "Unexpected end of file.");
        break;
      case UNZ_PARAMERROR:
        strcat(Tmp, "Parameter error.");
        break;
      case UNZ_BADZIPFILE:
        strcat(Tmp, "Wrong zip file.");
        break;
      case UNZ_INTERNALERROR:
        strcat(Tmp, "Internal error.");
        break;
      case UNZ_CRCERROR:
        strcat(Tmp, "CRC error.");
        break;
    }

    MessageBox(hWnd, Tmp, "Error", MB_OK);
    unzClose(Rom_File);
    Game = NULL;
    return(NULL);
  }

  unzClose(Rom_File);
  Rom_Size = Size;

  if(Type & 1) De_Interleave();

  Update_Rom_Name(File_Name);
  if     ((Type >> 1) < 3) Fill_Infos_MD();
  else if((Type >> 1) < 5)
  {
    if((Rom_Size / 512) & 1) // skip image header
    { Rom_Size -= 512;
      memmove(Rom_Data, Rom_Data + 512, Rom_Size); }
    Fill_Infos_MS();
  }

  return Rom_Info;
}


typedef struct _RARDATA {
  HWND   hWnd;
  UINT   uSize;
  LPBYTE pData;
} RARDATA, *PRARDATA, *LPRARDATA;

int PASCAL RARProc(UINT mes, LPRARDATA data, LPVOID ptr, UINT size)
{
  switch(mes)
  {
    case UCM_PROCESSDATA:
      memcpy(data->pData + data->uSize, ptr, size);
      data->uSize += size;
      break;
    case UCM_NEEDPASSWORD:
      MessageBox(data->hWnd, "Encrypted file", "Error", MB_OK);
      break;
  }
  return 1;
}

LPGROM_INFO __stdcall Load_Rom_RAR_Compressed(HWND hWnd, char *Name, int Type)
{
  HANDLE             file;
  RAROpenArchiveData ropen;
  RARHeaderData      rhdr;
  BOOL               bret;
  LPSTR              ext;
  RARDATA            data;

  SetCurrentDirectory(Gens_Path);

  memset(&ropen, 0, sizeof(RAROpenArchiveData));
  ropen.ArcName  = Name;
  ropen.OpenMode = RAR_OM_EXTRACT;
  file = RAROpenArchive(&ropen);
  if(!file) return NULL;
  RARSetCallback(file, (UNRARCALLBACK)RARProc, (LONG)&data);
  bret = FALSE;
  memset(&rhdr, 0, sizeof(RARHeaderData));
  data.hWnd  = hWnd;
  data.pData = Rom_Data;
  data.uSize = 0;
  while(!RARReadHeader(file, &rhdr))
  {
    ext = strrchr(rhdr.FileName, '.');
    if(ext)
    {
      if(!strcmpi(".SMD", ext) || !strcmpi(".BIN", ext) || !strcmpi(".GEN", ext) ||
         !strcmpi(".32X", ext) || !strcmpi(".SMS", ext) || !strcmpi(".GG", ext))
      { bret = TRUE; break; }
    }
    if(RARProcessFile(file, RAR_SKIP, NULL, NULL)) break;
    memset(&rhdr, 0, sizeof(RARHeaderData));
  }
  if(bret)
  { memset(Rom_Data, 0, 6 * 1024 * 1024);
    if(RARProcessFile(file, RAR_EXTRACT, NULL, NULL)) { bret = FALSE; } }
  RARCloseArchive(file);
  if(bret)
  {
    Rom_Info = (LPGROM_INFO)malloc(sizeof(GROM_INFO));
    if(!Rom_Info) { Game = NULL; return NULL; }
    Rom_Size = data.uSize;
    if(Type & 1) De_Interleave();
    Update_Rom_Name(rhdr.FileName);
    if     ((Type >> 1) < 3) Fill_Infos_MD();
    else if((Type >> 1) < 5)
    {
      if((Rom_Size / 512) & 1) // skip image header
      { Rom_Size -= 512;
        memmove(Rom_Data, Rom_Data + 512, Rom_Size); }
      Fill_Infos_MS();
    }
    return Rom_Info;
  }
  else return NULL;
}


unsigned short Calculate_Checksum(void)
{
  unsigned short checksum = 0;
  unsigned int i;

  if(!Game) return(0);

  for (i = 512; i < Rom_Size; i += 2)
  {
    checksum += (unsigned short)(Rom_Data[i + 0]);
    checksum += (unsigned short)(Rom_Data[i + 1] << 8);
  }

  return checksum;
}


void Fix_Checksum(void)
{
  unsigned short checks;

  if(!Game) return;

  checks = Calculate_Checksum();

  if(Rom_Size)
  {
    Rom_Data[0x18E] = checks & 0xFF;
    Rom_Data[0x18F] = checks >> 8;
    _32X_Rom[0x18E] = checks >> 8;
    _32X_Rom[0x18F] = checks & 0xFF;
    if(Game) Game->Genesis.Checksum = (Rom_Data[0x18F] << 8) | Rom_Data[0x18E];
  }
}


unsigned int Calculate_CRC32(void)
{
  unsigned int crc = 0;

  Byte_Swap(Rom_Data, Rom_Size);
  crc = crc32(0, Rom_Data, Rom_Size);
  Byte_Swap(Rom_Data, Rom_Size);

  return crc;
}


int IPS_Patching(void)
{
  FILE *IPS_File;
  char Name[1024];
  unsigned char buf[16];
  unsigned int adr, len, i;

  SetCurrentDirectory(Gens_Path);

  strcpy(Name, IPS_Dir);
  strcat(Name, Rom_Name);
  strcat(Name, ".ips");

  IPS_File = fopen(Name, "rb");

  if(IPS_File == NULL) return 1;

  fseek(IPS_File, 0, SEEK_SET);

  fread(buf, 1, 5, IPS_File);
  buf[5] = 0;

  if(strcmpi((char *) buf, "patch"))
  {
    fclose(IPS_File);
    return 2;
  }

  fread(buf, 1, 3, IPS_File);
  buf[3] = 0;

  while (strcmpi((char *) buf, "eof"))
  {
    adr = (unsigned int) buf[2];
    adr += (unsigned int) (buf[1] << 8);
    adr += (unsigned int) (buf[0] << 16);

    if(fread(buf, 1, 2, IPS_File) == 0)
    {
      fclose(IPS_File);
      return 3;
    }

    len = (unsigned int) buf[1];
    len += (unsigned int) (buf[0] << 8);

    for(i = 0; i < len; i++)
    {
      if(fread(buf, 1, 1, IPS_File) == 0)
      {
        fclose(IPS_File);
        return 3;
      }

      if(adr < Rom_Size) Rom_Data[adr++] = buf[0];
    }

    if(fread(buf, 1, 3, IPS_File) == 0)
    {
      fclose(IPS_File);
      return 3;
    }

    buf[3] = 0;
  }

  fclose(IPS_File);

  return 0;
}


void __stdcall Free_Rom(LPGROM_INFO Rom_MD)
{
  if     (Movie_Status & MOVIE_PLAYING)   StopPlayMovie();
  else if(Movie_Status & MOVIE_RECORDING) StopRecordMovie();

  if(Game == NULL) return;

#ifdef CC_SUPPORT
  CC_Close();
#endif

  if(SegaCD_Started) Save_BRAM();
  if(SMS_Started)    SMS_End_Emulation();
  else               Save_SRAM();
  Save_Patch_File();
  if(WAV_Dumping) Stop_WAV_Dump();
  if(GYM_Dumping) Stop_GYM_Dump();
  if(VGM_File)    Stop_VGM_Dump();
  if(SGM_File)    Stop_SGM_Dump();
  if(SegaCD_Started) Stop_CD();
  Net_Play = 0;
  Genesis_Started = 0;
  _32X_Started    = 0;
  SegaCD_Started  = 0;
  SMS_Started     = 0;
  Game = NULL;

  if(Rom_MD)
  {
    free(Rom_MD);
    Rom_MD = NULL;
  }

  Reset_VDP();
  if(Intro_Style == 3) Init_Genesis_Bios();

  // saving controller configuration (if necessary)
  if(Multiple_Controls_Config)
  {
    FILE *fcfg;
    strcpy(Str_Tmp, Config_Dir);
    strcat(Str_Tmp, Rom_Name);
    strcat(Str_Tmp, ".ini");
    fcfg = fopen(Str_Tmp, "wt");
    if(fcfg)
    {
      Save_Control_Config(fcfg);
      fclose(fcfg);
    }
    // reading the default controller configuration
    sprintf(Str_Tmp, "%s%s", Gens_Path, "gens.ini");
    Load_Control_Config(Str_Tmp);
  }

  Set_Title(NULL, NULL);
  memset(Rom_Name, 0, sizeof(Rom_Name));

  //Rewound - free_Memstates
  free_Memstates();
}
