/* script.cpp - projeto gens+. Rodrigo Cardoso.
   Rotinas para ler arquivos de script em geral. */

#include "script.h"


_STRINGS :: _STRINGS(void)
{
  ptr  = NULL;
  nptr = nstr = 0;
}

_STRINGS :: ~_STRINGS(void)
{
  delete_all();
}

BOOL __fastcall _STRINGS :: add_string(LPCSTR str, DWORD value)
{
  LPVOID ptmp;
  if(nstr == nptr)
  { ptmp = _realloc(ptr, (nptr + 64) * sizeof(struct _str));
    if(!ptmp) return FALSE;
    ptr = (struct _str *)ptmp; }
  nptr += 64;
  ptmp = _alloc(strlen(str) + 1);
  if(!ptmp) return FALSE;
  strcpy((LPSTR)ptmp, str);
  ptr[nstr].str     = (LPSTR)ptmp;
  ptr[nstr++].value = value;
  return TRUE;
}

void __fastcall _STRINGS :: delete_all(void)
{
  for(int i = 0; i < nstr; i++) { _free(ptr[i].str); }
  _free(ptr);
  ptr  = NULL;
  nptr = nstr = 0;
}


_SBUFFER :: _SBUFFER(void)
{
  ptr = NULL;
  len = buf_size = 0;
}

_SBUFFER :: ~_SBUFFER(void)
{
  _free(ptr);
  ptr = NULL;
  len = buf_size = 0;
}


BOOL __fastcall _SBUFFER :: expand(int size)
{
  LPVOID pv;
  pv = _realloc(ptr, size + buf_size);
  if(!pv) return FALSE;
  ptr = (LPSTR)pv;
  buf_size += size;
  return TRUE;
}


BOOL __inline _SBUFFER :: put_char(int c)
{
  if(buf_size <= len + 1) { if(!expand(256)) return FALSE; }
  ptr[len++] = (char)c;
  ptr[len]   = 0;
  return TRUE;
}


void __inline _SBUFFER :: reset_content(void)
{
  if(ptr) memset(ptr, 0, buf_size);
  len = 0;
}



_SCRIPT :: _SCRIPT(void)
{
  error_code = SFE_NO_ERROR;
  pfile      = NULL;
  cur_line   = 1;
}

int __inline _SCRIPT :: get_next_char(DWORD skip)
{
  while((c = getc(pfile)) != EOF)
  {
    if(c == ';' && (skip & SKIP_COMMENTS)) // pulando comentrios
    { while((c = getc(pfile)) != '\n') { if(c == EOF) return c; } }
    if(c == '\n') cur_line++;
    if(is_hspace(c))                       // pulando espaos
    { if(!(skip & SKIP_HSPACES)) break; }
    else if(is_vspace(c))                  // pulando quebras de linha
    { if(!(skip & SKIP_VSPACES)) break; }
    else break;
  }
  return c;
}

void __inline _SCRIPT :: unget_char(void)
{
  if(c == '\n') cur_line--;
  ungetc(c, pfile);
}

/* copia o prxima valor para o buffer, se o valor lido for uma string (entre "" ou '')
   atribui TRUE  varivel is_buf_string, os caracteres apontados por sp_chars so
   considerados especiais e retornam sozinhos */
BOOL __fastcall _SCRIPT :: read_next_value(DWORD skip, LPCSTR sp_chars)
{
  int brk;

  buf.reset_content();
  // pulando os espaos antes do valor
  get_next_char(skip);

  is_value_string = FALSE;
  // fim do arquivo
  if     (c == EOF)  return FALSE;
  else if(c == '\n') return TRUE;

  // caracteres especiais retornam sozinhos
  else if(is_character(c, sp_chars))
  { buf.put_char(c); }

  else if(c == '\"' || c == '\'') // string entre " ou '
  {
    is_value_string = TRUE;

    brk = c;
    while(get_next_char(SKIP_NOTHING) != brk)
    {
      if     (c == '\n') { error_code = SFE_INVALID_LINE_BREAK;  return FALSE; }
      else if(c == EOF)  { error_code = SFE_INVALID_END_OF_FILE; return FALSE; }
      else if(c == '\\')               // caracteres especiais dentro da string
      {
        get_next_char(SKIP_NOTHING);
        if(c == EOF) { error_code = SFE_INVALID_END_OF_FILE; return FALSE; }
        else if(c == '0')  c = 0;
        else if(c == 't')  c = '\t';
        else if(c == 'n')  c = '\n';
        else if(c != '\\' && c != '\'' && c != '\"' || c != '?')
        { unget_char(); c = '\\'; }
      }
      if(!buf.put_char(c)) // erro ao alocar memria
      { error_code = SFE_NOT_ENOUGH_MEMORY; return FALSE; }
    }
  }

  else
  {
    while(c != EOF)
    {
      // caracteres que quebram o valor
      if(is_space(c) || c == '\"' || c == '\'' || is_character(c, sp_chars))
      { unget_char(); break; }
      else
      {
        if(!buf.put_char(c))
        { error_code = SFE_NOT_ENOUGH_MEMORY; return FALSE; }
      }
      get_next_char(skip & SKIP_COMMENTS);
    }
  }

  return TRUE;
}


// retira caracteres de espao no incio e no fim da string
int __stdcall Clear_String_Spaces(LPSTR lpString)
{
  int i, iSize, ch, s;

  // tamanho da string
  iSize = strlen(lpString);
  // retirando espaos do fim da string
  for(i = iSize - 1; i > 0; i--)
  {
    ch = (int)lpString[i];
    if(is_space(ch)) lpString[i] = 0;
    else            break;
  }
  // novo tamanho da string
  iSize = i + 1;
  // contando o nmero de espaos no incio da string
  s = 0;
  for(i = 0; i < iSize; i++)
  {
    ch = (int)lpString[i];
    if(is_space(ch)) s++;
    else            break;
  }
  // movendo a string na memria para retirar os espaos no incio
  if(s)
  {
    // tamanho da string sem o 0x00 no final
    iSize -= s;
    MoveMemory(lpString, lpString + s, iSize);
    // colocando o 0x00 no fim
    lpString[iSize] = 0;
  }

  return iSize;
}
