/* script.h - projeto gens+. Rodrigo Cardoso.
   Rotinas para ler arquivos de script em geral. */

#ifndef _SCRIPT_H
#define _SCRIPT_H

#include <windows.h>
#include <stdio.h>

#define is_space(c)  ((c >= 0x09 && c <= 0x0D) || (c == 0x20))
#define is_hspace(c) (c == 0x09 || c == 0x0B || c == 0x0C || c == 0x20)
#define is_vspace(c) (c == 0x0A || c == 0x0D)
#define is_digit(c)  (c >= '0' && c <= '9')
BOOL __inline is_character(int c, LPCSTR characters)
{ int ch, i = 0;
  while((ch = characters[i++])) { if(c == ch) return TRUE; }
  return FALSE; }
extern "C" int __fastcall Str2Int(LPCSTR lpString, LPDWORD pResult);

int __stdcall Clear_String_Spaces(LPSTR);

/*
// funes em assembly
extern "C" BOOL __fastcall is_space (int c);
extern "C" BOOL __fastcall is_hspace(int c);
extern "C" BOOL __fastcall is_vspace(int c);
extern "C" BOOL __fastcall is_digit (int c);
extern "C" BOOL __fastcall is_character(int c, LPCSTR characters);
*/

// macros usadas para alocar memria
#define _alloc(size) malloc(size)
#define _realloc(ptr, size) ( ptr ? realloc(ptr, size) : malloc(size) )
#define _free(ptr) { if(ptr) { free(ptr); ptr = NULL; } }

// buffer que armazena diversas strings
typedef struct _STRINGS {
  _STRINGS(void);
  ~_STRINGS(void);
  int             nptr;                      // nmero de ponteiros alocados
  int             nstr;                      // nmero total de strings no buffer
  struct _str {
    LPSTR          str;                      // string
    DWORD          value;                    // valor relacionado
  } *ptr;
  BOOL __fastcall add_string(LPCSTR, DWORD); // adiciona a string ao buffer
  void __fastcall delete_all(void);
} STRINGS;


// buffer usado nas operaes com arquivos de script
typedef struct _SBUFFER {
  _SBUFFER (void);
  ~_SBUFFER(void);
  LPSTR           ptr;                     // aponta para o bloco de memria usado
  int             buf_size, len;           // tamanho do buffer e o tanto que est sendo usado
  DWORD           num;
  int             is_num_hex;              // TRUE se o nmero est no formato hexadecimal
  BOOL __fastcall expand(int);             // expande a memria
  BOOL __inline   put_char(int);           // insere o caracter no buffer (se for preciso expande)
  void __inline   reset_content(void);
  BOOL __inline   convert_to_number(void)  // converte a string em ptr para um nmero em num
  {
    int res;
    res = Str2Int(ptr, &num);
    return ((res & 0x00FF) && !(res & 0xFF00));
  }
} SBUFFER;


#define SKIP_NOTHING  0x00
#define SKIP_COMMENTS 0x01
#define SKIP_HSPACES  0x02
#define SKIP_VSPACES  0x04
#define SKIP_SPACES   (SKIP_HSPACES | SKIP_VSPACES)

// cdigos de erro relacionados ao arquivo de script
#define SFE_NO_ERROR            0x00      // sem erros
#define SFE_NOT_ENOUGH_MEMORY   0x01      // memria insuficinte
#define SFE_INVALID_END_OF_FILE 0x02      // fim de arquivo invlido
#define SFE_INVALID_LINE_BREAK  0x03      // quebra de linha invlida
#define SFE_INVALID_NUMBER      0x04      // nmero invlido

// define um arquivo de script
typedef struct _SCRIPT {
  _SCRIPT(void);
  DWORD           error_code;             // cdigo de erro
  SBUFFER         buf;                    // buffer usado para armazenar resultados parciais de leitura
  BOOL            is_value_string;        // TRUE se o valor lido  uma string (entre "" ou '')
  FILE           *pfile;                  // aponta para o arquivo sendo lido (aberto com acesso "rt")
  int             c;                      // ltimo caracter lido
  int             cur_line;               // linha atual sendo lida
  int  __inline   get_next_char(DWORD);   // retorna o prximo caracter no arquivo e copia para c
  void __inline   unget_char(void);
  BOOL __fastcall read_next_value(DWORD, LPCSTR);

  BOOL __inline   read_next_number(DWORD skip, LPCSTR sp_chars)
  {
    if(!read_next_value(skip, sp_chars)) return FALSE;
    if(is_value_string)          { error_code = SFE_INVALID_NUMBER; return FALSE; }
    if(!buf.convert_to_number()) { error_code = SFE_INVALID_NUMBER; return FALSE; }
    return TRUE;
  }
  BOOL __inline   is_value_char(int ch)
  {
    return (buf.len == 1 && (*buf.ptr) == ch && !is_value_string);
  }
  BOOL __inline   is_value_mnemonic(LPCSTR mn)
  {
    return (!strcmpi(buf.ptr, mn) && !is_value_string);
  }

} SCRIPT;


#endif // _SCRIPT_H
