; scripta.asm - projeto gens+. Rodrigo Cardoso
; contm funes usadas para ler arquivos de script

%macro DECL 1-2
  %if %0 > 1
    global @%1@%2
  %else
    global _%1
  %endif
    global %1
  %if %0 > 1
    @%1@%2:
  %else
    _%1:
  %endif
    %1:
%endmacro

section .bss  algin=64

  num_end:   resd 1
  error:     resb 1
  negative:  resb 1


section .text align=64

; --------------------------------------------------------------------------------------
; int __fastcall Str2Int(LPCSTR lpString, LPDWORD pResult)
; extrai nmero inteiro dentro da string, pode retornar os seguintes valores
; 0x0000 - erro: a string possui caracteres invlidos (*pResult permanece inalterado),
;          no foi possvel ler o nmero.
; 0xXX01 - nmero lido no formato decimal sem o sinal de menos.
; 0xXX03 - nmero lido no formato decimal com sinal de menos.
; 0xXX04 - nmero lido no formato formato hexadecimal (0x..).
; 0x00XX - o nmero foi lido e no havia erros na string.
; 0x01XX - foi possvel ler o nmero no formato, mas a string possui caracteres invlidos
;          depois do nmero.
; 0x02XX - o valor ultrapassa 32 bits, somente uma parte do valor foi lida.
DECL Str2Int, 8

; pula espaos na string
%macro SKIP_SPACES 0
  %%loop:
    mov   al, [ecx]
    test  al, al
    jz    near .end
    inc   ecx
    cmp   al, 0x20
    jz    %%loop
    cmp   al, 0x09
    js    %%end_loop
    cmp   al, 0x0E
    js    %%loop
    %%end_loop:
%endmacro

%macro VALIDATE_STRING 0
  mov   ecx, [num_end]
  %%loop:
    mov   al, [ecx]
    test  al, al
    jz    %%end_loop
    inc   ecx
    cmp   al, 0x20
    jz    %%loop
    cmp   al, 0x09
    js    %%no_space
    cmp   al, 0x0E
    js    %%loop
    %%no_space:
    or    byte [error], 0x01
    %%end_loop:
%endmacro

; verifica se o valor em al  um dgito hexadecimal vlido se no for, pula para o ponto
; em %1
%macro IS_HEX_DIGIT 1
  cmp     al, '0'     ; valor mnimo vlido
  js      %1
  cmp     al, 'f' + 1 ; valor mximo vlido
  jns     %1
  cmp     al, '9' + 1 ; intervalo entre '0' e '9'
  js      %%ok
  cmp     al, 'a'     ; intervalo entre 'a' e 'f'
  jns     %%ok
  cmp     al, 'A'     ; intervalo entre 'A' e 'F'
  js      %1
  cmp     al, 'F' + 1
  jns     %1
  %%ok:
%endmacro

; se o valor em al no for dgito decimal, pula para %1
%macro IS_DIGIT 1
  cmp     al, '0'
  js      %1
  cmp     al, '9' + 1
  jns     %1
%endmacro

  xor   eax, eax
  mov   [error], al
  mov   [negative], al
  test  ecx, ecx
  jz    near .end

  ; pulando espaos no incio da string
  SKIP_SPACES

  ; verificando o tipo de nmero  ser lido (hexadecimal ou decimal)
  cmp   al, '-'
  jz    near .read_signed
  cmp   al, '+'
  jz    near .read_unsig
  cmp   al, '0'
  jnz   near .read_unsig2
  mov   al, [ecx]
  inc   ecx
  cmp   al, 'X'
  jz    .read_hex
  cmp   al, 'x'
  jz    .read_hex
  test  al, al
  jnz   near .read_unsig2
  mov   [edx], eax
  mov   eax, 1
  ret


  ; *** lendo nmero no formato hexadecimal ***
  .read_hex:
  mov   al, [ecx]
  IS_HEX_DIGIT near .error
  ; o ltimo dgito do nmero est em al e [ecx]

  push  ebx
  push  ebp

  ; buscando o primeiro dgito do nmero
  xor   ebx, ebx
  mov   [edx], ebx
  .loop_hex_1:
    inc   ebx
    inc   ecx
    mov   al, [ecx]
    IS_HEX_DIGIT .end_loop_hex_1
    jmp   .loop_hex_1
    .end_loop_hex_1:

  mov   [num_end], ecx
  mov   ebp, 1
  ; mximo de 8 digitos para o nmero
  cmp   ebx, 8 + 1
  js    .loop_hex_2
  mov   ebx, 8
  or    byte [error], 0x02
  .loop_hex_2:
    dec  ecx
    xor  eax, eax
    mov  al, [ecx]
    cmp  al, '9' + 1
    jns  .cmp_hex_1
    sub  al, '0'
    jmp  .end_cmp_hex
    .cmp_hex_1:
    cmp  al, 'F' + 1
    jns  .cmp_hex_2
    sub  al, 'A' - 10
    jmp  .end_cmp_hex
    .cmp_hex_2:
    sub  al, 'a' - 10
    .end_cmp_hex:
    imul eax, ebp
    or   [edx], eax
    shl  ebp, 4
    dec  ebx
    jnz  .loop_hex_2

  ; nmero lido, liberando os registradores usados
  pop   ebp
  pop   ebx
  
  ; verificando se h caracteres invlidos no restante da string
  VALIDATE_STRING

  xor   eax, eax
  mov   al, 0x04
  mov   ah, [error]
  jmp   .end
  


  ; *** lendo nmero no formato decimal ***
  .read_signed:
  mov   byte [negative], 0x01  
  .read_unsig:
  SKIP_SPACES            ; pulando espaos entre o sinal de + e o nmero
  .read_unsig2:
  IS_DIGIT near .error
  dec   ecx
  ; o ltimo dgito do nmero est em al e [ecx]
  
  push  ebx
  push  ebp

  ; buscando o primeiro dgito do nmero
  xor   ebx, ebx
  mov   [edx], ebx
  .loop_dec_1:
    inc   ebx
    inc   ecx
    mov   al, [ecx]
    IS_DIGIT .end_loop_dec_1
    jmp   .loop_dec_1
    .end_loop_dec_1:

  mov   [num_end], ecx
  mov   ebp, 1
  ; mximo de 10 digitos para o nmero
  cmp   ebx, 9 + 1      ; ARRUMAR!!!
  js    .loop_dec_2
  mov   ebx, 9
  or    byte [error], 0x02
  .loop_dec_2:
    dec  ecx
    xor  eax, eax
    mov  al, [ecx]
    sub  al, '0'
    imul eax, ebp
    add  [edx], eax
    imul ebp, 10
    dec  ebx
    jnz  .loop_dec_2

  ; nmero lido, liberando os registradores usados
  pop   ebp
  pop   ebx
  
  ; verificando se h caracteres invlidos no restante da string
  VALIDATE_STRING

  xor   eax, eax
  mov   al, 0x01
  test  byte [negative], 0x01
  jz    .no_neg
  or    al, 0x02
  neg   dword [edx]
  .no_neg:
  mov   ah, [error]
  jmp   .end


  .error:
  xor   al, al
  .end:
  ret
; --------------------------------------------------------------------------------------


; --------------------------------------------------------------------------------------
; BOOL __fastcall is_space(int c)
; se o caracter for espao retorna TRUE
DECL is_space, 4
  cmp   ecx, 0x20
  jz    .ok
  xor   eax, eax
  cmp   ecx, 0x09
  js    .end
  cmp   ecx, 0x0E
  jns   .end
  .ok:
  mov   eax, 0x01
  .end:
  ret
; --------------------------------------------------------------------------------------


; --------------------------------------------------------------------------------------
; BOOL __fastcall is_hspace(int c)
; se o caracter for espao horizontal retorna TRUE
DECL is_hspace, 4
  cmp   ecx, 0x09
  jz    .ok
  cmp   ecx, 0x0B
  jz    .ok
  cmp   ecx, 0x0C
  jz    .ok
  cmp   ecx, 0x20
  jz    .ok
  xor   eax, eax
  jmp   .end
  .ok:
  mov   eax, 0x01
  .end:
  ret
; --------------------------------------------------------------------------------------


; --------------------------------------------------------------------------------------
; BOOL __fastcall is_vspace(int c)
; se o caracter for espao vertical retorna TRUE
DECL is_vspace, 4
  cmp   ecx, 0x0A
  jz    .ok
  cmp   ecx, 0x0D
  jz    .ok
  xor   eax, eax
  jmp   .end
  .ok:
  mov   eax, 0x01
  .end:
  ret
; --------------------------------------------------------------------------------------


; --------------------------------------------------------------------------------------
; BOOL __fastcall is_digit(int c)
; se o caracter for um dgito (0 a 9) retorna TRUE
DECL is_digit, 4
  xor   eax, eax
  cmp   ecx, 0x30
  js    .end
  cmp   ecx, 0x40
  jns   .end
  mov   eax, 0x01
  .end:
  ret
; --------------------------------------------------------------------------------------


; --------------------------------------------------------------------------------------
; BOOL __fastcall is_character(int c, LPCSTR characters)
; verifica se o caracter c  um dos da lista apontada em characters
DECL is_character, 8
  xor  eax, eax
  .loop:
  cmp  al, [edx]
  jz   .end
  cmp  cl, [edx]
  jz   .ok
  inc  edx
  jmp  .loop
  .ok:
  mov  eax, 0x01
  .end:
  ret
; --------------------------------------------------------------------------------------
