
#include "sms.h"
#include "vgm.h"

void __fastcall SMS_CPU_Init(void)
{
#ifdef ASMZ80
  sz80_Init(&M_SZ80);
#endif
  SMS_Reset();
}

void __fastcall OutZ80(WORD port, BYTE data)
{
  switch(port & 0xFF)
  {
    case 0x01: // GG SIO 
    case 0x02:
    case 0x03:
    case 0x04:
    case 0x05:
      break;

    case 0x06: // GG STEREO 
      sms.psg_mask = (data & 0xFF);
      if(VGM_File) Update_VGM_Dump(0x4F, 0, data);
      break;

    case 0x7E: // SN76489 PSG 
    case 0x7F:
      if(snd.init_flag) SN76496Write(data);
      if(VGM_File) Update_VGM_Dump(0x50, 0, data);
      break;

    case 0xBE: // VDP DATA 
      vdp_data_w(data);
      break;

    case 0xBD: // VDP CTRL  
    case 0xBF:
      vdp_ctrl_w(data);
      break;

    case 0xF0: // YM2413 
      if(snd.init_flag) sms.ym_cur_reg = data;
      break;

    case 0xF1:
      if(snd.init_flag) OPLL_writeReg(opll, sms.ym_cur_reg, data);
      if(VGM_File) Update_VGM_Dump(0x51, sms.ym_cur_reg, data);
      break;

    case 0xF2: // YM2413 DETECT 
      if(snd.init_flag) sms.port_F2 = (data & 1);
      break;

    case 0x3F:
      sms.port_3F = ((data & 0x80) | (data & 0x20) << 1) & 0xC0;
      if(sms.country == TYPE_DOMESTIC) sms.port_3F ^= 0xC0;
      break;

    case 0x3E: // memory enables (not suported yet)
      break;
  }
}


BYTE __fastcall InZ80(WORD port)
{
  BYTE temp = 0xFF;

  switch(port & 0xFF)
  {
    case 0x01: // GG SIO 
    case 0x02:
    case 0x03:
    case 0x04:
    case 0x05:
      return 0x00;

    case 0x7E: // V COUNTER 
      return vdp_vcounter_r();

    case 0x7F: // H COUNTER 
      return vdp_hcounter_r();

    case 0x00: // INPUT #2 
      temp = 0xFF;
      if(Control[0].Start == 0)        temp &= ~0x80;
      if(sms.country == TYPE_DOMESTIC) temp &= ~0x40;
      return temp;

    case 0xC0: // INPUT #0   
    case 0xDC:
      temp = 0xFF;
      if(Control[0].Up     == 0) temp &= ~0x01;
      if(Control[0].Down   == 0) temp &= ~0x02;
      if(Control[0].Left   == 0) temp &= ~0x04;
      if(Control[0].Right  == 0) temp &= ~0x08;
      if(Control[0].A      == 0) temp &= ~0x10;
      if(Control[0].B      == 0) temp &= ~0x20;
      if(Control[1].Up     == 0) temp &= ~0x40;
      if(Control[1].Down   == 0) temp &= ~0x80;
      return temp;

    case 0xC1: // INPUT #1 
    case 0xDD:
      temp = 0xFF;
      if(Control[1].Left   == 0) temp &= ~0x01;
      if(Control[1].Right  == 0) temp &= ~0x02;
      if(Control[1].A      == 0) temp &= ~0x04;
      if(Control[1].B      == 0) temp &= ~0x08;
      if(Control[0].Mode   == 0) temp &= ~0x10;
      return (temp & 0x3F) | (sms.port_3F & 0xC0);

    case 0xBE: // VDP DATA 
      return vdp_data_r();

    case 0xBD:
    case 0xBF: // VDP CTRL 
      return vdp_ctrl_r();

    // this is incorrect
    case 0xF2: // YM2413 DETECT 
      if(snd.init_flag) return 0xFF;
      else              return 0x00;
      break;
  }

  return 0xFF;
}

void __fastcall SMS_Map_Mem(WORD address, BYTE data)
{
  // calculate ROM page index 
  BYTE page = (data % SMS_Rom_Pages);

  switch(address)
  {
    case 0:
      if(data & 0x08)
      {
        // Page in ROM 
        z80_read_map[4]  = &SRAM[(data & 4) ? 0x4000 : 0x0000];
        z80_read_map[5]  = &SRAM[(data & 4) ? 0x6000 : 0x2000];
        z80_write_map[4] = &SRAM[(data & 4) ? 0x4000 : 0x0000];
        z80_write_map[5] = &SRAM[(data & 4) ? 0x6000 : 0x2000];
#ifdef ASMZ80
        sz80_Add_Fetch(&M_SZ80, 0x80, 0xBF, z80_read_map[4]);
#endif
      }
      else
      {
        // Page in RAM 
        z80_read_map[4]  = &Rom_Data[((sms.fcr[3] % SMS_Rom_Pages) << 14) + 0x0000];
        z80_read_map[5]  = &Rom_Data[((sms.fcr[3] % SMS_Rom_Pages) << 14) + 0x2000];
        z80_write_map[4] = SMS_Dummy;
        z80_write_map[5] = SMS_Dummy;
#ifdef ASMZ80
        sz80_Add_Fetch(&M_SZ80, 0x80, 0xBF, z80_read_map[4]);
#endif
      }
      break;

    case 1:
      z80_read_map[0] = &Rom_Data[(page << 14) + 0x0000];
      z80_read_map[1] = &Rom_Data[(page << 14) + 0x2000];
#ifdef ASMZ80
      sz80_Add_Fetch(&M_SZ80, 0x00, 0x3F, z80_read_map[0]);
#endif
      break;

    case 2:
      z80_read_map[2] = &Rom_Data[(page << 14) + 0x0000];
      z80_read_map[3] = &Rom_Data[(page << 14) + 0x2000];
#ifdef ASMZ80
      sz80_Add_Fetch(&M_SZ80, 0x40, 0x7F, z80_read_map[2]);
#endif
      break;

    case 3:
      if(!(sms.fcr[0] & 0x08))
      {
        z80_read_map[4] = &Rom_Data[(page << 14) + 0x0000];
        z80_read_map[5] = &Rom_Data[(page << 14) + 0x2000];
#ifdef ASMZ80
        sz80_Add_Fetch(&M_SZ80, 0x80, 0xBF, z80_read_map[4]);
#endif
      }
      break;
  }

  // save frame control register data 
  sms.fcr[address] = data;
}

#ifdef ASMZ80

void __fastcall SMS_Map_Mem_W(WORD address, WORD data)
{
  if     (address == 0xFFFB) SMS_Map_Mem(0xFFFC & 3, HIBYTE(data));
  else if(address == 0xFFFF) SMS_Map_Mem(0xFFFF & 3, LOBYTE(data));
  else
  {
//    SMS_Map_Mem(address++ & 3, LOBYTE(data));
//    SMS_Map_Mem(address & 3, HIBYTE(data));
  }
}

#endif
