/**********************************************************/
/*                                                        */
/* Z80 emulator 0.99                                      */
/* Copyright 2002 Stphane Dallongeville                  */
/* Used for the genesis emulation in Gens                 */
/*                                                        */
/* Gens+ SMS emulation by Rodrigo Cardoso                 */
/**********************************************************/

#ifndef _SZ80_H_
#define _SZ80_H_

#include <windows.h>

#ifdef __cplusplus
extern "C" {
#endif


/******************************/
/* Compiler dependant defines */
/******************************/

#ifndef FASTCALL
#define FASTCALL __fastcall
#endif


/****************************/
/* Structures & definitions */
/****************************/


typedef struct _SZ80_context
{
  union
  {
    struct
    {
      BYTE A;
      BYTE F;
      BYTE x;
      BYTE FXY;
    } b;
    struct
    {
      WORD AF;
      WORD FXYW;
    } w;
    DWORD d;
  } AF;
  union
  {
    struct
    {
      BYTE C;
      BYTE B;
      WORD x;
    } b;
    struct
    {
      WORD BC;
      WORD x;
    } w;
    DWORD d;
  } BC;
  union
  {
    struct
    {
      BYTE E;
      BYTE D;
      WORD x;
    } b;
    struct
    {
      WORD DE;
      WORD x;
    } w;
    DWORD d;
  } DE;
  union
  {
    struct
    {
      BYTE L;
      BYTE H;
      WORD x;
    } b;
    struct
    {
      WORD HL;
      WORD x;
    } w;
    DWORD d;
  } HL;
  union
  {
    struct
    {
      BYTE IXL;
      BYTE IXH;
      WORD x;
    } b;
    struct
    {
      WORD IX;
      WORD x;
    } w;
    DWORD d;
  } IX;
  union
  {
    struct
    {
      BYTE IYL;
      BYTE IYH;
      WORD x;
    } b;
    struct
    {
      WORD IY;
      WORD x;
    } w;
    DWORD d;
  } IY;
  union
  {
    struct
    {
      BYTE PCL;
      BYTE PCH;
      WORD x;
    } b;
    struct
    {
      WORD PC;
      WORD x;
    } w;
    DWORD d;
  } PC;
  union
  {
    struct
    {
      BYTE SPL;
      BYTE SPH;
      WORD x;
    } b;
    struct
    {
      WORD SP;
      WORD x;
    } w;
    DWORD d;
  } SP;
  union
  {
    struct
    {
      BYTE A2;
      BYTE F2;
      BYTE x;
      BYTE FXY2;
    } b;
    struct
    {
      WORD AF2;
      WORD FXYW2;
    } w;
    DWORD d;
  } AF2;
  union
  {
    struct
    {
      BYTE C2;
      BYTE B2;
      WORD x;
    } b;
    struct
    {
      WORD BC2;
      WORD x;
    } w;
    DWORD d;
  } BC2;
  union
  {
    struct
    {
      BYTE E2;
      BYTE D2;
      WORD x;
    } b;
    struct
    {
      WORD DE2;
      WORD x;
    } w;
    DWORD d;
  } DE2;
  union
  {
    struct
    {
      BYTE L2;
      BYTE H2;
      WORD x;
    } b;
    struct
    {
      WORD HL2;
      WORD x;
    } w;
    DWORD d;
  } HL2;
  union
  {
    struct
    {
      BYTE IFF1;
      BYTE IFF2;
      WORD x;
    } b;
    struct
    {
      WORD IFF;
      WORD x;
    } w;
    DWORD d;
  } IFF;
  union
  {
    struct
    {
      BYTE R1;
      BYTE R2;
      WORD x;
    } b;
    struct
    {
      WORD R;
      WORD x;
    } w;
    DWORD d;
  } R;

  BYTE I;
  BYTE IM;
  BYTE IntVect;
  BYTE IntLine;

  DWORD Status;
  DWORD BasePC;
  DWORD TmpSav0;
  DWORD TmpSav1;

  DWORD CycleCnt;
  DWORD CycleTD;
  DWORD CycleIO;
  DWORD CycleSup;

  BYTE *Fetch[0x100];

  DWORD RetIC;
  DWORD IntAckC;
} SZ80_CONTEXT;


/*************************/
/* Publics Z80 variables */
/*************************/


extern SZ80_CONTEXT M_SZ80;
extern LPBYTE z80_read_map[8];
extern LPBYTE z80_write_map[8];



/*************************/
/* Publics Z80 functions */
/*************************/


DWORD FASTCALL sz80_Init(SZ80_CONTEXT *sz80);
DWORD FASTCALL sz80_Reset(SZ80_CONTEXT *sz80);

DWORD FASTCALL sz80_Read_Odo(SZ80_CONTEXT *sz80);
void FASTCALL sz80_Clear_Odo(SZ80_CONTEXT *sz80);
void FASTCALL sz80_Set_Odo(SZ80_CONTEXT *sz80, DWORD Odo);
void FASTCALL sz80_Add_Cycles(SZ80_CONTEXT *sz80, DWORD cycles);

DWORD FASTCALL sz80_Exec(SZ80_CONTEXT *sz80, int odo);

DWORD FASTCALL sz80_NMI(SZ80_CONTEXT *sz80);
DWORD FASTCALL sz80_Interrupt(SZ80_CONTEXT *sz80, DWORD vector);

DWORD sz80_Add_Fetch(SZ80_CONTEXT *z80, DWORD low_adr, DWORD high_adr, BYTE *Region);

DWORD FASTCALL sz80_Get_PC(SZ80_CONTEXT *sz80);
DWORD FASTCALL sz80_Get_AF(SZ80_CONTEXT *sz80);
DWORD FASTCALL sz80_Get_AF2(SZ80_CONTEXT *sz80);
DWORD FASTCALL sz80_Set_PC(SZ80_CONTEXT *sz80, DWORD PC);
DWORD FASTCALL sz80_Set_AF(SZ80_CONTEXT *sz80, DWORD AF);
DWORD FASTCALL sz80_Set_AF2(SZ80_CONTEXT *sz80, DWORD AF2);


#ifdef __cplusplus
};
#endif

#endif