
#include "sms.h"
#include "vgm.h"
#include "movie.h"

SMS_CONTEXT sms;

// executa a emulao do console para um frame
int __fastcall Do_SMS_Frame(void)
{
  int i;

#ifdef ASMZ80
  int cycles;
  sz80_Clear_Odo(&M_SZ80);
  cycles = 0;
#endif

  if(Control[0].Start == 0)
  {
    if(!sms.paused)
    {
      sms.paused = 1;
#ifdef ASMZ80
      sz80_NMI(&M_SZ80);
#else
      IntZ80(&M_SZ80, INT_NMI);
#endif
    }
  }
  else sms.paused = 0;

  for(SMS_VDP.line = 0; SMS_VDP.line < 262; SMS_VDP.line++)
  {
    // executa possveis interrupes geradas pelo VDP
    SMS_VDP_Run();
    // imprime o quadro atual
    render_line(SMS_VDP.line);
    // executa o Z80 para esta linha
#ifdef ASMZ80
    cycles += 227;
    sz80_Exec(&M_SZ80, cycles);
#else
    RunZ80(&M_SZ80);
#endif
  }

  // update de som
  if(Sound_Enable && snd.init_flag)
  {
    SN76496Update(snd.psg_buffer, snd.bufsize, 0xFF);
    if(SMS_YM2413_Enable)
    {
      for(i = 0; i < snd.bufsize; i++)
      { snd.buffer[0][i] = snd.buffer[1][i] = (int)OPLL_calc(opll); }
    }
    for(i = 0; i < snd.bufsize; i++)
    {
      snd.buffer[0][i] += (int)snd.psg_buffer[0][i];
      snd.buffer[1][i] += (int)snd.psg_buffer[1][i];
    }
  }

  if(VGM_File) Update_VGM_Dump(0x62, 0, 0);

  return 1;
}


// idem a funo anterior, s que no executa o render
int __fastcall Do_SMS_Frame_Fast(void)
{
  int i;

#ifdef ASMZ80
  int cycles;
  sz80_Clear_Odo(&M_SZ80);
  cycles = 0;
#endif

  if(Control[0].Start == 0)
  {
    if(!sms.paused)
    {
      sms.paused = 1;
#ifdef ASMZ80
      sz80_NMI(&M_SZ80);
#else
      IntZ80(&M_SZ80, INT_NMI);
#endif
    }
  }
  else sms.paused = 0;

  for(SMS_VDP.line = 0; SMS_VDP.line < 262; SMS_VDP.line++)
  {
    // executa possveis interrupes geradas pelo VDP
    SMS_VDP_Run();
    // executa o Z80 para esta linha
#ifdef ASMZ80
    cycles += 227;
    sz80_Exec(&M_SZ80, cycles);
#else
    RunZ80(&M_SZ80);
#endif
  }

  // update de som
  if(Sound_Enable && snd.init_flag)
  {
    SN76496Update(snd.psg_buffer, snd.bufsize, 0xFF);
    if(SMS_YM2413_Enable)
    {
      for(i = 0; i < snd.bufsize; i++)
      { snd.buffer[0][i] = snd.buffer[1][i] = (int)OPLL_calc(opll); }
    }
    for(i = 0; i < snd.bufsize; i++)
    {
      snd.buffer[0][i] += (int)snd.psg_buffer[0][i];
      snd.buffer[1][i] += (int)snd.psg_buffer[1][i];
    }
  }

  if(VGM_File) Update_VGM_Dump(0x62, 0, 0);

  return 1;
}


// reseta o master (reseta MESMO)
void __fastcall SMS_Reset(void)
{
  if     (Movie_Status & MOVIE_PLAYING)   StopPlayMovie();
  else if(Movie_Status & MOVIE_RECORDING) StopRecordMovie();

  memset(SMS_Dummy, 0, 0x2000);
  memset(Ram_Z80, 0, 0x2000);
  sms.port_3F = sms.port_F2 = 0x00;
  sms.psg_mask   = 0xFF;
  sms.ym_cur_reg = 0x00;

  // mapeando a memria com os valores padres
  z80_read_map[0] = Rom_Data + 0x0000; // 0x0000 : 0x1FFF
  z80_read_map[1] = Rom_Data + 0x2000; // 0x2000 : 0x3FFF
  z80_read_map[2] = Rom_Data + 0x4000; // 0x4000 : 0x5FFF
  z80_read_map[3] = Rom_Data + 0x6000; // 0x6000 : 0x7FFF
  z80_read_map[4] = Rom_Data + 0x8000; // 0x8000 : 0x9FFF
  z80_read_map[5] = Rom_Data + 0xA000; // 0xA000 : 0xBFFF
  z80_read_map[6] = Ram_Z80;           // 0xC000 : 0xDFFF 
  z80_read_map[7] = Ram_Z80;           // 0xE000 : 0xFFFF

  z80_write_map[0] = SMS_Dummy;        // 0x0000 : 0x1FFF
  z80_write_map[1] = SMS_Dummy;        // 0x2000 : 0x3FFF
  z80_write_map[2] = SMS_Dummy;        // 0x4000 : 0x5FFF
  z80_write_map[3] = SMS_Dummy;        // 0x6000 : 0x7FFF
  z80_write_map[4] = SMS_Dummy;        // 0x8000 : 0x9FFF
  z80_write_map[5] = SMS_Dummy;        // 0xA000 : 0xBFFF
  z80_write_map[6] = Ram_Z80;          // 0xC000 : 0xDFFF 
  z80_write_map[7] = Ram_Z80;          // 0xE000 : 0xFFFF

  sms.fcr[0] = 0x00;
  sms.fcr[1] = 0x00;
  sms.fcr[2] = 0x01;
  sms.fcr[3] = 0x00;

  // iniciando o z80
#ifdef ASMZ80

  sz80_Add_Fetch(&M_SZ80, 0x00, 0xBF, Rom_Data);
  sz80_Add_Fetch(&M_SZ80, 0xC0, 0xDF, Ram_Z80);
  sz80_Add_Fetch(&M_SZ80, 0xE0, 0xFF, Ram_Z80);
  sz80_Reset(&M_SZ80);

#else

  ResetZ80(&M_SZ80);

#endif
}
