
#include "sms.h"
#include "vdp_io.h"
#include "movie.h"

int    SMS_Started   = 0;
int    SMS_Rom_Pages = 0;
int    Is_GG         = 0;
extern WORD MD_Screen[];

int __fastcall SMS_Init(LPGROM_INFO rom)
{
  if(rom->Type == SMS_ROM)
  {
    Is_GG = 0;
    Set_Title("SMS", Rom_Name);
  }
  else if(rom->Type == GG_ROM)
  {
    Is_GG = 1;
    Set_Title("GG", Rom_Name);
  }

  SMS_Rom_Pages = Rom_Size / 0x4000;
  // inicia o VDP
  SMS_VDP_Reset();
  // inicia Z80 e mapeia memria
  SMS_CPU_Init();

  SMS_Reset();
  render_init();
  SMS_Audio_Init();

  SMS_Load_SRAM();
  Frame_Number = 0;

  if(Sound_Enable)
  {
    End_Sound();
    if(!Init_Sound(ghWnd)) Sound_Enable = 0;
    else Play_Sound();
  }

  Update_Frame = Do_SMS_Frame;
  Update_Frame_Fast = Do_SMS_Frame_Fast;

  return 1;
}

void __fastcall SMS_End_Emulation(void)
{
  if     (Movie_Status & MOVIE_PLAYING)   StopPlayMovie();
  else if(Movie_Status & MOVIE_RECORDING) StopRecordMovie();

  SMS_Save_SRAM();
  SMS_Audio_End();
}

void __stdcall SMS_Load_SRAM(void)
{
  HANDLE hFile;
  DWORD  dw;
  char   pszName[GMAX_PATH + 32];

  memset(SRAM, 0, 0x8000);
  SetCurrentDirectory(Gens_Path);
  sprintf(pszName, "%s%s%s", SRAM_Dir, Rom_Name, ".srm");
  hFile = CreateFile(pszName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
  if(hFile != INVALID_HANDLE_VALUE)
  {
    ReadFile(hFile, SRAM, 0x8000, &dw, NULL);
    CloseHandle(hFile);
    sprintf(Str_Tmp, "SRAM loaded from %s", pszName);
    Put_Info(Str_Tmp, 2000);
  }
}


void __stdcall SMS_Save_SRAM(void)
{
  HANDLE hFile;
  DWORD  dw;
  char   pszName[GMAX_PATH + 32];
  int    i, size;

  i = 0x8000 - 1;
  while(i >= 0 && !SRAM[i]) i--;
  if(i < 0) return; // no need to save SRAM
  i++;
  size = 1;
  while(i > size) size <<= 1;

  SetCurrentDirectory(Gens_Path);
  sprintf(pszName, "%s%s%s", SRAM_Dir, Rom_Name, ".srm");
  hFile = CreateFile(pszName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
    FILE_ATTRIBUTE_ARCHIVE, NULL);
  if(hFile != INVALID_HANDLE_VALUE)
  {
    WriteFile(hFile, SRAM, size, &dw, NULL);
    CloseHandle(hFile);
    sprintf(Str_Tmp, "SRAM saved in %s", pszName);
    Put_Info(Str_Tmp, 2000);
  }
}

void __stdcall Export_SMS(LPBYTE data)
{
  int i, cur;

  data[0x00] = 'M';
  data[0x01] = 'S';
  data[0x02] = 'T';
  data[0x03] = 0x40;
  data[0x04] = 0xE0;

  data[0x50] = 5;    // Version
  data[0x51] = 0;    // Gens

  cur = 0x60;
  for(i = 0; i < sizeof(SMS_CONTEXT); i++) data[cur++] = ((LPBYTE)&sms)[i];
  // Z80
#ifdef ASMZ80
  for(i = 0; i < sizeof(SZ80_CONTEXT) - 0x108; i++) data[cur++] = ((LPBYTE)&M_SZ80)[i];
  for(i = 0; i < 0x2000; i++)              data[cur++] = Ram_Z80[i];
#else
  for(i = 0; i < sizeof(SZ80); i++) data[cur++] = ((LPBYTE)&M_SZ80)[i];
  for(i = 0; i < 0x2000; i++)       data[cur++] = Ram_Z80[i];
#endif
  // VDP
  for(i = 0; i < sizeof(SVDP); i++) data[cur++] = ((LPBYTE)&SMS_VDP)[i];
  for(i = 0; i < 0x40; i++)         data[cur++] = CRam[i];
  for(i = 0; i < 0x4000; i++)       data[cur++] = VRam[i];
  // ym2413
  for(i = 0; i < sizeof(OPLL); i++) data[cur++] = ((LPBYTE)opll)[i];
  // PSG
  for(i = 0; i < sizeof(t_SN76496); i++) data[cur++] = ((LPBYTE)&sn)[i];
}

void __stdcall Import_SMS(LPBYTE data)
{
  LPBYTE src;
  int    i;

  data += 0x60;

  src = (LPBYTE)&sms;
  for(i = 0; i < sizeof(SMS_CONTEXT); i++) src[i] = *data++;
  // Z80
  src = (LPBYTE)&M_SZ80;
#ifdef ASMZ80
  for(i = 0; i < sizeof(SZ80_CONTEXT) - 0x108; i++) src[i] = *data++;
#else
  for(i = 0; i < sizeof(SZ80); i++) src[i] = *data++;
#endif
  src = Ram_Z80;
  for(i = 0; i < 0x2000; i++)       src[i] = *data++;
  // VDP
  src = (LPBYTE)&SMS_VDP;
  for(i = 0; i < sizeof(SVDP); i++) src[i] = *data++;
  src = CRam;
  for(i = 0; i < 0x40; i++)         src[i] = *data++;
  src = VRam;
  for(i = 0; i < 0x4000; i++)       src[i] = *data++;
  if(snd.init_flag)
  {
    src = (LPBYTE)opll->reg;
    for(i = 0; i < 0x52; i++) src[i] = *data++;
    data += sizeof(OPLL) - 0x52;
    // PSG
    src = (LPBYTE)&sn;
    for(i = 0; i < sizeof(t_SN76496); i++) src[i] = *data++;
  }

  SMS_Map_Mem(0, sms.fcr[0]);
  SMS_Map_Mem(1, sms.fcr[1]);
  SMS_Map_Mem(2, sms.fcr[2]);
  SMS_Map_Mem(3, sms.fcr[3]);
  is_vram_dirty = 1;
  for(i = 0; i < 0x200; i++) vram_dirty[i] = 1;
  update_cache();
  for(i = 0; i < SMS_PALETTE_SIZE; i += 1) palette_sync(i);
}
