#ifndef _EMU2413_H_
#define _EMU2413_H_

#include <windows.h>

#define EMU2413_API
#define EMU2413_COMPACTION

#ifdef __cplusplus
extern "C" {
#endif

#define PI 3.14159265358979

enum {OPLL_2413_TONE=0, OPLL_VRC7_TONE=1};

// voice data
typedef struct {
  UINT32 TL,FB,EG,ML,AR,DR,SL,RR,KR,KL,AM,PM,WF;
} OPLL_PATCH;

// slot
typedef struct {

  OPLL_PATCH *patch;  

  int type;          // 0 : modulator 1 : carrier

  // OUTPUT
  int feedback;
  int output[5];   // Output value of slot

  // for Phase Generator (PG)
  WORD *sintbl;    // Wavetable
  UINT32 phase;      // Phase
  UINT32 dphase;     // Phase increment amount
  UINT32 pgout;      // output

  // for Envelope Generator (EG)
  int fnum;          // F-Number
  int block;         // Block
  int volume;        // Current volume
  int sustine;       // Sustine 1 = ON, 0 = OFF
  UINT32 tll;	      // Total Level + Key scale level*/
  UINT32 rks;        // Key scale offset (Rks)
  int eg_mode;       // Current state
  UINT32 eg_phase;   // Phase
  UINT32 eg_dphase;  // Phase increment amount
  UINT32 egout;      // output

  // refer to opll->
  int *plfo_pm;
  int *plfo_am;

} OPLL_SLOT;

// Channel
typedef struct {

  int patch_number;
  int key_status;
  OPLL_SLOT *mod, *car;

} OPLL_CH;

// Mask
#define OPLL_MASK_CH(x) (1<<(x))
#define OPLL_MASK_HH (1<<(9))
#define OPLL_MASK_CYM (1<<(10))
#define OPLL_MASK_TOM (1<<(11))
#define OPLL_MASK_SD (1<<(12))
#define OPLL_MASK_BD (1<<(13))
#define OPLL_MASK_RYTHM ( OPLL_MASK_HH | OPLL_MASK_CYM | OPLL_MASK_TOM | OPLL_MASK_SD | OPLL_MASK_BD )

// opll
typedef struct {

  // Register
  BYTE reg[0x40];
  int slot_on_flag[18];

  UINT32 realstep;
  UINT32 oplltime;
  UINT32 opllstep;

  UINT32 adr;

  int out, prev, next;
#ifndef EMU2413_COMPACTION
  int sprev[2],snext[2];
  UINT32 pan[16];
#endif

  // Pitch Modulator
  UINT32 pm_phase;
  int lfo_pm;

  // Amp Modulator
  int am_phase;
  int lfo_am;

  UINT32 quality;

  // Noise Generator
  UINT32 noise_seed;
  UINT32 whitenoise;
  UINT32 noiseA;
  UINT32 noiseB;
  UINT32 noiseA_phase;
  UINT32 noiseB_phase;
  UINT32 noiseA_dphase;
  UINT32 noiseB_dphase;

  // Channel & Slot
  OPLL_CH *ch[9];
  OPLL_SLOT *slot[18];

  // Voice Data
  OPLL_PATCH *patch[19*2];
  int patch_update[2]; // flag for check patch update

  UINT32 mask;

} OPLL;

// Create Object
EMU2413_API OPLL *OPLL_new(UINT32 clk, UINT32 rate);
EMU2413_API void OPLL_delete(OPLL *);

// Setup
EMU2413_API void OPLL_reset(OPLL *);
EMU2413_API void OPLL_reset_patch(OPLL *, int);
EMU2413_API void OPLL_set_rate(OPLL *opll, UINT32 r);
EMU2413_API void OPLL_set_quality(OPLL *opll, UINT32 q);
EMU2413_API void OPLL_set_pan(OPLL *, UINT32 ch, UINT32 pan);

// Port/Register access
EMU2413_API void OPLL_writeIO(OPLL *, UINT32 reg, UINT32 val);
EMU2413_API void OPLL_writeReg(OPLL *, UINT32 reg, UINT32 val);

// Synthsize
EMU2413_API short OPLL_calc(OPLL *);
EMU2413_API void OPLL_calc_stereo(OPLL *, int out[2]);

// Misc
EMU2413_API void OPLL_setPatch(OPLL *, LPCBYTE dump);
EMU2413_API void OPLL_copyPatch(OPLL *, int, OPLL_PATCH *);
EMU2413_API void OPLL_forceRefresh(OPLL *);
// Utility
EMU2413_API void OPLL_dump2patch(const BYTE *dump, OPLL_PATCH *patch);
EMU2413_API void OPLL_patch2dump(const OPLL_PATCH *patch, BYTE *dump);
EMU2413_API void OPLL_getDefaultPatch(int type, int num, OPLL_PATCH *);

// Channel Mask
EMU2413_API UINT32 OPLL_setMask(OPLL *, UINT32 mask);
EMU2413_API UINT32 OPLL_toggleMask(OPLL *, UINT32 mask);

#define dump2patch OPLL_dump2patch

#ifdef __cplusplus
}
#endif

#endif
