/* sgm.cpp - projeto gens+. Rodrigo Cardoso.
   Loga arquivos SGM. O formato VGM fica muito grande para genesis, estou tentando criar
   um formato que fique menor. */

#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include "zlib.h"
#include "gmain.h"
#include "grom.h"
#include "config.h"
#include "sgm.h"

extern "C" int CPU_Mode;
extern "C" BYTE State_Buffer[];
FILE      *SGM_File = NULL;
char       SGM_File_Name[GMAX_PATH] = "";
SGMHEADER  sgmhdr;

#define sgmputc(c) putc(c, SGM_File);

/* lista de comandos:
   00       - YM2612 e PSG update (735 samples)
   01       - YM2612 dac e timers update
   02 rr nn - YM2612 port 1, escreve valor nn para o reg rr e muda Cur_AAdr para rr
   03 rr nn - YM2612 port 2, escreve valor nn para o reg rr e muda Cur_BAdr para rr
   04 nn    - PSG, escreve valor nn
   05 rr    - YM2612 port 1, muda Cur_AAdr para rr
   06 rr    - YM2612 port 2, muda Cur_BAdr para rr
   07 nn    - YM2612 port 1, escreve valor nn para Cur_AAdr
   08 nn    - YM2612 port 2, escreve valor nn para Cur_BAdr
   2X nn    - YM2612 port 1, escreve valor nn para registro 2X
   FF       - EOF */

int Cur_AAdr, Cur_BAdr;

// cria o arquivo e inicia o cabealho
void __stdcall Start_SGM_Dump(void)
{
  LPSTR     sufix;
  int       i, j, d;
  int       file;
  CHIPSINFO ci;

  if(!(*Rom_Name)) return;

  i = sprintf(SGM_File_Name, "%s%s", Dump_Dir, Rom_Name);
  sufix = SGM_File_Name + i;
  d   = -1;
  do
  {
    if (d++ > 99999) return;
    sufix[0] = '_';
    i = 1;
    j = d / 10000;
    if (j) sufix[i++] = '0' + j;
    j = (d / 1000) % 10;
    if (j) sufix[i++] = '0' + j;
    j = (d / 100) % 10;
    sufix[i++] = '0' + j;
    j = (d / 10) % 10;
    sufix[i++] = '0' + j;
    j = d % 10;
    sufix[i++] = '0' + j;
    sufix[i++] = '.';
    sufix[i++] = 's';
    sufix[i++] = 'g';
    sufix[i++] = 'm';
    sufix[i]   = 0;
    file = open(SGM_File_Name, _O_CREAT | _O_EXCL | _O_WRONLY | _O_BINARY | _O_SEQUENTIAL);
  } while(file == -1);
  
  SGM_File = fdopen(file, "wbS");
  if(!SGM_File) return;

  memset(&sgmhdr, 0, sizeof(SGMHEADER));
  sgmhdr.dwType     = SGM_TYPE;
  sgmhdr.dwVersion  = SGM_VERSION;
  sgmhdr.dwMode     = CPU_Mode ? 1 : 0;
  // copiando os dados da rom
  memcpy(&sgmhdr.griRomInfo, &Game->Genesis, sizeof(GENESIS_ROM_INFO));
  fwrite(&sgmhdr, sizeof(SGMHEADER), 1, SGM_File);

  // salvando o estado inicial dos chips de som
  for(i = 0; i < 8; i++) ci.PSG_Reg[i] = PSG.Register[i];
  for(i = 0; i < 0x100; i++)
  {
    ci.YM2612_Reg[0x000 + i] = YM2612.REG[0][i];
    ci.YM2612_Reg[0x100 + i] = YM2612.REG[1][i];
  }
  fwrite(&ci, sizeof(CHIPSINFO), 1, SGM_File);
  Cur_AAdr = YM2612.OPNAadr;
  Cur_BAdr = YM2612.OPNBadr;
  sgmputc(0x05); sgmputc(Cur_AAdr);
  sgmputc(0x06); sgmputc(Cur_BAdr);
}

// fim
void __stdcall Stop_SGM_Dump(void)
{
  if(!SGM_File) return;
  sgmputc(0xFF);
  sgmhdr.dwFileSize = ftell(SGM_File); // salvando o nmero total de samples
  fseek(SGM_File, 0, SEEK_SET);
  fwrite(&sgmhdr, sizeof(SGMHEADER), 1, SGM_File);
  fclose(SGM_File);
  SGM_File = NULL;
}

// adiciona comando
void __stdcall Update_SGM_Dump(int cmd, int b1, int b2)
{
  switch(cmd)
  {
    case 0x00:
      sgmhdr.dwTotalLength += CPU_Mode ? 882 : 735;
    case 0x01:
      sgmputc(cmd);
      break;
    case 0x02:
      if((b1 & 0xF0) == 0x20)
      { sgmputc(b1); sgmputc(b2); }
      else if(Cur_AAdr == b1)
      { sgmputc(0x07); sgmputc(b2); }
      else
      { sgmputc(0x02); sgmputc(b1); sgmputc(b2); Cur_AAdr = b1; }
      break;
    case 0x03:
      if(Cur_BAdr == b1)
      { sgmputc(0x08); sgmputc(b2); }
      else
      { sgmputc(0x03); sgmputc(b1); sgmputc(b2); Cur_BAdr = b1; }
      break;
    case 0x04:
      sgmputc(0x04); sgmputc(b1);
      break;
  }
}

// compress the movie file in Movie_File_Name
int __stdcall Compress_SGM_File(void)
{
  HANDLE  fIn;
  GZFILE  fOut;
  char    pOutName[GMAX_PATH];
  LPSTR   ext;
  int     i;
  DWORD   sz;

  // open the input file for reading
  fIn = CreateFile(SGM_File_Name, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
  if(fIn == INVALID_HANDLE_VALUE) return 0;
  // creating output compressed file
  ext = strrchr(SGM_File_Name, '.');
  if(ext) *ext = 0;
  sprintf(pOutName, "%s.%s", SGM_File_Name, "sgz");
  fOut = gzopen(pOutName, "wb5");
  if(!fOut) { CloseHandle(fIn); return 0; }

  i  = 1;
  sz = 0x4000;
  while(sz == 0x4000)
  {
    // reading the input file
    if(ReadFile(fIn, State_Buffer, 0x4000, &sz, NULL))
    { 
      // compressing data to output file
      if(gzwrite(fOut, State_Buffer, sz) != sz) { sz = 0; i = 0; break; }
    }
    else break;
  }

  gzclose(fOut);
  CloseHandle(fIn);

  return i;
}

