; sound.asm - este arquivo pertence ao projeto gens+
; Rodrigo Cardoso

%include "nasmhead.inc"

section .text align=64

ALIGN64

; --------------------------------------------------------------------------------------
; void __cdecl Write_Sound_Mono_MMX(int *Left, int *Right, short *Dest, int lenght)
; copia Seg_L e Seg_R para Dest
DECL Write_Sound_Mono_MMX

  push ebx
  push ecx
  push edi
  push esi

  mov edi, [esp + 20]      ; Left
  mov esi, [esp + 24]      ; Right
  mov ecx, [esp + 32]      ; Lenght
  mov ebx, [esp + 28]      ; Dest

  shr ecx, 1
  jnc short .Double_Trans

.Simple_Trans
  mov eax, [edi]
  mov dword [edi], 0
  add eax, [esi]
  mov dword [esi], 0

  cmp eax, 0xFFFF
  jle short .lower_s

  mov word [ebx], 0x7FFF
  jmp short .ok_s

.lower_s
  cmp eax, -0xFFFF
  jge short .greater_s

  mov word [ebx], -0x7FFF
  jmp short .ok_s

.greater_s
  shr eax, 1
  mov [ebx], ax

.ok_s
  add edi, 4
  add esi, 4
  add ebx, 2

.Double_Trans
  mov eax, 32
  pxor mm4, mm4
  movd mm5, eax
  mov eax, 1
  test ecx, ecx
  movd mm6, eax
  jnz short .Loop
  jmp short .End

ALIGN64

.Loop
  movq mm0, [edi]      ; L2 | L1
  add ebx, 4
  movq [edi], mm4
  movq mm1, [esi]      ; R2 | R1
  add edi, 8
  movq [esi], mm4
  packssdw mm0, mm0    ; 0 | 0 | L2 | L1
  packssdw mm1, mm1    ; 0 | 0 | R2 | R1
  psraw mm0, 1
  psraw mm1, 1
  add esi, 8
  paddw mm0, mm1      ; 0 | 0 | R2 + L2 | R1 + L1
  dec ecx
  movd [ebx - 4], mm0
  jnz short .Loop

  emms

.End
  pop esi
  pop edi
  pop ecx
  pop ebx
  ret
; --------------------------------------------------------------------------------------

ALIGN64

; --------------------------------------------------------------------------------------
; void __cdecl Write_Sound_Stereo_MMX(int *Left, int *Right, short *Dest, int lenght)
DECL Write_Sound_Stereo_MMX

  push ebx
  push ecx
  push edx
  push edi
  push esi

  mov edi, [esp + 24]      ; Left
  mov esi, [esp + 28]      ; Right
  mov ecx, [esp + 36]      ; Lenght
  mov ebx, [esp + 32]      ; Dest

  shr ecx, 1
  jnc short .Double_Trans

.Simple_Trans
  mov eax, [edi]
  cmp eax, 0x7FFF
  mov dword [edi], 0
  jle short .lower_s1

  mov word [ebx + 0], 0x7FFF
  jmp short .right_s1

.lower_s1
  cmp eax, -0x7FFF
  jge short .greater_s1

  mov word [ebx + 0], -0x7FFF
  jmp short .right_s1

.greater_s1
  mov [ebx + 0], ax

.right_s1
  mov edx, [esi]
  cmp edx, 0x7FFF
  mov dword [esi], 0
  jle short .lower_s2

  mov word [ebx + 2], 0x7FFF
  jmp short .ok_s1

.lower_s2
  cmp edx, -0x7FFF
  jge short .greater_s2

  mov word [ebx + 2], -0x7FFF
  jmp short .ok_s1

.greater_s2
  mov [ebx + 2], dx

.ok_s1
  add edi, 4
  add esi, 4
  add ebx, 4

.Double_Trans
  mov eax, 32
  pxor mm4, mm4
  test ecx, ecx
  movd mm5, eax
  jnz short .Loop
  jmp short .End

ALIGN64

.Loop
  movd mm0, [esi]      ; 0  | R1
  add edi, 8
  movd mm1, [esi + 4]    ; 0  | R2
  psllq mm0, mm5      ; R1 |  0
  movq [esi], mm4
  psllq mm1, mm5      ; R2 |  0
  movd mm2, [edi - 8]    ; 0  | L1
  add esi, 8
  movd mm3, [edi - 8 + 4]  ; 0  | L2
  add ebx, 8
  paddd mm0, mm2      ; R1 | L1
  paddd mm1, mm3      ; R2 | L2
  movq [edi - 8], mm4
  packssdw mm0, mm1    ; R2 | L2 | R1 | L1

  dec ecx
  movq [ebx - 8], mm0
  jnz short .Loop

  emms

.End
  pop esi
  pop edi
  pop edx
  pop ecx
  pop ebx
  ret
; --------------------------------------------------------------------------------------

ALIGN64

; --------------------------------------------------------------------------------------
; void __cdecl Set_Sound_Volume(short *buf, int len, int percent)
; muda o volume do som
DECL Set_Sound_Volume
  push  ebx
  push  edi
  push  esi

  ; verificando a validade dos argumentos
  mov   edi, [esp + 16]  ; buffer
  test  edi, edi         ; nulo
  jz    .end
  mov   ecx, [esp + 20]
  test  ecx, ecx
  jz    .end
  mov   ebx, [esp + 24]
  mov   esi, 100
  cmp   ebx, esi
  jz    .end
  cmp   ebx, 500         ; verificando os valores mximos [-500%, +500]
  js    .conf
  mov   ebx, 500
  jmp   .loop
  .conf:
  cmp   ebx, 0
  jns   .loop
  xor   ebx, ebx

  .loop:
    ; alterando o volume do som
    movsx eax, word [edi]
    imul  ebx
    idiv  esi
    ; verificando se os valores obtidos esto dentro dos limites (16 bits)
    cmp   eax, 32766        ; limite mximo
    js    .no_max
    mov   ax, 0x7FFF
    jmp   .value_ok
    .no_max:
    cmp   eax, -32767       ; limite minimo
    jns   .no_min
    mov   ax, 0x8000
    jmp   .value_ok
    .no_min:
    test  eax, 0x80000000   ; convertendo o valor de 32 para 16 bits
    jz    .value_ok
    or    ax, 0x8000
    .value_ok:
    mov   [edi], ax
    
    add   edi, byte 2
    dec   ecx
    jnz  .loop

  .end:
  pop   esi
  pop   edi
  pop   ebx
  ret
; --------------------------------------------------------------------------------------

ALIGN64

section .bss  align=64
  Seg_Buffer: resd  882
  Seg_Len:    resd  1
  Buf_Len:    resd  1
  
section .data align=64
  DECL   Echo_Buffer           ; ponteiro para o buffer usado no eco
  dd 0
  extern _Sound_Rate           ; sample rate
  extern _Sound_Stereo
  
%macro MOVMEM 0
  ALIGN4
  %%loop1:
    sub   ecx, 4
    js    %%next
    mov   eax, [esi]
    mov   edx, [esi + 4]
    mov   [edi],     eax
    mov   [edi + 4], edx
    add   esi, 8
    add   edi, 8
    jmp   %%loop1
  %%next:
	add   ecx, 4
  jz    %%end
  cld
  %%loop2:
    movsw
    dec   ecx
    jnz   %%loop2
  %%end
%endmacro

%macro MOVMEM_MMX 0
  ALIGN4
  %%loop1:
    sub   ecx, 32
    js    %%next
    movq  mm0, [esi]
    movq  mm1, [esi + 8]
    movq  mm2, [esi + 16]
    movq  mm3, [esi + 24]
    movq  mm4, [esi + 32]
    movq  mm5, [esi + 40]
    movq  mm6, [esi + 48]
    movq  mm7, [esi + 56]
    movq  [edi],      mm0
    movq  [edi + 8],  mm1
    movq  [edi + 16], mm2
    movq  [edi + 24], mm3
    movq  [edi + 32], mm4
    movq  [edi + 40], mm5
    movq  [edi + 48], mm6
    movq  [edi + 56], mm7
    add   esi, 64
    add   edi, 64
    jmp   %%loop1
  %%next:
  add   ecx, 32
  jz    %%end
  cld
  %%loop2:
    movsw
    dec   ecx
    jnz   %%loop2
  %%end:
%endmacro

%macro SOUND_ECHO 1
  ; verificando os argumentos da funo
  test  ecx, ecx
  jz    near %%end2
  test  edx, edx
  jz    near %%end2
  ; verificando se o buffer est allocado
  cmp   [Echo_Buffer], dword 0
  jz    near %%end2

  push  ebx
  push  edi
  push  esi
  
  mov   esi, ecx
  mov   ecx, edx
  mov   ebx, [_Sound_Rate]
  test  [_Sound_Stereo], byte 1
  jz    %%no_stereo
  add   ebx, ebx
  add   ecx, ecx
  %%no_stereo
  shr   ebx, 3              ; *** DELAY *** aviso: se aumentar o delay tem que aumentar o buffer
  mov   [Buf_Len], ebx
  mov   [Seg_Len], ecx

  ; preparando o buffer
  mov   ebx, esi
  mov   esi, [Echo_Buffer]
  mov   edi, Seg_Buffer
%if(%1 == 1)
  MOVMEM2
%else
  MOVMEM
%endif
  
  ; aplicando o eco
  mov   esi, ebx
  mov   ecx, [Seg_Len]
  mov   edi, Seg_Buffer
  %%echo_loop:
    movsx eax, word [esi]
    movsx edx, word [edi]
    imul  eax, 3            ; *** VOLUME ***
    add   eax, edx
    add   eax, edx
    xor   edx, edx
    cdq
    mov   ebx, 5
    idiv  ebx

    test  eax, 0x80000000   ; convertendo o valor de 32 para 16 bits
    jz    %%copy_ok
    or    ax, 0x8000
    %%copy_ok:
    mov   [esi], ax
    mov   [edi], ax

    add   esi, 2
    add   edi, 2
    dec   ecx
    jnz   %%echo_loop

  ; atualizando o buffer
  mov   edi, [Echo_Buffer]
  mov   eax, [Seg_Len]
  shl   eax, 1
  mov   esi, edi
  add   esi, eax
  mov   ecx, [Buf_Len]
  sub   ecx, [Seg_Len]
%if(%1 == 1)
  MOVMEM_MMX
%else
  MOVMEM
%endif

  mov   esi, Seg_Buffer
  mov   ecx, [Seg_Len]
%if(%1 == 1)
  MOVMEM_MMX
%else
  MOVMEM
%endif

  %%end:
  pop   esi
  pop   edi
  pop   ebx
  emms
  %%end2:
  ret
%endmacro

ALIGN64

; --------------------------------------------------------------------------------------
; void __fastcall Set_Sound_Echo(short *buf, int len)
; atribui eco ao som
DECLF Set_Sound_Echo, 8
  SOUND_ECHO 0
; --------------------------------------------------------------------------------------

ALIGN64

; --------------------------------------------------------------------------------------
; void __fastcall Set_Sound_Echo_MMX(short *buf, int len)
; atribui eco ao som usando registradores MMX
DECLF Set_Sound_Echo_MMX, 8
  SOUND_ECHO 1
; --------------------------------------------------------------------------------------
