
#include "ggraphic.h"
#include "misc.h"
#include "blit.h"
#include "vdp_io.h"
#include "config.h"
#include "state.h"
#include "sms.h"

extern LPDIRECTDRAW         lpDD_Init;
extern LPDIRECTDRAW7        lpDD;
extern LPDIRECTDRAWSURFACE7 lpDDS_Primary;
extern LPDIRECTDRAWSURFACE7 lpDDS_Flip;
extern LPDIRECTDRAWSURFACE7 lpDDS_Back;
extern LPDIRECTDRAWSURFACE7 lpDDS_Blit;
extern LPDIRECTDRAWCLIPPER  lpDDC_Clipper;

LPDIRECT3D7                 lpD3D = NULL;
LPDIRECT3DDEVICE7           lpD3D_Device = NULL;
LPDIRECTDRAWSURFACE7        lpDDS_Texture = NULL;

void D3D_Fail(HWND hWnd, char *err)
{
  End_D3D();
  MessageBox(hWnd, err, "Error Direct3D", MB_ICONSTOP | MB_OK);
  DestroyWindow(hWnd);
  exit(0);
}

int __stdcall Init_D3D(HWND hWnd)
{
  int            rend;
  HRESULT        rval;
  DDSURFACEDESC2 ddsd;
  D3DVIEWPORT7   d3dvp;
  int            rx, ry;

  if(DirectDrawCreateEx(0, (LPVOID *)&lpDD_Init, IID_IDirectDraw7, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: DirectDrawCreate.\nCheck your DirectX version."); return 0; }
  if(lpDD_Init->QueryInterface(IID_IDirectDraw7, (LPVOID *)&lpDD) != DD_OK)
  { D3D_Fail(hWnd, "Error: QueryInterface.\nCheck your DirectX version."); return 0; }
  if(lpDD_Init->QueryInterface(IID_IDirect3D7, (LPVOID *)&lpD3D) != DD_OK)
  { D3D_Fail(hWnd, "Error: QueryInterface.\nCheck your DirectX version."); return 0; }

  lpDD_Init->Release();
  lpDD_Init = NULL;

  if(Full_Screen) rend = Render_FS;
  else            rend = Render_W;

  if(Force_Mode == 2)
  {
    memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
    ddsd.dwSize = sizeof(DDSURFACEDESC2);
    lpDD->GetDisplayMode(&ddsd);
    if(ddsd.ddpfPixelFormat.dwGBitMask == 0x03E0) Mode_555 = 1;
    else                                          Mode_555 = 0;
  }
  else Mode_555 = Force_Mode;

  Recalculate_Palettes();

  if(Full_Screen)
    rval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN/* | DDSCL_FPUSETUP */);
  else
    rval = lpDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL/* | DDSCL_FPUSETUP */);

  if(rval != DD_OK)
  { D3D_Fail(hWnd, "Error: SetCooperativeLevel.\nCheck your DirectX version."); return 0; }

  if(Full_Screen)
  {
    if(Res_X < 640 || Res_Y < 480)
    { rend = 0;
      if(Have_MMX) Blit_FS = Blit_X1_MMX;
      else         Blit_FS = Blit_X1; }

    if(!rend)
    { rx = 320;  ry = 240; }
    else
    { rx = Res_X;  ry = Res_Y; }
		if(lpDD->SetDisplayMode(rx, ry, 16, 0, 0) != DD_OK) //MK2k: Keeping 16bit per pixel, fullscreen didn't work with 32bitPP
		//if(lpDD->SetDisplayMode(rx, ry, 32, 0, 0) != DD_OK)	//MK2k: Set to 32bits per pixel in order to avoid display mode change upon start
    { D3D_Fail(hWnd, "Error: SetDisplayMode."); return 0; }

    if((Res_X < (320 * 3) || Res_Y < (240 * 3)) && Render_FS > 14) 
    {
      Render_FS = 1;
      if(Have_MMX) Blit_FS = Blit_X2_MMX;
      else         Blit_FS = Blit_X2;
    }
  }

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);

  if(Full_Screen && VSync_FS)
  {
    ddsd.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 2;
  }
  else
  {
    ddsd.dwFlags        = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
  }

  if(lpDD->CreateSurface(&ddsd, &lpDDS_Primary, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: CreateSurface."); return 0; }

  if(Full_Screen)
  {
    if(VSync_FS)
    {
      ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
      if(lpDDS_Primary->GetAttachedSurface(&ddsd.ddsCaps, &lpDDS_Flip) != DD_OK)
      { D3D_Fail(hWnd, "Error: GetAttachedSurface."); return 0; }
      lpDDS_Blit = lpDDS_Flip;
    }
    else lpDDS_Blit = lpDDS_Primary;
  }
  else
  {
    if(lpDD->CreateClipper(0, &lpDDC_Clipper, NULL ) != DD_OK)
    { D3D_Fail(hWnd, "Error: CreateClipper."); return 0; }

    if(lpDDC_Clipper->SetHWnd(0, hWnd) != DD_OK)
    { D3D_Fail(hWnd, "Error: SetghWnd."); return 0; }

    if(lpDDS_Primary->SetClipper(lpDDC_Clipper) != DD_OK)
    { D3D_Fail(hWnd, "Error: SetClipper."); return 0; }
  }

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize  = sizeof(DDSURFACEDESC2);
  ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;

  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY | DDSCAPS_3DDEVICE;
  if(rend > 13)      { ddsd.dwWidth = 960; ddsd.dwHeight = 720; }
  else if(rend > 1)  { ddsd.dwWidth = 640; ddsd.dwHeight = 480; }
	else               { ddsd.dwWidth = 320; ddsd.dwHeight = 240; }

  if(lpDD->CreateSurface(&ddsd, &lpDDS_Back, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: CreateSurface."); return 0; }

  if(!Full_Screen) lpDDS_Blit = lpDDS_Back;

  if(lpD3D->CreateDevice(IID_IDirect3DHALDevice, lpDDS_Back, &lpD3D_Device) != DD_OK)
  {
    if(lpD3D->CreateDevice(IID_IDirect3DMMXDevice, lpDDS_Back, &lpD3D_Device) != DD_OK)
    {
      if(lpD3D->CreateDevice(IID_IDirect3DRefDevice, lpDDS_Back, &lpD3D_Device) != DD_OK)
      { D3D_Fail(hWnd, "Error: CreateDevice."); return 0; }
    }
  }

  memset(&d3dvp, 0, sizeof(D3DVIEWPORT7));
  if(Full_Screen) { d3dvp.dwWidth = Res_X; d3dvp.dwHeight = Res_Y; }
  else            { d3dvp.dwWidth = ddsd.dwWidth; d3dvp.dwHeight = ddsd.dwHeight; }
  if(lpD3D_Device->SetViewport(&d3dvp) != DD_OK)
  { D3D_Fail(hWnd, "Error: SetViewport."); return 0; }

  lpD3D_Device->LightEnable(0, FALSE);
  lpD3D_Device->LightEnable(1, FALSE);

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize         = sizeof(DDSURFACEDESC2);
  ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
  ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
  if(rend > 13)      { ddsd.dwWidth = 960; ddsd.dwHeight = 720; }
  else if(rend > 1)  { ddsd.dwWidth = 640; ddsd.dwHeight = 480; }
	else               { ddsd.dwWidth = 320; ddsd.dwHeight = 240; }
  if(lpDD->CreateSurface(&ddsd, &lpDDS_Texture, NULL) != DD_OK)
  { D3D_Fail(hWnd, "Error: CreateSurface."); return 0; }

  return 1;
}

void __stdcall End_D3D(void)
{
  if(lpDDC_Clipper) { lpDDC_Clipper->Release(); lpDDC_Clipper = NULL; }
  if(lpDDS_Back)    { lpDDS_Back->Release();    lpDDS_Back = NULL; }
  if(lpDDS_Texture) { lpDDS_Texture->Release(); lpDDS_Texture = NULL; }
  if(lpDDS_Flip)    { lpDDS_Flip->Release();    lpDDS_Flip = NULL; }
  if(lpDDS_Primary) { lpDDS_Primary->Release(); lpDDS_Primary = NULL; }
  if(lpD3D_Device)  { lpD3D_Device->Release();  lpD3D_Device = NULL; }
  if(lpD3D)         { lpD3D->Release();         lpD3D = NULL; }
  if(lpDD)
  { lpDD->SetCooperativeLevel(ghWnd, DDSCL_NORMAL);
    lpDD->Release();
    lpDD = NULL; }
  lpDDS_Blit = NULL;
}


int __stdcall Clear_D3D_Primary_Screen(HWND hWnd)
{
  return Clear_DDraw_Primary_Screen(hWnd);
}


int __stdcall Clear_D3D_Back_Screen(HWND hWnd)
{
  return Clear_DDraw_Back_Screen(hWnd);
}


void __stdcall Restore_D3D_Primary(void)
{
  Restore_DDraw_Primary();
}


void __stdcall Flip_D3D_FS(HWND hWnd)
{
}

HRESULT RestoreGraphics(void);

void __stdcall Flip_D3D_W(HWND hWnd)
{
  RECT           RectSrc, RectDest;
  int            sx, sy;
  HRESULT        rval;
  DDSURFACEDESC2 ddsd;
  D3DTLVERTEX    lpD3D_Vertex[4];
  POINT          p;

  lpD3D_Device->BeginScene();

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);

  if(SMS_Started && !Debug) // SMS
  {
    if(Flag_Clr_Scr != 24)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 24; }

    sx = 256; sy = 196;
  }
  else if((VDP_Reg.Set4 & 0x01) || Debug)
  {
    if(Flag_Clr_Scr != 40)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 40; }

    sx = 320;  sy = VDP_Num_Vis_Lines;
  }
  else
  {
    if(Flag_Clr_Scr != 32)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 32; }

    sx = 256; sy = VDP_Num_Vis_Lines;
  }

  p.x = p.y = RectSrc.left = RectSrc.top = 0;
  GetClientRect(hWnd, &RectDest);
  ClientToScreen(hWnd, &p);

  if(Render_W > 13)     { RectSrc.right = sx * 3;  RectSrc.bottom = sy * 3; }
  else if(Render_W > 1) { RectSrc.right = sx << 1; RectSrc.bottom = sy << 1; }
  else                  { RectSrc.right = sx;      RectSrc.bottom = sy; }

  if(Stretch_W)
  { RectDest.left    = p.x;  RectDest.top     = p.y;
    RectDest.right  += p.x;  RectDest.bottom += p.y; }
  else
  {
    if(Render_W == 1)
    { RectDest.left    = ((RectDest.right - (RectSrc.right << 1)) >> 1) + p.x;
      RectDest.top     = ((RectDest.bottom - (RectSrc.bottom << 1)) >> 1) + p.y;
      RectDest.right   = RectDest.left + (RectSrc.right << 1);
      RectDest.bottom  = RectDest.top + (RectSrc.bottom << 1); }
    else
    { RectDest.left    = ((RectDest.right - RectSrc.right) >> 1) + p.x;
      RectDest.top     = ((RectDest.bottom - RectSrc.bottom) >> 1) + p.y;
      RectDest.right   = RectDest.left + RectSrc.right;
      RectDest.bottom  = RectDest.top + RectSrc.bottom; }
  }


  rval = lpDDS_Texture->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
  if(rval != DD_OK) return;
  Blit_W((LPBYTE)ddsd.lpSurface, ddsd.lPitch, sx, sy, (336 - sx) << 1);
  lpDDS_Texture->Unlock(NULL);

  memset(lpD3D_Vertex, 0, 4 * sizeof(D3DTLVERTEX));
  lpD3D_Vertex[1].sy    = (float)RectSrc.bottom;
  lpD3D_Vertex[2].sx    = (float)RectSrc.right;
  lpD3D_Vertex[2].sy    = (float)RectSrc.bottom;
  lpD3D_Vertex[3].sx    = (float)RectSrc.right;
  lpD3D_Vertex[0].tu    = 0.0f;
  lpD3D_Vertex[0].tv    = 0.0f;
  lpD3D_Vertex[1].tu    = 1.0f;
  lpD3D_Vertex[1].tv    = 0.0f;
  lpD3D_Vertex[2].tu    = 0.0f;
  lpD3D_Vertex[2].tv    = 1.0f;
  lpD3D_Vertex[3].tu    = 1.0f;
  lpD3D_Vertex[3].tv    = 1.0f;
  lpD3D_Vertex[0].rhw   = 1.0f;
  lpD3D_Vertex[1].rhw   = 1.0f;
  lpD3D_Vertex[2].rhw   = 1.0f;
  lpD3D_Vertex[3].rhw   = 1.0f;
  lpD3D_Vertex[0].color = 0xFFFFFF;
  lpD3D_Vertex[1].color = 0xFFFFFF;
  lpD3D_Vertex[2].color = 0xFFFFFF;
  lpD3D_Vertex[3].color = 0xFFFFFF;
  lpD3D_Device->SetRenderTarget(lpDDS_Back, 0);
  lpD3D_Device->SetRenderState(D3DRENDERSTATE_FILLMODE, D3DFILL_WIREFRAME);
  lpD3D_Device->SetTexture(0, lpDDS_Texture);
  lpD3D_Device->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
  lpD3D_Device->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTFG_POINT);
  lpD3D_Device->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
  lpD3D_Device->SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
  lpD3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX, lpD3D_Vertex, 4, D3DDP_WAIT);
  lpD3D_Device->EndScene();

  if(VSync_W)
  {
    int vb;
    lpDD->GetVerticalBlankStatus(&vb);
    if(!vb) lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);
  }
  rval = lpDDS_Primary->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL);
  if(rval == DDERR_SURFACELOST) rval = RestoreGraphics();

}
