
#include "ggraphic.h"
#include "gdsound.h"
#include "gdinput.h"
#include "gmain.h"
#include "resource.h"
#include "gens.h"
#include "mem_M68K.h"
#include "vdp_io.h"
#include "vdp_rend.h"
#include "misc.h"
#include "blit.h"
#include "net.h"
#include "cdda_mp3.h"
#include "config.h"
#include "state.h"
#include "sms.h"

LPDIRECTDRAW         lpDD_Init     = NULL;
LPDIRECTDRAW7        lpDD          = NULL;
LPDIRECTDRAWSURFACE7 lpDDS_Primary = NULL;
LPDIRECTDRAWSURFACE7 lpDDS_Flip    = NULL;
LPDIRECTDRAWSURFACE7 lpDDS_Back    = NULL;
LPDIRECTDRAWSURFACE7 lpDDS_Blit    = NULL;
LPDIRECTDRAWCLIPPER  lpDDC_Clipper = NULL;

void DDraw_Fail(HWND hWnd, char *err)
{
  End_DDraw();
  MessageBox(hWnd, err, "Error DirectDraw", MB_ICONSTOP | MB_OK);
  DestroyWindow(hWnd);
  exit(0);
}

int __stdcall Init_DDraw(HWND hWnd)
{
  int            rend;
  HRESULT        rval;
  DDSURFACEDESC2 ddsd;
  int            rx, ry;

  if(DirectDrawCreate(0, &lpDD_Init, NULL) != DD_OK)
  { DDraw_Fail(hWnd, "Error: DirectDrawCreate.\nCheck your DirectX version."); return 0; }
  if(lpDD_Init->QueryInterface(IID_IDirectDraw7, (LPVOID *)&lpDD) != DD_OK)
  { DDraw_Fail(hWnd, "Error: QueryInterface.\nCheck your DirectX version."); return 0; }

  lpDD_Init->Release();
  lpDD_Init = NULL;

  if(Full_Screen) rend = Render_FS;
  else            rend = Render_W;

  if(Force_Mode == 2)
  {
    memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
    ddsd.dwSize = sizeof(DDSURFACEDESC2);
    lpDD->GetDisplayMode(&ddsd);
    if(ddsd.ddpfPixelFormat.dwGBitMask == 0x03E0) Mode_555 = 1;
    else                                          Mode_555 = 0;
  }
  else Mode_555 = Force_Mode;

  Recalculate_Palettes();

  if(Full_Screen)
    rval = lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
  else
    rval = lpDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL);

  if(rval != DD_OK)
  { DDraw_Fail(hWnd, "Error: SetCooperativeLevel.\nCheck your DirectX version."); return 0; }

  if(Full_Screen)
  {
    if(Res_X < 640 || Res_Y < 480)
    { rend = 0;
      if(Have_MMX) Blit_FS = Blit_X1_MMX;
      else         Blit_FS = Blit_X1; }

    if(!rend)
    { rx = 320;  ry = 240; }
    else
    { rx = Res_X;  ry = Res_Y; }
		//if(lpDD->SetDisplayMode(rx, ry, 16, 0, 0) != DD_OK) // MK2k: Keep 16bitPP, full screen didn't work with 32bitPP
		if(lpDD->SetDisplayMode(rx, ry, 32, 0, 0) != DD_OK)	//MK2k: Set to 32bits per pixel in order to avoid display mode change upon start
    { DDraw_Fail(hWnd, "Error: SetDisplayMode."); return 0; }

    if((Res_X < (320 * 3) || Res_Y < (240 * 3)) && Render_FS > 14) 
    {
      Render_FS = 1;
      if(Have_MMX) Blit_FS = Blit_X2_MMX;
      else         Blit_FS = Blit_X2;
    }
  }

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);

  if(Full_Screen && VSync_FS)
  {
    ddsd.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 2;
  }
  else
  {
    ddsd.dwFlags        = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
  }

  if(lpDD->CreateSurface(&ddsd, &lpDDS_Primary, NULL) != DD_OK)
  { DDraw_Fail(hWnd, "Error: CreateSurface."); return 0; }

  if(Full_Screen)
  {
    if(VSync_FS)
    {
      ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;
      if(lpDDS_Primary->GetAttachedSurface(&ddsd.ddsCaps, &lpDDS_Flip) != DD_OK)
      { DDraw_Fail(hWnd, "Error: GetAttachedSurface."); return 0; }
      lpDDS_Blit = lpDDS_Flip;
    }
    else lpDDS_Blit = lpDDS_Primary;
  }
  else
  {
    if(lpDD->CreateClipper(0, &lpDDC_Clipper, NULL ) != DD_OK)
    { DDraw_Fail(hWnd, "Error: CreateClipper."); return 0; }

    if(lpDDC_Clipper->SetHWnd(0, hWnd) != DD_OK)
    { DDraw_Fail(hWnd, "Error: SetghWnd."); return 0; }

    if(lpDDS_Primary->SetClipper(lpDDC_Clipper) != DD_OK)
    { DDraw_Fail(hWnd, "Error: SetClipper."); return 0; }
  }

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize  = sizeof(DDSURFACEDESC2);
  ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;

  if(rend < 2 && !Blit_Soft)
  {
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
    ddsd.dwWidth        = 336;
    ddsd.dwHeight       = 240;
  }
  else
  {
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
    if(rend > 13)      { ddsd.dwWidth = 960; ddsd.dwHeight = 720; }
    else if(rend > 1)  { ddsd.dwWidth = 640; ddsd.dwHeight = 480; }
	  else               { ddsd.dwWidth = 320; ddsd.dwHeight = 240; }
  }

  if(lpDD->CreateSurface(&ddsd, &lpDDS_Back, NULL) != DD_OK)
  { DDraw_Fail(hWnd, "Error: CreateSurface."); return 0; }

  if(!Full_Screen) lpDDS_Blit = lpDDS_Back;

  if(rend < 2 && !Blit_Soft)
  {
    memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
    ddsd.dwSize = sizeof(DDSURFACEDESC2);

    if(lpDDS_Back->GetSurfaceDesc(&ddsd) != DD_OK)
    { DDraw_Fail(hWnd, "Error: GetSurfaceDesc"); return 0; }

    ddsd.dwFlags   = DDSD_WIDTH | DDSD_HEIGHT | DDSD_PITCH | DDSD_LPSURFACE | DDSD_PIXELFORMAT;
    ddsd.dwWidth   = 336;
    ddsd.dwHeight  = 240;
    ddsd.lPitch    = 336 * 2;
    ddsd.lpSurface = MD_Screen;
    if(lpDDS_Back->SetSurfaceDesc(&ddsd, 0) != DD_OK)
    { DDraw_Fail(hWnd, "Error: SetSurfaceDesc."); return 0; }
  }

  return 1;
}

void __stdcall End_DDraw(void)
{
  if(lpDDC_Clipper) { lpDDC_Clipper->Release(); lpDDC_Clipper = NULL; }
  if(lpDDS_Back)    { lpDDS_Back->Release();    lpDDS_Back = NULL; }
  if(lpDDS_Flip)    { lpDDS_Flip->Release();    lpDDS_Flip = NULL; }
  if(lpDDS_Primary) { lpDDS_Primary->Release(); lpDDS_Primary = NULL; }
  if(lpDD)
  { lpDD->SetCooperativeLevel(ghWnd, DDSCL_NORMAL);
    lpDD->Release();
    lpDD = NULL; }
  lpDDS_Blit = NULL;
}


HRESULT RestoreGraphics(void)
{
  HRESULT rval;

  rval = lpDDS_Primary->Restore();
  rval = lpDDS_Back->Restore();

  return rval;
}


int __stdcall Clear_DDraw_Primary_Screen(HWND hWnd)
{
  DDSURFACEDESC2 ddsd;
  DDBLTFX        ddbltfx;
  RECT           RD;
  POINT          p;

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);

  memset(&ddbltfx, 0, sizeof(DDBLTFX));
  ddbltfx.dwSize      = sizeof(DDBLTFX);

  if(Full_Screen)
  {
    if(VSync_FS)
    {
      lpDDS_Flip->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
      lpDDS_Primary->Flip(NULL, DDFLIP_WAIT);

      lpDDS_Flip->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
      lpDDS_Primary->Flip(NULL, DDFLIP_WAIT);

      lpDDS_Flip->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
      lpDDS_Primary->Flip(NULL, DDFLIP_WAIT);
    }
    else lpDDS_Primary->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
  }
  else
  {
    p.x = p.y = 0;
    GetClientRect(hWnd, &RD);
    ClientToScreen(hWnd, &p);

    RD.left = p.x;
    RD.top = p.y;
    RD.right += p.x;
    RD.bottom += p.y;

    if(RD.top < RD.bottom)
      lpDDS_Primary->Blt(&RD, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
  }

  return 1;
}


int __stdcall Clear_DDraw_Back_Screen(HWND hWnd)
{
  DDSURFACEDESC2 ddsd;
  DDBLTFX ddbltfx;

  memset(&ddsd, 0, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);

  memset(&ddbltfx, 0, sizeof(ddbltfx));
  ddbltfx.dwSize = sizeof(ddbltfx);
  ddbltfx.dwFillColor = 0;

  lpDDS_Back->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);

  return 1;
}


void __stdcall Restore_DDraw_Primary(void)
{
  if(lpDD && Full_Screen && VSync_FS)
  {
    while (lpDDS_Primary->GetFlipStatus(DDGFS_ISFLIPDONE) != DD_OK);
    lpDD->FlipToGDISurface();
  }
}


void __stdcall Flip_DDraw_FS(HWND hWnd)
{
  HRESULT        rval;
  RECT           RectSrc, RectDest;
  int            sx, sy;
  DDSURFACEDESC2 ddsd;

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);

  if(SMS_Started && !Debug) // SMS
  {
    if(Flag_Clr_Scr != 24)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 24; }

    sx = 256; sy = 196;
  }
  else if((VDP_Reg.Set4 & 0x01) || Debug)
  {
    if(Flag_Clr_Scr != 40)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 40; }

    sx = 320;  sy = VDP_Num_Vis_Lines;
  }
  else
  {
    if(Flag_Clr_Scr != 32)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 32; }

    sx = 256; sy = VDP_Num_Vis_Lines;
  }

  // blit hardware
  if(Render_FS < 2 && !Blit_Soft)
  {
    RectSrc.top  = 0; RectSrc.bottom = sy;
    RectSrc.left = 8; RectSrc.right  = sx + 8;
    if(Stretch_FS)
    {
      RectDest.left = 0; RectDest.top  = 0;
      if(Render_FS) { RectDest.right  = Res_X; RectDest.bottom = Res_Y; }
      else          { RectDest.right  = 320;   RectDest.bottom   = 240; }
    }
    else
    {
      if(Render_FS)
      { sx <<= 1; sy <<= 1;
        RectDest.left = (Res_X - sx) >> 1; RectDest.top  = (Res_Y - sy) >> 1; }
      else
      { RectDest.left = (320 - sx) >> 1; RectDest.top  = (240 - sy) >> 1; }
      RectDest.right  = RectDest.left + sx; RectDest.bottom = RectDest.top + sy;
    }

    if(VSync_FS)
    { lpDDS_Flip->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL);
      lpDDS_Primary->Flip(NULL, DDFLIP_WAIT); }
    else
    { lpDDS_Primary->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL); }
  }

  // blit software
  else
  {
    if(Stretch_FS)
    {
      rval = lpDDS_Back->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
      if(rval != DD_OK) return;
      Blit_FS((LPBYTE)ddsd.lpSurface, ddsd.lPitch, sx, sy, (336 - sx) << 1);
      lpDDS_Back->Unlock(NULL);

      if(Render_FS > 13)     { sx *= 3;  sy *= 3; }
      else if(Render_FS > 1) { sx <<= 1; sy <<= 1; }
      RectSrc.left    = 0;  RectSrc.right   = sx;
      RectSrc.top     = 0;  RectSrc.bottom  = sy;
      RectDest.left   = 0;  RectDest.right  = Res_X;
      RectDest.top    = 0;  RectDest.bottom = Res_Y;

      if(VSync_FS)
      { lpDDS_Flip->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL);
        lpDDS_Primary->Flip(NULL, DDFLIP_WAIT); }
      else
      { lpDDS_Primary->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL); }
    }
    else
    {
      if(Render_FS < 1)
      { RectSrc.left = 320 - sx; RectSrc.top = 240 - sy; }
      else if(Render_FS < 14)
      { RectSrc.left = Res_X - (sx << 1); RectSrc.top = Res_Y - (sy << 1); }
      else
      { RectSrc.left = Res_X - (sx * 3);  RectSrc.top = Res_Y - (sy * 3); }

      rval = lpDDS_Blit->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
      if(rval != DD_OK) return;
      Blit_FS((LPBYTE)ddsd.lpSurface + ((ddsd.lPitch * RectSrc.top) >> 1) + RectSrc.left, ddsd.lPitch, sx, sy, (336 - sx) << 1);
      lpDDS_Blit->Unlock(NULL);

      if(VSync_FS) lpDDS_Primary->Flip(NULL, DDFLIP_WAIT);
    }
  }

  if(rval == DDERR_SURFACELOST) RestoreGraphics();
}


void __stdcall Flip_DDraw_W(HWND hWnd)
{
  RECT           RectSrc, RectDest;
  int            sx, sy;
  HRESULT        rval;
  DDSURFACEDESC2 ddsd;
  POINT          p;

  memset(&ddsd, 0, sizeof(DDSURFACEDESC2));
  ddsd.dwSize = sizeof(DDSURFACEDESC2);

  if(SMS_Started && !Debug) // SMS
  {
    if(Flag_Clr_Scr != 24)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 24; }

    sx = 256; sy = 196;
  }
  else if((VDP_Reg.Set4 & 0x01) || Debug)
  {
    if(Flag_Clr_Scr != 40)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 40; }

    sx = 320;  sy = VDP_Num_Vis_Lines;
  }
  else
  {
    if(Flag_Clr_Scr != 32)
    { Clear_DDraw_Primary_Screen(hWnd);
      Clear_DDraw_Back_Screen(hWnd);
      Flag_Clr_Scr = 32; }

    sx = 256; sy = VDP_Num_Vis_Lines;
  }

  p.x = p.y = RectSrc.left = RectSrc.top = 0;
  GetClientRect(hWnd, &RectDest);
  ClientToScreen(hWnd, &p);

  if(Render_W > 13)     { RectSrc.right = sx * 3;  RectSrc.bottom = sy * 3; }
  else if(Render_W > 1) { RectSrc.right = sx << 1; RectSrc.bottom = sy << 1; }
  else                  { RectSrc.right = sx;      RectSrc.bottom = sy; }

  if(Stretch_W)
  { RectDest.left    = p.x;  RectDest.top     = p.y;
    RectDest.right  += p.x;  RectDest.bottom += p.y; }
  else
  {
    if(Render_W == 1)
    { RectDest.left    = ((RectDest.right - (RectSrc.right << 1)) >> 1) + p.x;
      RectDest.top     = ((RectDest.bottom - (RectSrc.bottom << 1)) >> 1) + p.y;
      RectDest.right   = RectDest.left + (RectSrc.right << 1);
      RectDest.bottom  = RectDest.top + (RectSrc.bottom << 1); }
    else
    { RectDest.left    = ((RectDest.right - RectSrc.right) >> 1) + p.x;
      RectDest.top     = ((RectDest.bottom - RectSrc.bottom) >> 1) + p.y;
      RectDest.right   = RectDest.left + RectSrc.right;
      RectDest.bottom  = RectDest.top + RectSrc.bottom; }
  }


  if(Render_W > 1 || Blit_Soft)
  {
    rval = lpDDS_Back->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
    if(rval != DD_OK) return;
    Blit_W((LPBYTE)ddsd.lpSurface, ddsd.lPitch, sx, sy, (336 - sx) << 1);
    lpDDS_Back->Unlock(NULL);
  }
  else { RectSrc.right += 8; RectSrc.left = 8; }

  if(VSync_W)
  {
    int vb;
    lpDD->GetVerticalBlankStatus(&vb);
    if(!vb) lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);
  }
  rval = lpDDS_Primary->Blt(&RectDest, lpDDS_Back, &RectSrc, DDBLT_WAIT | DDBLT_ASYNC, NULL);
  if(rval == DDERR_SURFACELOST) rval = RestoreGraphics();
}