
#include "ggdi.h"
#include "vdp_io.h"
#include "sms.h"

HDC     hdcComp  = NULL;    // DC associado ao bitmap
HBITMAP hbmpDraw = NULL;    // bitmap usado como buffer para imprimir a imagem
LPBYTE  pbmpData = NULL;    // buffer
SIZE    szGDIBuf;

void __stdcall GDI_Error(HWND hWnd, LPCSTR pszMessage)
{
  End_GDI();
  MessageBox(hWnd, pszMessage, "Error GDI", MB_ICONSTOP);
  DestroyWindow(hWnd);
  exit(0);
}


int __stdcall Init_GDI(HWND hWnd)
{
  BITMAPINFOHEADER bih;

  if(Full_Screen) return 0; // gdi no  vlido para full screen

  if(Render_W < 1)       { szGDIBuf.cx = 320; szGDIBuf.cy = 240; }  // render 1x
  else if(Render_W < 14) { szGDIBuf.cx = 640; szGDIBuf.cy = 480; }  // render 2x
  else                   { szGDIBuf.cx = 960; szGDIBuf.cy = 720; }  // render 3x

  // criando DC
  hdcComp = CreateCompatibleDC(GetDC(hWnd));
  if(!hdcComp)
  { GDI_Error(hWnd, "Error CreateCompatibleDC."); return 0; }

  // tentando criar o bitmap que ser usado como buffer
  memset(&bih, 0, sizeof(BITMAPINFOHEADER));
  bih.biSize     = sizeof(BITMAPINFOHEADER);
  bih.biPlanes   = 1;
  bih.biBitCount = 16;
  //bih.biBitCount = 32;	//MK2k: Set to 32bits per pixel in order to avoid display mode change upon start
  bih.biWidth    = szGDIBuf.cx;
  bih.biHeight   = -szGDIBuf.cy;
  hbmpDraw = CreateDIBSection(hdcComp, (BITMAPINFO *)&bih, DIB_RGB_COLORS,
    (LPVOID *)&pbmpData, NULL, 0);
  if(!hbmpDraw)
  { GDI_Error(hWnd, "Error CreateDIBSection."); return 0; }
  SelectObject(hdcComp, hbmpDraw);

  Mode_555 = 1;
  Recalculate_Palettes();

  return 1;
}


void __stdcall End_GDI(void)
{
  if(hdcComp)  { DeleteDC(hdcComp); hdcComp = NULL; }
  if(hbmpDraw) { DeleteObject(hbmpDraw); hbmpDraw = NULL; pbmpData = NULL; }
}


void __stdcall Flip_GDI(HWND hWnd)
{
  HDC         hdcDest;
  int         sx, sy;
  RECT        RectDest;

  GetClientRect(hWnd, &RectDest);
  hdcDest = GetDC(hWnd);

  if(SMS_Started && !Debug) // SMS
  {
    if(Flag_Clr_Scr != 24)
    { InvalidateRect(hWnd, NULL, FALSE);
      FillRect(hdcDest, &RectDest, (HBRUSH)GetStockObject(BLACK_BRUSH));
      if(pbmpData) memset(pbmpData, 0, (szGDIBuf.cx << 1) * szGDIBuf.cy);
      Flag_Clr_Scr = 24; }

    sx = 256; sy = 196;
  }
  else if((VDP_Reg.Set4 & 0x01) || Debug)
  {
    if(Flag_Clr_Scr != 40)
    { InvalidateRect(hWnd, NULL, FALSE);
      FillRect(hdcDest, &RectDest, (HBRUSH)GetStockObject(BLACK_BRUSH));
      if(pbmpData) memset(pbmpData, 0, (szGDIBuf.cx << 1) * szGDIBuf.cy);
      Flag_Clr_Scr = 40; }
    sx = 320;  sy = VDP_Num_Vis_Lines;
  }
  else
  {
    if(Flag_Clr_Scr != 32)
    { InvalidateRect(hWnd, NULL, FALSE);
      FillRect(hdcDest, &RectDest, (HBRUSH)GetStockObject(BLACK_BRUSH));
      if(pbmpData) memset(pbmpData, 0, (szGDIBuf.cx << 1) * szGDIBuf.cy);
      Flag_Clr_Scr = 32; }
    sx = 256; sy = VDP_Num_Vis_Lines;
  }

  Blit_W(pbmpData, szGDIBuf.cx << 1, sx, sy, (336 - sx) << 1);
  if     (Render_W > 13) { sx *= 3;  sy *= 3;  }
  else if(Render_W > 0)  { sx <<= 1; sy <<= 1; }

  if(Stretch_W)
  {
    StretchBlt(hdcDest, 0, 0, RectDest.right, RectDest.bottom, hdcComp, 0, 0, sx, sy, SRCCOPY);
  }
  else
  {
    RectDest.left = (RectDest.right - sx) >> 1;
    RectDest.top  = (RectDest.bottom - sy) >> 1;
    BitBlt(hdcDest, RectDest.left, RectDest.top, sx, sy, hdcComp, 0, 0, SRCCOPY);
  }

  ReleaseDC(hWnd, hdcDest);
}


int  __stdcall Clear_GDI_Primary_Screen(HWND hWnd)
{
  if(pbmpData) memset(pbmpData, 0, (szGDIBuf.cx << 1) * szGDIBuf.cy);
  return 0;
}


int  __stdcall Clear_GDI_Back_Screen(HWND hWnd)
{
  HDC  hdcDest;
  RECT RectDest;
  hdcDest = GetDC(hWnd);
  InvalidateRect(hWnd, NULL, FALSE);
  GetClientRect(hWnd, &RectDest);
  FillRect(hdcDest, &RectDest, (HBRUSH)GetStockObject(BLACK_BRUSH));
  return 0;
}
