; scale2x.asm - este arquivo  parte do projeto gens+
; Rodrigo Cardoso
; Utiliza o algortimo scale2x para dobrar o tamanho da imagem no monitor

%include "nasmhead.inc"

section .text align=64

extern MD_Screen

%define REG_PUSH 4
%define DEST     esp + ((REG_PUSH + 1) * 4)
%define PITCH    esp + ((REG_PUSH + 2) * 4)
%define X        esp + ((REG_PUSH + 3) * 4)
%define Y        esp + ((REG_PUSH + 4) * 4)
%define OFFSET   esp + ((REG_PUSH + 5) * 4)
%define SOURCE   MD_Screen + 16


%macro DUP_LINE 0
  ALIGN4
  %%loop:
    mov  eax, [esi]
    add  esi, 4
    mov  edx, eax
    rol  eax, 16
    xchg ax, dx
    mov  [edi], eax
    mov  [edi + 4], edx
    mov  [edi + ebx], eax
    mov  [edi + ebx + 4], edx
    add  edi, 8
    dec  ecx
    jnz  %%loop
%endmacro


%macro DUP_PIXEL 0
  mov  dx, [esi]
  add  esi, 2
  mov  ax, dx
  shl  eax, 16
  mov  ax, dx
  mov  [edi], eax
  mov  [edi + ebx], eax
  add  edi, 4
%endmacro


%define B  esi + ebp
%define D  esi - 2
%define E  esi
%define F  esi + 2
%define H  esi + ebp

%define E0 edi
%define E1 edi + 2
%define E2 edi + ebx
%define E3 edi + ebx + 2

; se B == H ou D == F pula para DO_2
%macro CHECK_1 2
  mov  eax, [D + %1]         ; eax = DE
  cmp  ax,  [F + %1]         ; D == F ?
  jz   %2
  neg  ebp
  mov  dx,  [B + %1]         ; dx = B
  neg  ebp
  cmp  dx,  [H + %1]         ; B == H ?
  jz   %2
%endmacro

; se B == D pula para DO_3
%macro CHECK_2 2
  cmp  ax, dx                ; B == D ?
  jz   %2
%endmacro

; se H == F pula para DO_4
%macro CHECK_3 2
  mov  eax, [E + %1]         ; eax = EF
  ror  eax, 16               ; eax = FE
  cmp  ax,  [H + %1]         ; H == F ?
  jz   %2
%endmacro

; caso nenhuma das comparaes sejam verdadeiras
%macro DO_1 1
  cmp  ax, dx                ; B == F ?
  jz   %%set1
  shr  eax, 16               ; ax = E
  %%set1:
  mov  [E1 + (%1 * 2)], ax   ; E1
  mov  eax, [D + %1]         ; eax = DE
  mov  edx, eax              ; edx = DE
  cmp  ax,  [H + %1]         ; D == H ?
  jz   %%set2
  shr  eax, 16               ; ax = E
  %%set2:
  mov  [E2 + (%1 * 2)], ax   ; E2
  shr  edx, 16               ; dx = E
  mov  [E0 + (%1 * 2)], dx   ; E0 = E
  mov  [E3 + (%1 * 2)], dx   ; E3 = E
%endmacro

; caso B == H || D == F
%macro DO_2 1
  mov  edx, eax
  shr  edx, 16
  mov  ax, dx
  mov  [E0 + (%1 * 2)], eax  ; E0 e E1 = E
  mov  [E2 + (%1 * 2)], eax  ; E2 e E3 = E
%endmacro

; caso B != H && D != F && B == D
%macro DO_3 1
  mov  [E0 + (%1 * 2)], eax  ; E0 = D e E1 = E
  shr  eax, 16
  mov  [E2 + (%1 * 2)], ax   ; E2 = E
  mov  eax, [E + %1]         ; eax = EF
  ror  eax, 16
  cmp  ax,  [H + %1]         ; H == F ?
  jz   %%set
  shr  eax, 16
  %%set:
  mov  [E3 + (%1 * 2)], ax   ; E3
%endmacro

%macro DO_4 1
  mov  edx, eax              ; edx = FE
  ror  eax, 16               ; eax = EF
  mov  [E2 + (%1 * 2)], eax  ; E2 = E e E3 = F
  mov  dx, ax                ; edx = EE
  mov  [E0 + (%1 * 2)], edx  ; E0 e E1 = E
%endmacro

ALIGN64

; --------------------------------------------------------------------------------------
; void __cdecl Blit_Scale2x(LPBYTE dest, int pitch, int x, int y, int offset)
; usa o algortimo scale2x para dobrar o tamanho da imagem na tela (sem MMX)
DECL Blit_Scale2x
  push ebx
  push ebp
  push edi
  push esi

  mov  edi, [DEST]            ; destino
  lea  esi, [SOURCE]          ; origem
  mov  ecx, [X]               ; nmero de pixels por linha na origem
  lea  ebp, [ecx * 2]         ; ebp = tamanho em bytes de uma linha na origem (impresso)
  add  ebp, [OFFSET]          ; tamanho em bytes de uma linha na origem
  shl  ecx, 2                 ; nmero de bytes impressos por linha no destino
  mov  ebx, [PITCH]           ; ebx = tamanho em bytes de uma linha no destino
  mov  edx, ebx
  sub  edx, ecx               ; parte da linha do destino que no  impressa
  add  edx, ebx
  mov  [PITCH], edx
  sub  dword [X], 2           ; excluindo o primeiro e ltimo pixel da cada linha
  shr  dword [X], 1           ; divide por dois, pois cada etapa do loop faz dois pixels


  ; fazendo a primeira linha (simplesmente duplica os pixels)
  shr  ecx, 3
  DUP_LINE

  add  esi, [OFFSET]
  add  edi, [PITCH]
  sub  dword [Y], 2           ; excluindo a primeira e a ltima linha do loop principal
  jmp  .loop_y

  ALIGN64
  .loop_y:

    ; fazendo o primeiro pixel da linha (duplica)
    DUP_PIXEL
    mov  ecx, [X]             ; nmero de pixels por linha na origem
    jmp  .loop_x

    ALIGN64
    .loop_x:

      CHECK_1 0, .do_2
      CHECK_2 0, .do_3
      CHECK_3 0, .do_4
      DO_1    0
      jmp  .next_pixel
      ALIGN4
      .do_2: DO_2 0
      jmp  .next_pixel
      ALIGN4
      .do_3: DO_3 0
      jmp  .next_pixel
      ALIGN4
      .do_4: DO_4 0
      jmp  .next_pixel

      ALIGN64
      .next_pixel:

      CHECK_1 2, .2do_2
      CHECK_2 2, .2do_3
      CHECK_3 2, near .2do_4
      DO_1    2
      add  esi, 4
      add  edi, 8
      dec  ecx
      jnz  .loop_x
      jmp  .end_loop_x
      ALIGN4
      .2do_2: DO_2 2
      add  esi, 4
      add  edi, 8
      dec  ecx
      jnz  .loop_x
      jmp  .end_loop_x
      ALIGN4
      .2do_3: DO_3 2
      add  esi, 4
      add  edi, 8
      dec  ecx
      jnz  .loop_x
      jmp  .end_loop_x
      ALIGN4
      .2do_4: DO_4 2
      add  esi, 4
      add  edi, 8
      dec  ecx
      jnz  .loop_x
      jmp  .end_loop_x

      ALIGN64
      .end_loop_x:

    ; fazendo o ltimo pixel da linha
    DUP_PIXEL

    add  esi, [OFFSET]
    add  edi, [PITCH]
    dec  dword [Y]
    jnz  .loop_y


  ; fazendo a ltima linha (simplesmente duplica os pixels)
  mov  ecx, [X]
  add  ecx, 2
  shr  ecx, 1
  DUP_LINE

  pop  esi
  pop  edi
  pop  ebp
  pop  ebx
  ret
; --------------------------------------------------------------------------------------


%macro DUP_MMX 2
    movq mm0, [%1]
    movq mm2, [%1 + 8]
    movq mm4, [%1 + 16]
    movq mm6, [%1 + 24]
    movq mm1, mm0
    punpcklwd mm1, mm1
    punpckhwd mm0, mm0
    movq mm3, mm2
    punpcklwd mm3, mm3
    punpckhwd mm2, mm2
    movq mm5, mm4
    punpcklwd mm5, mm5
    punpckhwd mm4, mm4
    movq mm7, mm6
    punpcklwd mm7, mm7
    punpckhwd mm6, mm6
    movq [%2 + 0],  mm1
    movq [%2 + 8],  mm0
    movq [%2 + 16], mm3
    movq [%2 + 24], mm2
    movq [%2 + 32], mm5
    movq [%2 + 40], mm4
    movq [%2 + 48], mm7
    movq [%2 + 56], mm6
    movq [%2 + ebx + 0],  mm1
    movq [%2 + ebx + 8],  mm0
    movq [%2 + ebx + 16], mm3
    movq [%2 + ebx + 24], mm2
    movq [%2 + ebx + 32], mm5
    movq [%2 + ebx + 40], mm4
    movq [%2 + ebx + 48], mm7
    movq [%2 + ebx + 56], mm6
    add %2, 64
    add %1, 32
%endmacro

; duplica a primeira (esi, edi) e a ultima linha (eax, edx)
%macro DUP_LINE_MMX 0
  ALIGN4
  %%loop:
    DUP_MMX esi, edi ; duplicando a primeira linha
    DUP_MMX eax, edx ; duplicando a ltima linha
    dec ecx
    jnz  %%loop
%endmacro

ALIGN64

; --------------------------------------------------------------------------------------
; void __cdecl Blit_Scale2x_MMX(LPBYTE dest, int pitch, int x, int y, int offset)
; scale2x mmx
DECL Blit_Scale2x_MMX
  push ebx
  push ebp
  push edi
  push esi

  mov  edi, [DEST]            ; destino
  lea  esi, [SOURCE]          ; origem
  mov  ecx, [X]               ; nmero de pixels por linha na origem
  lea  ebp, [ecx * 2]         ; ebp = tamanho em bytes de uma linha na origem (impresso)
  add  ebp, [OFFSET]          ; tamanho em bytes de uma linha na origem
  shl  ecx, 2                 ; nmero de bytes impressos por linha no destino
  mov  ebx, [PITCH]           ; ebx = tamanho em bytes de uma linha no destino
  mov  edx, ebx
  sub  edx, ecx               ; parte da linha do destino que no  impressa
  add  edx, ebx
  mov  [PITCH], edx
  shr  dword [X], 2           ; divide por quatro, pois cada etapa do loop faz quatro pixels


  ; fazendo a primeira e ltima linhas (simplesmente duplica os pixels)
  shr  ecx, 6
  mov  eax, [Y]
  dec  eax
  mov  edx, eax
  imul eax, ebp
  add  eax, esi
  imul edx, ebx
  shl  edx, 1
  add  edx, edi
  DUP_LINE_MMX

  add     esi, [OFFSET]
  add     edi, [PITCH]
  mov     edx, edi
  mov     eax, esi
  sub     dword [Y], 2        ; excluindo a primeira e a ltima linha do loop principal
  jmp     .loop_x

  ALIGN64
  .loop_x:

    neg  ebp
    movq mm0, [B]
    neg  ebp
    movq mm2, [E]
    mov  ecx, [Y]             ; nmero de linhas na imagem
    jmp  .loop_y

    ALIGN64
    .loop_y:

      movq      mm1, mm0      ; mm1 = B
      movq      mm0, mm2      ; mm0 = E
      movq      mm2, [H]      ; mm2 = H
      movq      mm3, [D]      ; mm3 = D
      movq      mm4, [F]      ; mm4 = F

      movq      mm5, mm1
      pcmpeqw   mm5, mm2      ; mm5 = B == H ?
      movq      mm6, mm3
      pcmpeqw   mm6, mm4      ; mm6 = D == F ?
      por       mm6, mm5
      pcmpeqd   mm5, mm5
      pxor      mm5, mm6      ; mm5 = B != H && D != F ?

      ; E0
      movq      mm6, mm1
      pcmpeqw   mm6, mm3      ; mm6 = B == D ?
      pand      mm6, mm5      ; mm6 = B == D && B != H && D != F ?
      pcmpeqd   mm7, mm7
      pxor      mm7, mm6      ; mm7 = ~mm6
      pand      mm6, mm1
      pand      mm7, mm0
      por       mm6, mm7      ; mm6 = E0
      ; E1
      pcmpeqw   mm1, mm4      ; mm1 = B == F ?
      pand      mm1, mm5      ; mm1 = B == F && B != H && D != F ?
      pcmpeqd   mm7, mm7
      pxor      mm7, mm1      ; mm7 = ~mm1
      pand      mm1, mm4
      pand      mm7, mm0
      por       mm1, mm7      ; mm1 = E1
      movq      mm7, mm6
      punpcklwd mm6, mm1
      punpckhwd mm7, mm1
      movq      [E0], mm6
      movq      [E0 + 8], mm7
      ; E2
      movq      mm1, mm2
      pcmpeqw   mm1, mm3      ; mm1 = D == H ?
      pand      mm1, mm5      ; mm1 = D == H && B != H && D != F ?
      pcmpeqd   mm7, mm7
      pxor      mm7, mm1      ; mm7 = ~mm1
      pand      mm1, mm2
      pand      mm7, mm0
      por       mm1, mm7      ; mm1 = E2
      ; E3
      pcmpeqw   mm4, mm2      ; mm4 = F == H ?
      pand      mm4, mm5      ; mm4 = F == H && B != H && D != F ?
      pcmpeqd   mm7, mm7
      pxor      mm7, mm4      ; mm7 = ~mm4
      pand      mm4, mm2
      pand      mm7, mm0
      por       mm4, mm7      ; mm2 = E3
      movq      mm7, mm1
      punpcklwd mm1, mm4
      punpckhwd mm7, mm4
      movq      [E2], mm1
      movq      [E2 + 8], mm7

      add  esi, ebp
      add  edi, ebx
      add  edi, ebx
      dec  ecx
      jnz  .loop_y

    add  edx, 16
    mov  edi, edx
    add  eax, 8
    mov  esi, eax
    dec  dword [X]
    jnz  .loop_x


  pop  esi
  pop  edi
  pop  ebp
  pop  ebx
  emms
  ret
; --------------------------------------------------------------------------------------
